/* qartus.h
 * FORTH-79 Standard with Double-Number-Standard Extensions (+ qlibrary) 
 * Copyleft 2011-2020
 * Antonio Maschio <ing dot antonio dot maschio at gmail dot com>
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * It's GPL! Enjoy! */

#include <stdio.h>
#include <stdlib.h>
#include <readline/readline.h>
#include <readline/history.h>
#include <string.h>
#include <termios.h>
#include <math.h>
#include <unistd.h>
#include <signal.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/time.h>
#include <errno.h>
#include <dirent.h>
#include<sys/socket.h>
#include<arpa/inet.h>	//inet_addr
#include <pwd.h>
#include "custom.h"
#include "wordlist.h"
#include "errors.h"

/***************************************************************************/

#define VERSION_MAJ	1			// versioning data
#define VERSION_MIN	0
#define VERSION_STR     "Alpha"		
#define YEARS		"2011-2020"		
#define ADDRLIMIT	65535+1			// default memory dimension
#define FALSE 		0			// truth values of Forth-79
#define TRUE 		1

#define ILINE 		256			// max length in input line
#define CHUNKS		255			// file arrays capacity
#define FILENAMELEN	255			// max characters in file name
#define FILELIMIT       6			// max number of open files
#define SCREEN		75			// screen length for outputs
#define PROMPTDIM	16			// max length of prompt
#define PROMPT 		""			// default prompt (void)
#define BL		32			// blank ASCII code
#define ESCL		127			// blank ASCII code
#define LPS             15                      // lines per screen
#define CPL             64      		// characters per line
/* memory organization */
#define WORDNAMELENBASE	31
#define PADLEN		84			// PAD's default dimension
#define PADMAX		1024			// PAD's maximum dimension
#define PADWIDTH	200                     // PAD's offset from HERE
#define QUERYLEN	80			// QUERY requirement
#define BUFFERSLENGTH	512			// mem buffers dimension
#define BYTE		256			// HEX 100
#define FFCHAR		255			// HEX FF
#define CHAR		127			// HEX 7F
#define DOUBLE		65536			// limit for unsigned int 2^16
#define TEMPBUFFER	512			// buffers dimension
#define ZDSSTACKLENGTH	2048			// data stack dim
#define ZRSSTACKLENGTH	2048			// return stack dim
#define WDSSTACKMAX     32767			// data stack max limit
#define WRSSTACKMAX     16384			// return stack max limit
#define CFSTACKLENGTH	9999999+1		// control-flow stack dim
#define VOCS            32			// number of vocabularies
#define BASELOW		2			// lower limit for base 
#define BASELIMIT	36			// higher limit for base
#define EDITORLINE      65			// editor line length (+1)
#define B_BUF		1024			// block buffer width
#define INTERPRET	0			// interpreter mode (STATE=0)
#define COMPILE		1			// compile mode (STATE=1)
#define MARKVAR		0xaacaada		// mark start of variable data
#define DOESMARK	0xdaafaac		// mark does> portion in prog
#define VOC_MARK	0xceafacc		// mark vocabulary in program
#define OK 		" ok"			// standard ok
#define OK_COMP		" (compiled)"		// ok for a compiling word
#define SINGLE		32768	 		// limit for signed int 2^16
#define UDOUBLE		"4294967295 "		// limit for unsigned int 2^32
#define LPDOUBLE	"2147483647 " 		// limit for signed int 2^31
#define LNDOUBLE 	"-2147483648 "		// limit for double int -2^31
#define ALL		0	// used in erasestacks
#define DATA_STACK	1	// erase only data stack
#define RETURN_STACK	2	// erase only return stack
#define R_O		1	// file access mode Read-Only
#define W_O		2	// file access mode Write-Only
#define PORTNUM		64	// number of ports
/* Precedence bits */
#define B_NULL		0	// no bits on
#define B_PREC		1	// precedence bit (1=IMMEDIATE)
#define B_SMUDGE	2	// smudge bit
#define B_BUFFER	4	// word is buffered (needs following arg)
#define B_COMPILE	8	// word is compile-only
#define B_DEFINING	16	// word is a defining word
#define B_ADDCOMP       32	// word needs more control while compiling
#define B_EXECUTION     64	// word is exec-only
#define RES_NAME	".qartus.rc"
#define HID_PATH        ".qartus/"
#define HISTORY_NAME	".qartus_history"	// readline facility
#define BLOCKS_NAME     "blocks/"
#define LIBS_NAME	"qlibrary.ft"
/* screen picture */
#define INVCOLOR 	"[7;37;40m"
#define RESETCOLOR 	"[1;m"
#define REDCOLOR        "[1;31m"
/* EDIT status lines */
#define VOIDLINE 	"                                                                "
#define ISTAMP   	"\\ Program name        Screen name        qartus %3s %02d-%3s-%4d"		// add 3-char INITIALS; 2-char day; 3-char month; 4-char year
#define EHEADER  	"BLOCK:%5d (%d)  %3s  LINE: %02d  POS: %02d  %s  DATE: %02d-%3s-%4d  \r\n" // addr block, BFR, INS state (insert/overwrite), eline, epos, day, month, year
#define EFOOTER  	"   CTRL-Q=Quit   CTRL-S=Select   CTRL-C=Copy   CTRL-V=Paste      \r\n"
#define TFOOTER  	"Text not found                                                   \r\n"
#define FFOOTER  	"Search:                                                          \r\n" // epos=8,eline=16
#define RFOOTER  	"Replace:                                                         \r\n" // epos=9
#define BFOOTER  	"Block:                                                           \r\n" // epos=7
#define EASTEREGG 	"   Welcome to the quartus FORTH editor! I hope you'll like it!   \r\n"
#define SFOOTER  	"Block file %5d was saved to file system.                       \r\n"
#define ZFOOTER  	"Block file was removed from file system.                         \r\n"
#define ZFOOTERQ  	"Are you sure? (Y/N)  Block %5d will be deleted:                \r\n"
#define ONWARD	 	10
#define BACKWARD 	20
/* speaker devices  (default SPEAKER1) */
#define SPEAKER1        system("play -q -n synth 0.14 sin 880 &")
#define SPEAKER2        system("( speaker-test -t sine -f 880 >/dev/null)& pid=$! ; sleep 0.14s ; kill -9 $pid")
#define SPEAKER3	fputc(7,stdout)
#define SPEAKER4	system("echo > /dev/null")
#define SPEAKER5	system("call")
/* sound devices  (default SOUND1) */
/* Notes for play:
 * -q means --no-show-progress
 * -n means --null (no file) 
 * synth is the effect (sinusoidal synthesis) 
 * %f is the length in seconds (qartus asks for milliseconds to refine 
 * the pitch duration)
 * sin is the generative tone 
 * %d is the frequency */
#define SOUND1		snprintf(comm,ILINE,"play -q -n synth %f sin %d",rdur,freq)
/* Notes for the speaker-test:
 * -t means type sine
 * -f means frequency %d 
 * %f is the length in seconds used with the sleep tool (qartus asks for 
 * milliseconds to refine the pitch duration */
#define SOUND2		snprintf(comm,ILINE,"( speaker-test -t sine -f %d >/dev/null)& pid=$! ; sleep %fs ; kill -9 $pid",freq,rdur)
/* Note: echo should work even in Windows (though the string "/dev/null" appears
 * in the output...)*/
#define SOUND3		snprintf(comm,ILINE,"echo > /dev/null") 
#define SOUND4		snprintf(comm,ILINE,"call") 

/* memory macros */
#define prev(X) 	(mem[(X)])		// addr of previous word
#define cbits(X) 	(mem[(X)+1])		// control bits
#define voccode(X)	(mem[(X)+2])		// vocabulary code
#define wordlen(X) 	(mem[(X)+4])		// length of word
#define wordstart(X)	((X)+5)			// addr of first word char
#define codestart(X)	((X)+5+mem[(X)+4])	// addr of first code token
/* system vars locators */
#define BASE		mem[11]
#define STATE		mem[25]
#define to_in		((unsigned short) (mem[37]+BYTE*mem[38]))
#define to_in_(X)       (mem[37]=(int)((X)%BYTE),mem[38]=((X)/BYTE))
#define CSP		((unsigned short) (mem[49]+BYTE*mem[50]))
#define CSP_(X)         (mem[49]=(int)((X)%BYTE),mem[50]=((X)/BYTE))
#define SPAN		mem[62]
#define BLK1		mem[74]
#define BLK		mem[75]
#define SCR1		mem[86]
#define SCR		mem[87]
#define BFR		mem[98]
#define CONTEXT 	((unsigned short) (mem[114]+BYTE*mem[115]))
#define CONTEXT_(X) 	(mem[114]=(int)((X)%BYTE),mem[115]=((X)/BYTE))
#define CURRENT 	((unsigned short) (mem[130]+BYTE*mem[131]))
#define CURRENT_(X) 	(mem[130]=(int)((X)%BYTE),mem[131]=((X)/BYTE))
#define ROW_GRID	mem[141]	// #L
#define POS_GRID	mem[152]	// #P
#define DPL		mem[164]
#define FLD		mem[176]
#define FENCE1		mem[190]
#define FENCE		mem[191]
#define LAST		((unsigned short) (mem[203]+BYTE*mem[204]))
#define LAST_(X)        (mem[203]=(int)((X)%BYTE),mem[204]=((X)/BYTE))
#define SHARP_TIB1	mem[216]	// #TIB
#define SHARP_TIB	mem[217]
#define RSHARP1		mem[227]	// R#
#define RSHARP		mem[228]
#define _PC		((unsigned short) (mem[238]+BYTE*mem[239]))
#define PC_(X)          (mem[238]=(int)((X)%BYTE),mem[239]=((X)/BYTE))
#define OFFSET1		mem[253]
#define OFFSET		mem[254]
#define OUT		mem[265]
#define WARNING		mem[281]
#define BUFFERS		mem[298]
#define NVOCS		mem[312]
#define ERR1		mem[324]
#define ERR		mem[325]
#define WORDNAMELEN	mem[338]
#define HLD1		mem[350]
#define HLD		mem[351]
#define isFigCompatEnabled mem[368]
#define isDebug         mem[388]
#define ip1         	mem[400]
#define ip2         	mem[412]
#define ip3         	mem[424]
#define ip4         	mem[436]
#define NUMCLOCK	mem[451]
/* build macros */
#define insertvar(T,A,B,C,D,E) PREVHERE=HERE;LAST_(HERE);HERE=_add_var(T,HERE,hb,A,B,C,D,E);hb=PREVHERE;if(HERE>HIGHMEM-100)_perror(30,to_in)
#define insertconst(A,B,C,D,E) PREVHERE=HERE;LAST_(HERE);HERE=_add_var('k',HERE,hb,A,B,C,D,E);hb=PREVHERE;if(HERE>HIGHMEM-100)_perror(30,to_in)
#define insertbui(T,A,B,C,D,E) PREVHERE=HERE;LAST_(HERE);HERE=_add_word(T,HERE,hb,A,B,C,D,E);hb=PREVHERE;if(HERE>HIGHMEM-100)_perror(30,to_in)
#define line_buffer(X)  mem[TERMINALINPUTBUFFERSTART+(X)] // buffer management facility
/* vocabulary macros */
#define UWD		220	// user word code (was 85)
#define BWD		230	// builtin word code (was 95)
#define	isImm(X)	(cbits(X)&B_PREC)
#define	isSmudge(X)	(cbits(X)&B_SMUDGE)
#define	isBuffered(X)	(cbits(X)&B_BUFFER)
#define	isComp(X)	(cbits(X)&B_COMPILE)
#define	isExec(X)	(cbits(X)&B_EXECUTION)
#define	isDef(X)	(cbits(X)&B_DEFINING)
#define	isAddComp(X)	(cbits(X)&B_ADDCOMP)
#define codeup(X)	mem[HERE++]=BWD,mem[HERE++]=(X)
/* general macros */
#define isBlank(i)	((i)==32 || (i)=='\t')
#define isFinal(i)	((i)==32 || (i)=='\t' || !(i))
#define dobfr   	if(!BFR)return
#define CRS		printf("\n    ")
#define NOCRS 		printf("\b\b\b\b")
#define STDIN 10
#define STDOUT 20
#define STDERR 30
#define _doerr(A,N);	if ((A)) {_perror((N),(to_in));return;}	
/* stack pop macros */
#define _pop()          _POP();if(isQuit||isInterrupt)return
#define _rspop()        _RSPOP();if(isQuit||isInterrupt)return
#define _cfpop()        _CFPOP();if(isQuit||isInterrupt)return

/* system strings */
static char *month[]=
{"", "Jan","Feb","Mar","Apr","May","Jun","Jul","Aug","Sep","Oct","Nov","Dec" };
char helpScreen[]= "\
\\ SCREEN EDITOR COMMANDS:        - Use CTRL + the following:    \
\\  A select all text             B Edit next block              \
\\  C copy selected text          D delete current line          \
\\  E edit another block          F find string                  \
\\  G repeat last search          H show this help screen        \
\\  I insert an empty line        J set insert/overwrite         \
\\  K empty current line          L step to Linear Editor        \
\\  N clear current buffer        O reload buffer from disk      \
\\  P default stamp               Q quit                         \
\\  R replace a string            S toggle selection state       \
\\  T put the cursor at 0,0       U undo last operation          \
\\  V paste selected text         W write to disk                \
\\  X cut selected text           Z delete block file            \
\\  BACKSPACE delete left char    CANC delete character          \
\\  END go to end of line         ESC return from help screen    \
\\  HOME go to start of line      arrows: move around the text   ";

/* system struct entities */
/* block structure */
struct t_buffer {
	int start;		// first byte
	int end;		// last byte
	int in_use;		// is it engaged?
	int order;		// order in reading 
	unsigned int number;	// block number (and name)
	int update;		// TRUE=need saving; FALSE=saved
};
typedef struct t_buffer TBuffer;

/* user vocabulary structure */
struct t_voc {
	unsigned short code;	// presently unused
	char name[WORDNAMELENBASE];
};
typedef struct t_voc TVoc;

/* file structure */
struct t_yfile {
	int state;		// channel state (in use or not)
	FILE * fd;
};
typedef struct t_yfile TFILE;

/* EDIT structure */
struct abuf {
	char *b;
	int len;
};

/* PORT struct entities */
struct t_port {
	int port_number;
	int socket_desc;
	struct sockaddr_in server;
};
typedef struct t_port TPort;

#define ABUF_INIT {NULL, 0}
#define CTRL_KEY(k) ((k) & 0x1f)

enum editorKey {
	BACKSPACE = 127,
	ARROW_LEFT = 1000,
	ARROW_RIGHT,
	ARROW_UP,
	ARROW_DOWN,
	DEL_KEY,
	HOME_KEY,
	END_KEY,
	PAGE_UP,
	PAGE_DOWN,
	CTRL_BAR
};


/* user input variables */
int DSSTACKLENGTH=ZDSSTACKLENGTH;		// data stack/return stack dim
int RSSTACKLENGTH=ZRSSTACKLENGTH;		// data stack/return stack dim
int WCOUNT=0;
int shour,smin,ssec;				// starting time values
int syear,smon,sday;				// starting day values
int newline=0;		 // default for option --newline=<n>
int LAST_BUILTIN=0;
int EXIT_CODE=0;
char iline[ILINE+1];
char cline[ILINE+1];
char nline[ILINE+1];
char* inputline=iline;
char word[ILINE+1];
char *w=word;
char backbuffer[B_BUF+1];
char *b=backbuffer;
int isLoad=FALSE;
int isBye=FALSE;			// if true, exit qartus
int isMinusTo=FALSE;			// if true, stop LOAD and set for next
int isSemiS=FALSE;			// if true, stop LOAD
int wasTrace=FALSE;
int isQuit=FALSE;
int compile_ON=FALSE;			// COMPILE action
int isOpen=FALSE;			// if TRUE a ( comment was found
char prompt[PROMPTDIM+1];
int isPrintSeparator=FALSE;
int isPict=FALSE;			// if true, we are between <# and #>
long long int beginS, beginM;
struct timeval tv; 
int isStat=FALSE;
int isRequire=FALSE;
int isRes=FALSE;
int isExecute=FALSE;
int provokeSTATE=FALSE;
/* stacks variables */
int ds=0,cfs=0,rs=0;
int backcfs=0;				// for colon() and semicolon()	
/* dictionary variables */
int mem[ADDRLIMIT];			// the memory
int cfmem[CFSTACKLENGTH];		// the control flow indep. memory
int HERE=1;				// next free mem cell pointer
int PREVHERE=0;
int countWords=0;
int isSetup=FALSE;
char compiling[ILINE+1]; 
int picp=0;
int padlen=PADLEN;
unsigned short currentDef=0;
int loops=0;				// loops counter
int currPC=0;
int isExists=FALSE;
int isInterrupt=FALSE;
int doesmark=FALSE;
int doDoes=FALSE;
unsigned short doesaddr=0;
/* debugging */
unsigned short traceVal=0;
int stopKey=FALSE;
/* file variables */
char path[FILENAMELEN];
char filearray[CHUNKS][FILENAMELEN];
int fc=0, fr=0;		// file counter
char xhomedir[4*ILINE+1];	// home directory
char currdir[4*ILINE+1];	// current directory
char BASEHOME[4*ILINE+1];	// $HOME path
char HID_HOME[4*ILINE+1];	// BASEHOME + HID_PATH ($HOME/.qartus/)
char RES_PATH[4*ILINE+1];	// resouce file default path
char HISTORY_PATH[4*ILINE+1];	// readline history path
char BLOCKS_PATH[4*ILINE+1];	// default blocks path
char LIBS_PATH[4*ILINE+1];	// default library path
TFILE yfile[FILELIMIT+1];
/* other */
struct termios base_termios; //registered by main
struct termios raw, orig_termios; // EDIT
char digits[] = "0123456789abcdefghijklmnopqrstuvwxyz";
struct sigaction new_action;
struct sigaction old_action;
int exitOnFile=TRUE;
int qr=0;
/* blocks */
TBuffer _buffer_data[4];	// used 1,2,3
unsigned int last_block_accessed=0;
unsigned int last_word_with_success=0;
unsigned int LOWVAR=0, LOWMEM=0, HIGHMEM=0;	// limits of the user space	
unsigned int WORDBUFFERSTART, WORDBUFFERLIMIT;
unsigned int TERMINALINPUTBUFFERSTART, TERMINALINPUTBUFFERLIMIT;
unsigned int SCRATCHAREASTART, SCRATCHAREALIMIT, SC=0;	 // scratch area pointer
unsigned int DATASTACKSTART, DATASTACKLIMIT;
unsigned int RETURNSTACKSTART, RETURNSTACKLIMIT;
unsigned int CFSTACKSTART, CFSTACKLIMIT;
int isFile=FALSE; 
int isBlock=FALSE; 
int isStopFile=FALSE; 
int isError=FALSE;
int RunningErrorInstance=FALSE;
int isInterpret=FALSE;
int isExitEdit=FALSE;
int epos=0, eline=0;
int isPage=FALSE;
int reqOffset=0;
int isTotalTiming=FALSE;
long long int mbeginS, mendS;
/* ports*/
TPort yport[PORTNUM+1];
int portTOS=0;
/* vocabularies */
#define VFORTH	0
#define VEDITOR 1
int vocCount=0; // the default editors are 0=FORTH and 1=EDITOR
int vocChain[VOCS+1]; 
TVoc vocs[VOCS+1];

/* functions prototypes */
void _forth(void);
void _definitions(void);
void __exit(unsigned short a);
void _plus(void);
void _dot(void);
void _pad(void);
void _bye(void);
void _word_ender(void);
void _push(int c);
int _POP(void);
void _cfpush(int c);
int  _CFPOP(void);
int  _cfcopy(void);
void _rspush(int c);
int  _RSPOP(void);
void _interpret(void);
void _ok(void);
int  _add_word(int t, unsigned short w, unsigned short a, int L, int CB, int V, char *n, int c);
int  _add_var(int t, unsigned short w, unsigned short a, int L, int CB, int V, char *n, int v);
int  _add_2_var(int t, unsigned short w,unsigned short a, int L, int CB, int V, char *n, int v1, int v2);
void _add(void);
void _setup(void);
void _do_see(char *s, unsigned short a);
int _fill_buff(char *s);
void _uprint_s(void);
void _uprint_r(void);
void _print_cf(void);
int  _compare(unsigned short a, int l, char* w);
int  _ncompare(unsigned short a, const char* w, int l);
char *_dupstr(unsigned short a);
void _database_search(char* word);
void _exec(unsigned short addr);
void _exec_one(void);
void _exec_flow(unsigned short ad);
void _exec_imm(unsigned short ad);
void _builtin(unsigned short code);
void _execute(void);
void _evaluate_input_buffer(void);
void _evaluate_block_buffer(int i);
void _perror(int c, int p);
int _print(int val, int base, char *dest);
int _print_unsigned(int val, int base, char *d);
int _print_double_unsigned(int val, int base, char *d);
void _print_version(int g);
int  _get_and_push(char* w, int b);
long long int int2ll(long long int v1, long long int v2);
void ll2int(long long int n, long long int* h, long long int*l);
void _erase_stacks(int i);
int  _print_word_name(unsigned short h);
int  _print_builtin_name(unsigned short h);
int  _print_builtin_name(unsigned short h);
unsigned short  _find_word(char* w, int m);
void _idle(void);
char _dottize(int v);
void sigint(int s);
void setsig(void);
void _parse_file(char *f);
void _include(void);
void _require(int m);
void _erase_stacks(int flag);
int _isdigit(char v);
void _remove_nasty(char *);
char *dupstr(char *s);
char* tf_generator(const char *t, int s);
char* tf_fgenerator(const char *t, int s);
static char **tf_completion(const char *t, int s, int e);
void _over(void);
void qquit(int sig);
void update_buffer_order(void);
int choose_available_buffer(unsigned short b);
void put_line(int row, char *line);
void shift(int a, int b, int c);
void save_buffer(unsigned short blo, int buf);
void load_buffer(unsigned short blo, int buf);
void remove_block(unsigned short block);
void _block(void);
void _buffer(void);
void _m(void);
void _does_to(void);
void _tick(unsigned short addr);
void _find(void);
void print_long_banner(void);
void _print_license(void);
void _license(void);
void _add(void);
void _paren(void);
void _act_word(unsigned short a);
void _where(void);
void _dot_s(void);
void _inc_to_in_(void);
void _l(void);
void _editor(void);
void _print_reverse(int i);
void _print_(int i);
void _two_drop(void);
void _rot(void);
int editorReadKey(void);
struct tm *getTime(void);
int getDay(void);
int getMonth(void);
int getYear(void);
int getMin(void);
int getSec(void);
int getHour(void);
void editorRefreshScreen(int n);
int check_buffer(int c);
void _abort(void);
void _error(void);
void _question_error(void);
void _message(void);
FILE* _file_open(char* original);
void _check_required(int m, FILE *f, char *word, int* done);
void _load_required(FILE *, char *word);

/* editors data */
int isIns=FALSE;
int isSel=FALSE;
char sel[B_BUF+2];
char footer_buffer[64+5];
char undo_buffer[B_BUF+2];
int SelStart=0;
int SelEnd=0;
int isHelpScreen=FALSE;
char matchString[64+1];
char replaceString[64+1];
int isStepToLinear=FALSE;
int TextLine=0;
int TextPos=0;
int OTextLine=0;
int OTextPos=0;
int isTextInsert=FALSE;
int isBackSpace=FALSE;
int isBanner=TRUE;

/* logo string */
#define LOGO            "\n                   |\n                 _  _|\n    \\  __ \\   __\\  |  |   |   _\\\n  _\\ |  _\\ |  |    |  |_  |  \\ \n \\__ | \\___| _|   _| \\____/ \\__/\n     |\n    _|  Forth-79"
