/* qartus.c
 * Forth-79 Standard with Double-Number-Standard Extensions (+ qlibrary)
 * Copyleft 2011-2020
 * Antonio Maschio <ing dot antonio dot maschio at gmail dot com>
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * It's GPL! Enjoy! */

#include "qartus.h"

/************************************
 * USE and LIB
 ***********************************/

/* setblockdir() */
int setblockdir(char *new) {
	DIR *dir;
	struct stat st = {0};
	/* if input is an absolute path, set it as the block dir */
	if (*new=='/') strncpy(BLOCKS_PATH,new,ILINE*4);
	/* else, if input is not null, patch currdir with input dir 
	 * to get the block dir*/
	else if (*new){
		strncpy(BLOCKS_PATH,currdir,ILINE*4);
		strncat(BLOCKS_PATH,new,ILINE*4);
	}
	/* otherwise reset to default in case of empty argument */
	else {
		strcpy(BLOCKS_PATH,HID_HOME);
		strcat(BLOCKS_PATH,BLOCKS_NAME);
	}
	/* check if it ends with '/' */
	if (BLOCKS_PATH[strlen(BLOCKS_PATH)-1]!='/')
		strncat(BLOCKS_PATH,"/",ILINE*4);
	/* if the block directory does not exist, create it */
	dir = opendir(BLOCKS_PATH);
	if (dir) {
		/* directory exists: all good */
    		closedir(dir);
	}
	else if (ENOENT == errno) {
		/* directory does not exist: create it */
		if (stat(BLOCKS_PATH, &st) == -1) {
			mkdir(BLOCKS_PATH,0755);
		}
	}
	else {
    		/* invalid name */
		return FALSE;
	}
	return TRUE;
}


/* built-in LIB
 * set the current library file */
void _lib(void) {
	if (STATE==INTERPRET) {
		FILE *f;
		/* get new block dir */
		b=backbuffer;
		while (isBlank(line_buffer(to_in))) _inc_to_in_(); 
		while(line_buffer(to_in) && !isBlank(line_buffer(to_in))) { 
			*b++=line_buffer(to_in);
			_inc_to_in_();
		}
		*b='\0';
		b=backbuffer;
		/* if buffer was not empty, use argument as library file */
		if (*b) strncpy(LIBS_PATH,b,4*ILINE);
		/* otherwise set default library path */
		else {
			strcpy(LIBS_PATH,HID_HOME);
			strcat(LIBS_PATH,LIBS_NAME);
		}
		/* check for library file; if not existing create it empty */
		f=_file_open(LIBS_PATH);
		if (!f) {
			f=fopen(LIBS_PATH,"w");
			_doerr(!f,53);
		}
		fclose(f);
	}
	else codeup(_LIB);
}


/* built-in USE 
 * set the new blocks directory */
void _use(void) {
	if (STATE==INTERPRET) {
		/* get new block dir */
		b=backbuffer;
		while (isBlank(line_buffer(to_in))) _inc_to_in_(); 
		while(line_buffer(to_in) && !isBlank(line_buffer(to_in))) { 
			*b++=line_buffer(to_in);
			_inc_to_in_();
		}
		*b='\0';
		_doerr(!setblockdir(backbuffer),61);
	}
	else codeup(_USE);
}


/************************************
 * PORT LISTEN and SEND 
 ***********************************/

/* built-in PORT 
 * nport ON PORT return -1 in case port nport could not be opened
 * or in case no connection couldn't be established with port
 * return 0 in case of socket established 
 * nport OFF PORT 
 * return -1 regardless of the closing state */
void _port(void) {
	if (STATE==INTERPRET) {
		int socket_desc=0;
		unsigned int port_number=0;
		struct sockaddr_in server;
		int state=_pop(); // ON or OFF
		port_number=(unsigned int)_pop();
		/* PORT ON */
		if (state) {
			char ipaddr[ILINE];
			_doerr(portTOS>PORTNUM,46);
			/* Create socket */
			socket_desc=socket(AF_INET, SOCK_STREAM, 0);
			if (socket_desc==-1) {
				_push(-1);
				return;
			}
			/* Connect to remote server */
			snprintf(ipaddr,ILINE,"%d.%d.%d.%d",ip1,ip2,ip3,ip4);
			server.sin_addr.s_addr = inet_addr(ipaddr);
			server.sin_family = AF_INET;
			server.sin_port = htons(port_number);
			if (connect(socket_desc, (struct sockaddr *) & server, sizeof(server))<0) {
				_push(-2);
				return;
			}
			yport[portTOS].socket_desc=socket_desc;
			yport[portTOS].server=server;
			yport[portTOS].port_number=port_number;
			portTOS++;
			_push(0);
		}
		/* PORT OFF */
		else {
			int i=portTOS;
			while (i>=0 && yport[i].port_number!=port_number) i--;
			if (i>=0) {
				static const struct sockaddr_in zero;
				close(yport[i].socket_desc);
				yport[i].socket_desc=0;
				yport[portTOS].server=zero;
				yport[portTOS].port_number=0;
			}
			_push(-1);
		}
	}
	else codeup(_PORT);
}


/* LISTEN
 * nport LISTEN
 * listen port nport and push byte received;
 * in case no listening could be done, push 0 */
void _listen(void) {
	if (STATE==INTERPRET) {
		int i=portTOS;
		char server_reply[2];
		unsigned int port_number=_pop();
		while (i>=0 && yport[i].port_number!=port_number) i--;
		if (i<0) _push(0); // not found
		else {
   			if (recv(yport[i].socket_desc, server_reply, 1, 0)<0) {
				_push(0); // not received
			}
			else _push(server_reply[0]);
		}
	}
	else codeup(_LISTEN);
}


/* SEND 
 * n nport SEND 
 * send n to port nport; no return. */
void _send(void) {
	if (STATE==INTERPRET) {
		int i=portTOS;
		char message[2];
		unsigned int port_number=_pop();
		unsigned int val=_pop();
		while (i>=0 && yport[i].port_number!=port_number) i--;
		if (i>=0) {
			message[0]=val; 
			message[1]='\0';
			send(yport[i].socket_desc, message, 1, 0);
		}
	}
	else codeup(_SEND);
}


/************************************
 * CSP related words !CSP and ?CSP
 ***********************************/

/* built-in ?CSP */
void _question_csp(void) {
	if (STATE==INTERPRET) {
		unsigned short now=ds-DATASTACKSTART;
		_doerr(now!=CSP,58);
	}
	else codeup(_QUESTION_CSP);
}

/* built-in !CSP */
void _store_csp(void) {
	if (STATE==INTERPRET) {
		unsigned short sp=ds-DATASTACKSTART;
		CSP_(sp);
	}
	else codeup(_STORE_CSP);
}


/************************************************************
 * ON-DEBUG, OFF-DEBUG, TRACE and debugging utilities
 ************************************************************/

/* built-in TRACE (not Forth 79) */
void _trace(void) {
	if (STATE==INTERPRET) {
		unsigned short addr=0;
		/* get buffer and put in addr the flag filled/unfilled */
		addr=_fill_buff(backbuffer);
		/* TRACE was used without argument (unfilled): 
		 * reset traceVal */
	       	if (!addr) traceVal=0;
		/* find user TRACE word and set traceVal as the LFA 
		 * Note: if word is not in the dictionary 
		 * or it is set to a built-in, issue an error */
		else {
			addr=_find_word(backbuffer,TRUE);
			_doerr(!addr,49);
			_doerr(addr<LOWMEM,56);
			traceVal=addr;
		}
		isDebug=FALSE;
		stopKey=FALSE;
		wasTrace=TRUE;
	}
}


/* built-in DEBUG (not Forth 79) */
void _debug(void) {
	if (STATE==INTERPRET) {
		int re=_pop();
		/* perform ON DEBUG */
		if (re) {
			putchar('\n');
			isDebug=TRUE;
		}
		/* perform OFF DEBUG */
		else isDebug=FALSE;
		stopKey=FALSE;
		/* erase TRACE in any case */
		traceVal=0;
	}
}

	
/* waitKey()
 * execute debugger commands */
void waitKey(void) {
	char cc=1;
	struct termios thismodes, thissavemodes;
	static int isCalledE=0;
	if (stopKey) return;
	/* set tcgetattr values backing up current state */
	tcgetattr(0, &thismodes);
	thissavemodes = thismodes;
	thismodes.c_lflag &= ~ICANON;
	thismodes.c_lflag &= ~ECHO;
	tcsetattr(0, TCSANOW, &thismodes);
	cc = toupper(fgetc(stdin));
	tcsetattr(0, TCSANOW, &thissavemodes);
	if (isInterrupt) return;
	switch (cc) {
		case 'H':
			fputc('\n',stdout);
			fprintf(stdout,"%s\n","C-continue in debug mode   D-data stack             E-execute user code");
			fprintf(stdout,"%s\n","F-control flow stack       G-continue off debug     H-help");
			fprintf(stdout,"%s\n","R-return stack             X-stop and exit debug");
			fputc('\n',stdout);
			waitKey();
			break;
		case 'X':
			fprintf(stdout,"\n%s\n","(terminated)");
			isQuit=TRUE;
			stopKey=FALSE;
			break;
		case 'C':
			stopKey=TRUE;
			break;
		case 'G':
			isDebug=FALSE;
			traceVal=0;
			stopKey=TRUE;
			break;
		case 'R':
			fputc('\n',stdout);
			_uprint_r();
			fputc('\n',stdout);
			waitKey();
			break;
		case 'D':
			fputc('\n',stdout);
			_uprint_s();
			fputc('\n',stdout);
			waitKey();
			break;
		case 'F':
			fputc('\n',stdout);
			_print_cf();
			fputc('\n',stdout);
			waitKey();
			break;
		case 'E':
			{
				char btib[B_BUF+1], *bt=btib;
				char brl[B_BUF+1];
				int bin=to_in, bpc=_PC, bp=rl_point;
				int btv, bd, i=0;
				isCalledE++;
				/* backup line buffer and debug*/
				while (line_buffer(i) && i<B_BUF) 
					*bt++=line_buffer(i++);
				*bt='\0';
				btv=traceVal;
				bd=isDebug;
				strncpy(brl,rl_line_buffer,B_BUF);
				rl_delete_text(0,rl_end);
				/* ask for the line to interpret
				 * evaluate user input in no-debug mode */
				traceVal=0;
				isDebug=FALSE;
				to_in_(0);
				/* Note: readline passes automatically
				 * the line to process_input() */
				readline("\n\n(suspended)\n\n(D) ");
				/* restore current TIB, >IN, PC and debug */
				bt=btib;
				i=0;
				while (*bt) line_buffer(i++)=*bt++;
				line_buffer(i)='\0';
				to_in_(bin);
				isDebug=bd;
				traceVal=btv;
				PC_(bpc);
				/* restore readline stuff */
				rl_delete_text(0,rl_end);
				rl_point=0;
				rl_insert_text(brl);
				rl_point=bp;
				fputc('\n',stdout);
				_dot_s();
				fputc('\n',stdout);
				waitKey();
				/* re-print the debug line of the suspended
				 * user word that was delayed */
				if (!isQuit && isDebug && isCalledE==1) {
					fprintf(stdout,"\n\n(resumed)\n\nPC:%05d  %sNow suspended word is executed%s\n",_PC,REDCOLOR,RESETCOLOR);
				}
				isCalledE--;
			}
			break;
	}
}


/******************************************
 *  INCLUDE, REQUIRE, SEARCH and SEARCHLIB
 ******************************************/

/* built-in SEARCHLIB */
void _search_lib(void) {
	if (STATE==INTERPRET) {
		FILE *ff;
		char name[ILINE], *n=NULL;
		char listw[ILINE], *lw=NULL;
		int count=0, found=FALSE;

		/* open qlibrary.ft */
		ff=fopen(LIBS_PATH,"r");
		_doerr(!ff,53);

		/* get match string */
		b=backbuffer;
		while (isBlank(line_buffer(to_in))) _inc_to_in_(); 
		/* on empty string list all words */
		if (!line_buffer(to_in)) {
			fprintf(stdout,"\n");
			/* start reading the library */
			while (fgets(name,ILINE-1,ff)) {
				n=name;
				_remove_nasty(name);
				while (isblank(*n)) n++;
				if (*n==':') {
					n++;
					while (isblank(*n)) n++;
					lw=listw;
					while (!isblank(*n) && *n) *lw++=*n++;
					*lw='\0';
					/* control Carriage Return */
					if (count + (int)strlen(listw)>SCREEN) { 
						fputc('\n',stdout); 
						count=0; 
					}
					/* print word */
					count+=(int)strlen(listw);
					fprintf(stdout,"%s ",listw);
					count++;
				}
			}

		}

		/* proceed with the required search for a specific name
		 * contained in backbuffer */
		while(line_buffer(to_in) && !isBlank(line_buffer(to_in))) { 
			*b++=tolower(line_buffer(to_in));
			_inc_to_in_();
		}

		*b='\0';
		/* start reading the library */
		while (fgets(name,ILINE-1,ff)) {
			n=name;
			_remove_nasty(name);
			while (isblank(*n)) n++;
			if (*n==':') {
				n++;
				while (isblank(*n)) n++;
				lw=listw;
				while (!isblank(*n) && *n) *lw++=*n++;
				*lw='\0';
				if (strstr(listw,backbuffer)) {
					fprintf(stdout,"\n%s",listw);
					found=TRUE;
				}
			}
		}
		fclose(ff);
		if (!found) isError=TRUE;
	}
	else codeup(_SEARCH);

}


/* built-in INCLUDE */
void _include(void) {
	if (STATE==INTERPRET) {
		unsigned int i=0, count=0, btoin;
		char fname[ILINE+1], *fn=fname;
		char bbuff[B_BUF+1], *bb=bbuff;
	
		/* backup stream buffer */
		i=0;
		while (i<B_BUF) *bb++=line_buffer(i++);
	
		/* build file name and read file content */
		while (line_buffer(to_in)==BL) _inc_to_in_();
		if (line_buffer(to_in)=='\"') _inc_to_in_(); // starting quote
		while (line_buffer(to_in) && line_buffer(to_in)!=BL &&\
			       count<FILENAMELEN) {
			if (line_buffer(to_in)=='\"') { // ending quote
				_inc_to_in_(); count++;
				break;
			}
			else {
				*fn++=line_buffer(to_in);
				_inc_to_in_();
			}
		}
		*fn='\0'; btoin=to_in;
		/* return silently in case no file name was given */
		if (!fname[0]) return;
		_parse_file(fname);
	
		/* restore buffers */
		i=0; bb=bbuff;to_in_(btoin);
		while (i<B_BUF) line_buffer(i++)=*bb++;
	}
	else codeup(_INCLUDE);
}
		

/* load_required()
 * load required word from file opened at ff, with name wname 
 * Note: the loading process follows the qartus rules: if a word is not
 * found, compilation stops with an error 
 * System function for REQUIRE */
void _load_required(FILE *ff, char *wname) {
	unsigned int go_on=FALSE;
	char *p;
	char iline[ILINE+1];
	
	/* exit if wname is empty */
	if (!strlen(wname)) return;

	/* read all lines and search for : <any blanks> <name> 
	 * restarting from the beginning of library */
	fseek(ff,0,SEEK_SET);
	while (fgets(iline, ILINE, ff)) {
		_remove_nasty(iline);
		/* skip blanks */
		p=iline;
		while (isBlank(*p)) p++;
		/* skip slash comments at the start */
		if (*p=='\\' && isBlank(*(p+1))) continue;
		/* found the colon definition */
		if (*p==':') {
			p++;
			while (isBlank(*p)) p++;
			/* found word? */
			if (!strncasecmp(p,wname,strlen(wname))) {
				/* No. word is only part of a word */
				if (!isFinal(*(p+strlen(wname)))) continue;
				go_on=TRUE;
				/* Yes. begin a cycle to read all program */
				do {
					p=iline;
					to_in_(0);
					while (*p) {
						line_buffer(to_in)=*p++;
						_inc_to_in_();
					}
					line_buffer(to_in)='\0';
					/* set #TIB */
					SHARP_TIB1=to_in%BYTE;
					SHARP_TIB=to_in/BYTE;
					/* reset current >IN */
					to_in_(0);
					if (line_buffer(to_in)) 
					   _evaluate_input_buffer();
					if (fgets(iline, ILINE, ff)) {
						_remove_nasty(iline);
						p=iline;
						while (isBlank(*p)) p++;
						if (!*p || !strlen(iline))
						       continue;
						if (*p==':' &&isBlank(*(p+1))) 
							go_on=FALSE;
					}
					/* end of file */
					else go_on=FALSE;
				} while (go_on);
				break;
			}
			if (go_on) break;
		}
	}
	if (!isQuit && !isFile && !isBlock) {
		int i=0;
		fputc('\n',stdout);
		while (i<reqOffset) fputc(' ',stdout),i++;
		fprintf(stdout,"%s added ",wname);
	}	
	/* if state is still compile, there is an 
	 * unfinished colon definition */
	if (STATE==COMPILE) {
		_perror(48,to_in); // resets STATE
	}
}


/* _not_a_number()
 * return TRUE if number is not a number in Forth sense, that is
 * it's made of a starting '-' or '+', and is made of dots and digits
 * belonging to BASE.
 * Return FALSE if it is a number. 
 * System function for get_and_push() and REQUIRE */
int _not_a_number(char *num) {
	char *n=num;
	if (*n=='+' || *n=='-') n++;
	while (*n) {
		if (*n=='.') n++;
		else if (*n==',') n++;
		else if (*n<'0' || *n>digits[BASE-1]) return TRUE;
		else n++;
	}
	return FALSE;
}


/* _check_required()
 * check if the required word has some dependencies that can be solved
 * in the file ff itself, and load them in turn before loading the
 * required word passed as argument
 * if mode=1 perform REQUIRE, if mode=0 perform ENSURE
 * return zero in case of success  
 * System function for REQUIRE */
void _check_required(int mode, FILE *ff, char *wname, int* done) {
	char ns[WORDNAMELEN+1], *c, *p;
	char iline[ILINE+1];
	long pos,flag;
	unsigned int go_on=FALSE, isParen=FALSE;

	/* exit if wname is empty */
	if (!strlen(wname)) return;

	/* read all lines and search for : <any blanks> <name> 
	 * restarting from the beginning of library */
	pos=ftell(ff)+strlen(wname);
	fseek(ff,0,SEEK_SET);
	while (fgets(iline, ILINE, ff)) {
		_remove_nasty(iline);
		/* skip blanks */
		p=iline;
		while (isBlank(*p)) p++;
		/* skip slash comments at the start */
		if (*p=='\\' && isBlank(*(p+1))) continue;
		/* found a colon definition */
		if (*p==':') {
			p++;
			/* skip blanks */
			while (isBlank(*p)) p++;
			/* found word? */
			if (!strncasecmp(p,wname,strlen(wname))) {
				/* No. word is only part of a word */
				if (!isFinal(*(p+strlen(wname)))) continue;
				*done=TRUE;
				/* Yes. check for ENSURE  */
				if (!mode && _find_word(wname,TRUE)) return;
				go_on=TRUE;
				p+=strlen(wname);
				/* match is only part of another word */
				if (!(isBlank(*p) || !*p)) continue; 
				/* begin a cycle to read all program 
				 * and check every word */
				while (go_on) {
					while (*p) {
						c=ns;
						flag=TRUE;
						while (isBlank(*p)) p++;
						if (!*p && !isParen) break;
						else if (*p=='\\' && !isParen) 
							break;
						else if (*p=='(' &&\
							       isBlank(*(p+1))\
							       && !isParen)
							p++, isParen=TRUE;
						else if (*p==')' && isParen)
							p++, isParen=FALSE;
						if (isParen) {
							p++;
							continue;
						}
						while (*p && !isBlank(*p)) 
							*c++=*p++;
						*c++='\0';
						if (!strlen(ns)) continue;
						/* exclude existing words
						 * and pure numbers */
						if (_find_word(ns,TRUE)) {
							flag=FALSE;
						}
						else if (!_not_a_number(ns)) {
							flag=FALSE;
						}
						/* proceed with analysis*/
						if (flag) {
							int i=0;
							if(!isFile&&!isBlock){
							   fputc('\n',stdout);
							   while (i<reqOffset)
							     fputc(' ',stdout),
							     i++;
							   fprintf(stdout,"%s requires %s",wname,ns);
							}
							long ps=ftell(ff);
							reqOffset+=3;
							_check_required(mode,ff,ns, done);
							reqOffset-=3;
							fseek(ff,ps,SEEK_SET);
						}
						/* skip various */
						if (!strcmp(".\"",ns)) {
							while (*p!='\"') p++;
							p++; // skip closing
						}
						else if (!strcmp(ns,"ascii")) {
							while (isBlank(*p))p++;
							p++; // skip ASCII char
						}
						else if (!strcmp(ns,";")) {
							go_on=FALSE;
							break;
						}
					}
					fgets(iline, ILINE, ff);
					_remove_nasty(iline);
					p=iline;
					while (isBlank(*p)) p++;

				}
				break;
			}
		}
	}
	/* now we have solved all dependencies, so load word if found */
	if (*done) _load_required(ff,wname);
	fseek(ff,pos,SEEK_SET);
}


/* built-in REQUIRE 
 * if mode=1 perform REQUIRE
 * if mode=0 perform ENSURE */
void _require(int mode) {
	if (STATE==INTERPRET) {
		unsigned int i, btoin, len=0;
		int done=FALSE, bbase=BASE;
		char wname[B_BUF+1], *wn=wname;
		char bname[B_BUF+1], *bb=bname;
		FILE *ff;

		/* back up TIB */
		i=0;
		while (i<B_BUF) *bb++=line_buffer(i++);
	
		/* build word name and store in wname */
		while (line_buffer(to_in)==BL) _inc_to_in_();
		while (line_buffer(to_in) && !isBlank(line_buffer(to_in)) &&\
			       len<WORDNAMELEN) {
			*wn++=tolower(line_buffer(to_in));
			len++;
			_inc_to_in_();
		}
		*wn='\0';
		btoin=to_in;

		/* check if name is missing */
		_doerr(!strlen(wname),24); 

		/* open qlibrary.ft */
		ff=fopen(LIBS_PATH,"r");
		_doerr(!ff,53);

		/* check wname (word is loaded by _load_required() called
		 * by _check_required() */
		reqOffset=0;
		BASE=10;
		_check_required(mode, ff, wname, &done);

		/* close qlibrary.ft */
		if (ff) fclose(ff);
		
		/* restore TIB */
		i=0; bb=bname; to_in_(btoin);
		while (i<B_BUF) line_buffer(i++)=*bb++;
		BASE=bbase;

		/* check error condition */
		_doerr(!done,54);
	}
	else {
		if (mode) codeup(_REQUIRE);
		else codeup(_ENSURE);
	}	
}


/* built-in SEARCH */
void _search(void) {
	if (STATE==INTERPRET) {
		unsigned short block=1;
		char buf[B_BUF+1], ibuf[B_BUF+1];
		char bname[ILINE+1], nname[ILINE+1];
		FILE *f=NULL;
		/* get match string */
		b=backbuffer;
		while (isBlank(line_buffer(to_in))) _inc_to_in_(); 
		/* return on empty string */
		if (!line_buffer(to_in)) {
			return;
		}
		while(line_buffer(to_in) && !isBlank(line_buffer(to_in))) { 
			*b++=tolower(line_buffer(to_in));
			_inc_to_in_();
		}
		*b='\0';
		block=1;
		while (block<DOUBLE-1) {
			/* get file name and open file */
			strcpy(bname,BLOCKS_PATH);
			sprintf(nname, "%d", block);
			strcat(bname,nname);
			f=fopen(bname,"r");

			/* if file does exist, get content */ 
			if (f) {
				int d,c,o;
				char *p;
				register unsigned int i=0;
				while (i<B_BUF) {
					c=fgetc(f);
					if (c=='\n'||c==13||c==10) continue;
				       	ibuf[i]=c;
				       	buf[i]=tolower(c);
					i++;
				}
				buf[i]=ibuf[i]='\0';
				o=0;
				/* check occurrence; if found, print 
				 * line where the string begins  */
				while ((p=strstr(buf+o,backbuffer))) {
					d=(((int)(p-buf))/CPL);
					fputc('\n',stdout);
					fprintf(stdout,"%5d (%2d): ",block,d);
					for (i=0;i<CPL;i++) 
						fputc(ibuf[d*CPL+i],stdout);
					o=(int)(p-buf)+1;
				}
				fclose(f);
			}
			block++;
		}
		fputc('\n',stdout);
		OUT=0;
	}
	else codeup(_SEARCH);
}


/************************************
 *         SCREEN EDITOR
 ***********************************/
/* For what about the Screen Editor, I want to thank here Jeremy Ruten 
 * (in art yjerem), the author of the site "Build Your Own Text Editor", 
 * that can be found at 
 * http://viewsourcecode.org/snaptoken/kilo/index.html   
 * where all the necessary explanations about how to build your own textual 
 * console editor (vi-like) are contained. This site was the inspiration 
 * for my personal (and simpler, indeed) editor for qartus, called EDIT 
 * and whose source code begins here. Thanks, Jeremy. 
 * Note for lurkers: 
 * Jeremy work has the form jeremyWorkFunctions() 
 * My work has the form _antonio_work_functions() */


/* _search_occurrence()
 * search text 'match' in the buffer 'buf', starting from position 'pos'.
 * Return global position of match if found, or -1 
 * System function for the CTRL-F, CTRL-G and CTRL-R commands 
 * Note: I cannot use strcasestr() here because it's not standard 
 * and many systems don't have it */
int _search_occurrence(int buf, char* match, int pos) {
	int i,j;
	char base[B_BUF+1], *p;

	/* turn to lower the search string */
	p=match;
	while (*p) {
		*p=tolower(*p);
		p++;
	}
	/* build, turned to lower, the base string from buffer 'buf' */
	i=pos; j=0;
	while (i<1024) {
		base[j]=tolower(mem[_buffer_data[buf].start+i]); 
		i++, j++;
	}
	base[j]='\0';
	/* search */
	p=strstr(base,match);
	if (!p) return -1;
	else return pos+(int)(p-base);
}


/* _change_occurrence()
 * search text 'match' in the buffer 'buf', starting from position pos, and
 * substitute the search occurrence with the string 'change' 
 * Return global position of match if found, or -1 
 * System function for the CTRL-R and CTRL-V commands */
int _change_occurrence(int buf, char *match, char *change, int pos) {
	int i,j, l,p;

	/* first, search if match string is found: CTRL-R command 
	 * set return values to the match position */
	if (strlen(match)) { 
		i=_search_occurrence(buf,match,pos);
		/* not found */
		if (i<0) return i;
		/* found: get position of match */
		l=i/CPL;
		p=i%CPL;
	}
	/* otherwise it's the CTRL-V command: set return values to pos */
	else {
		l=pos/CPL;
		p=pos%CPL; 
	}

	/* easy case: substitute if lesser than or equal to original*/
	if (strlen(match)>=strlen(change)) {
		j=strlen(match)-strlen(change);
		i=p;
		/* if strings are different, shift back until end of line */
		if (j>0) {
			while (i<CPL-j) {
				mem[_buffer_data[buf].start+CPL*l+i]=
				mem[_buffer_data[buf].start+CPL*l+j+i];
				i++;
			}
			/* put blanks on the holes created */
			while (i<CPL) { 
				mem[_buffer_data[buf].start+CPL*l+i]=BL;
				i++;
			}
		}
		/* copy the change string in place */
		i=0;
		j=strlen(change);
		while (i<j) {
			mem[_buffer_data[buf].start+CPL*l+p+i]=*(change+i);
			i++;
		}
	}
	/* hard case: substitute if longer than the original, enlarging
	 * text space in case of insertion */
	else if (isIns) {
		int TextLine=l, TextPos=0;
		register unsigned int i=p;
		int diff=strlen(change)-strlen(match);
		int m=strlen(match);
		while (i<CPL) {
			if (mem[_buffer_data[buf].start+CPL*l+i]!=BL)
				TextPos=i;
			i++;
		}
		/* we're in a zone of blanks, since TextPos is still null */
		if (!TextPos) TextPos=p;
		/* don't insert if in the final zone and there's no room */
		if (TextLine==LPS && TextPos>=CPL-1-diff) return -1;
		/* if there's no space in current line, insert a new line */
		if (TextPos>=CPL-1-diff) {
			if (TextLine<LPS) shift(BFR,TextLine+1,ONWARD);
		}
		/* shift right all non blanks characters */
		for (i=CPL*TextLine+TextPos; i>=CPL*l+p+m; i--)	
			mem[_buffer_data[BFR].start+i+diff]=
			mem[_buffer_data[BFR].start+i];
		/* copy the change string in place */
		i=0; 
		j=strlen(change);
		while (i<j) {
			mem[_buffer_data[BFR].start+CPL*l+p+i]=*(change+i);
			i++;
		}
	}
	/* hard case: overwrite the text with insertion, not beyond screen */
	else {
		i=0; 
		j=strlen(change);
		while (i<j) {
			mem[_buffer_data[BFR].start+CPL*l+p+i]=*(change+i);
			i++;
			if (i>=1024) break;
		}
	}
	return CPL*l+p;
}


/* abAppend()
 * assemble the string to be built for the screen picturing 
 * Note: this function is almost entirely a Jeremy Ruten conception */
void abAppend(struct abuf *ab, const char *s, int len) {
	char *new = realloc(ab->b, ab->len + len);
	if (new == NULL) return;
	memcpy(&new[ab->len], s, len);
	ab->b = new;
	ab->len += len;
}


/* abFree()
 * free the string used for picturing 
 * Note: this function is almost entirely a Jeremy Ruten conception */
void abFree(struct abuf *ab) {
	free(ab->b);
}


/* editorMoveCursor() 
 * check cursor values and update
 * Note: this function maintains the Jeremy Ruten conception, but was
 * adapted to my needs */
void editorMoveCursor(int key) {
	switch (key) {
		case ARROW_LEFT:  
			if (epos>0) epos--; 
			else if (eline>0) {
				epos=CPL-1; 
				eline--;
			}
			if (isSel) SelEnd=CPL*eline+epos;
			break;
		case ARROW_RIGHT: 
			if (epos<CPL-1) epos++; 
			else if (eline<LPS) {
				epos=0; 
				eline++;
			}
			if (isSel) SelEnd=CPL*eline+epos;
			break;
		case ARROW_UP: 	  
			if (eline>0) eline--;
			else if (epos>0) {
				eline=LPS;
				epos--;
			}	
			if (isSel) SelEnd=CPL*eline+epos;
			break;
		case ARROW_DOWN:  
			if (eline<LPS) eline++;
			else if (epos<CPL-1) {
				eline=0;
				epos++;
			}	
			if (isSel) SelEnd=CPL*eline+epos;
			break;
	}
}


/* editorReadKey()
 * read keyboard keys hit by the user, intercepting also special cases
 * System routine for EDIT and KEY
 * Note: this function is almost entirely a Jeremy Ruten conception */
int editorReadKey(void) {
	int nread;
	char c;
	char seq[3];
	
	/* read one char */
	while ((nread = read(STDIN_FILENO, &c, 1)) != 1) {
		if (nread == -1 && errno != EAGAIN) return -1;
	}
	/* evaluate a valid char */
	if (c == '\x1b') {
		if (read(STDIN_FILENO, &seq[0], 1) != 1) return '\x1b';
	 	if (read(STDIN_FILENO, &seq[1], 1) != 1) return '\x1b';
	 	if (seq[0] == '[') {
			if (seq[1] >= '0' && seq[1] <= '9') {
				if (read(STDIN_FILENO,&seq[2],1)!=1) 
					return '\x1b';
				if (seq[2] == '~') {
					switch (seq[1]) {
						case '1': return HOME_KEY;
						case '3': return DEL_KEY;
						case '4': return END_KEY;
						case '5': return PAGE_UP;
						case '6': return PAGE_DOWN;
						case '7': return HOME_KEY;
						case '8': return END_KEY;
						case ' ': return CTRL_BAR;
					}
				}
			} 
			else {
				switch (seq[1]) {
				 	case 'A': return ARROW_UP;
				 	case 'B': return ARROW_DOWN;
				 	case 'C': return ARROW_RIGHT;
				 	case 'D': return ARROW_LEFT;
				 	case 'H': return HOME_KEY;
				 	case 'F': return END_KEY;
					case ' ': return CTRL_BAR;
				}
			}
		} else if (seq[0] == 'O') {
			switch (seq[1]) {
				case 'H': return HOME_KEY;
				case 'F': return END_KEY;
				case ' ': return CTRL_BAR;
			}
		}
		return '\x1b';
	} 
	else return c;
}


/* _assure_undo()
 * Store in undo_buffer the current screen data for a subsequent UNDO */
void _assure_undo(void) {
	register unsigned int i;
	for (i=0;i<B_BUF;i++)
		undo_buffer[i]=mem[_buffer_data[BFR].start+i];
	undo_buffer[i]='\0';
}


/* _check_selection() 
 * Check values of current selection, and if lower index is greater than
 * the higher (case of backward selection), exchange the two */
void _check_selection(void) {
	if (SelEnd<SelStart) {
		int back=SelStart;
		SelStart=SelEnd;
		SelEnd=back;
	}
}


/* _print_reverse()
 * print a character in the footer, reversing colors */
void _print_reverse(int c) {
	char output[ILINE+1];
	sprintf(output, "%s%c%s",INVCOLOR,c,RESETCOLOR);
	write(STDOUT_FILENO,output,strlen(output)); 
}


/* editorPrompt()
 * get string 's' in raw mode; 'lim' is the size of the string (when
 * reached, the string is returned); 'pos' is the lower input position; 
 * 'low' and 'high' delimit the input characters that the user can type, 
 * Note: this function maintains the Jeremy Ruten conception, but was
 * adapted to my needs */
void editorPrompt(char *s, int lim, int pos, int low, int high) {
	int c, i=0;
	while (i<lim) {
		c=editorReadKey();
		if (c==127) {  // backspace
			int u=' ';
			_print_reverse(u);
			if (epos>pos) epos--;
			_print_reverse(u);
			editorRefreshScreen(FALSE);
			if (i>0) i--, s--;
			continue;
		}
		if (c=='\r') break;
		if (c<low || c>high) continue;
		*s++=c; i++;
		_print_reverse(c);
		epos++;
	}
	*s='\0';
}


/* editorProcessKeypress()
 * interpret each key pressed. This is the heart of the editor.
 * Note: this function maintains the Jeremy Ruten conception, but was
 * heavily adapted to my needs */
void editorProcessKeypress(void) {
	int CH,G,i,j,l=0,c=editorReadKey(); //,bbfr;
	int rstart, rend, pstart, pend, delta; // for CTRL-X
	char stamp[EDITORLINE], *q;
	
	switch (c) {
		/* ESC: do nothing, except exit from Help Screen */
		case 27:
			if (!isHelpScreen) break;
			/* substitute HELP screen with current screen */
			for (i=0;i<B_BUF;i++)
				mem[_buffer_data[BFR].start+i]=undo_buffer[i];
			isHelpScreen=FALSE;
			break;
		/* CTRL-A select all */
		case CTRL_KEY('a'):
			if (isHelpScreen) break;
			isSel=TRUE;
			SelStart=0;
			SelEnd=1023;
			break;
		/* CTRL-B Page Down - edit block+1 */
		case CTRL_KEY('b'):
		case PAGE_DOWN: // UNDOCUMENTED
			G=SCR1+SCR*BYTE+1;
			_push(G);
			_block();
			_pop(); // consume address left by BLOCK 
			last_block_accessed=G;
			SCR1=G%BYTE;
			SCR=G/BYTE;
			_buffer_data[BFR].number=G;
			eline=0; epos=0;
			strncpy(footer_buffer,EFOOTER,68);
			break;
		/* CTRL-\ Page Up - edit block-1 */
		case 28:	// UNDOCUMENTED
		case PAGE_UP:	// UNDOCUMENTED
			G=SCR1+SCR*BYTE-1;
			_push(G);
			_block();
			_pop(); // consume address left by BLOCK 
			last_block_accessed=G;
			SCR1=G%BYTE;
			SCR=G/BYTE;
			_buffer_data[BFR].number=G;
			eline=0; epos=0;
			strncpy(footer_buffer,EFOOTER,68);
			break;
		/* CTRL-C copy selection */
		case CTRL_KEY('c'):
			if (isHelpScreen) break;
			if (SelStart-SelEnd==0) break;
			_check_selection();
			for (i=SelStart,j=0; i<=SelEnd; i++,j++)
				sel[j]=mem[_buffer_data[BFR].start+i];
			sel[j]='\0';
			isSel=FALSE;
			break;
		/* CTRL-D delete line and scroll up */
		case CTRL_KEY('d'):
			if (isHelpScreen) break;
			isTextInsert=FALSE;
			isBackSpace=FALSE;
			_assure_undo();
			shift(BFR,eline,BACKWARD);
			_buffer_data[BFR].update=TRUE; 
			break;
		/* CTRL-E edit another block */
		case CTRL_KEY('e'):
			eline=LPS+1; epos=7;
			strncpy(footer_buffer,BFOOTER,68);
			editorRefreshScreen(TRUE);
			q=matchString;
			editorPrompt(q,5,7,'0','9');
			G=strtol(matchString,NULL,10);
			if (matchString[0] && G>0 && G<DOUBLE) {
				_push(G);
				_block();
				_pop(); // consume address left by BLOCK 
				last_block_accessed=G;
				SCR1=G%BYTE;
				SCR=G/BYTE;
				_buffer_data[BFR].number=G;
				eline=0; epos=0;
			}
			strncpy(footer_buffer,EFOOTER,68);
			break;
		/* CTRL-F find a string in the text 
		 * Note: this section actually only asks for the match */
		case CTRL_KEY('f'):
			if (isHelpScreen) break;
			j=eline; l=epos;
			eline=LPS+1; epos=8;
			strncpy(footer_buffer,FFOOTER,68);
			editorRefreshScreen(TRUE);
			q=matchString;
			editorPrompt(q,48,8,' ',126);
			eline=j; epos=l;
			/* decrease & set flag for CTRL-F (first search) */
			epos--; j=-999; 
			/* no break here! process continues with CTRL-G */
		/* CTRL-G repeat find */
		case CTRL_KEY('g'):
			/* increase for next search 
			 * (avoid finding the same string again) */
			epos++; 
			if (isHelpScreen) break;
			if (strlen(matchString)) {
				i=_search_occurrence(BFR,matchString,CPL*eline+epos);
				/* text found: reposition cursor */
				if (i>=0) {
					eline=i/CPL;
					epos=i%CPL;
					strncpy(footer_buffer,EFOOTER,68);
					break;
				}
				else {
					/* text not found; reset cursor, 
					 * but if this is the first search, 
					 * don't reset, because epos was 
					 * already decreased within CTRL-F */
					if (j!=-999) epos--;
				}
			}
			strncpy(footer_buffer,TFOOTER,68);
			break;
		/* CTRL-H help screen - ESC to exit */
		case CTRL_KEY('h'):
			if (isHelpScreen) break;
			isHelpScreen=TRUE;
			isSel=FALSE;
			_assure_undo();
			eline=0; epos=0;
			/* substitute current screen with HELP screen*/
			for (i=0; i<B_BUF; i++)
				mem[_buffer_data[BFR].start+i]=helpScreen[i];
			break;
		/* CTRL-I insert new line and scroll down */
		case CTRL_KEY('i'): // = TAB
			if (isHelpScreen) break;
			isTextInsert=FALSE;
			isBackSpace=FALSE;
			_assure_undo();
			shift(BFR,eline,ONWARD);
			_buffer_data[BFR].update=TRUE; 
			break;
		/* CTRL-J toggle insert/overwrite */
		case CTRL_KEY('j'):
			if (isHelpScreen) break;
			isIns=isIns?FALSE:TRUE;
			break;
		/* CTRL-K empty current line */
		case CTRL_KEY('k'):	
			if (isHelpScreen) break;
			isTextInsert=FALSE;
			isBackSpace=FALSE;
			_assure_undo();
			shift(BFR,eline,BACKWARD);
			shift(BFR,eline,ONWARD);
			_buffer_data[BFR].update=TRUE; 
			break;
		/* CTRL-L pass to the line editor */
		case CTRL_KEY('l'):
			if (isHelpScreen) break;
			_editor(); // enable EDITOR vocabulary
			isStepToLinear=TRUE; // postpone list printing 
			isExitEdit=TRUE;
			break;
		/* CTRL-M = ENTER key: move to start of next line */
		case CTRL_KEY('m'): // = 'r'
			if (isHelpScreen) break;
			if (isIns && eline<LPS) {
				_assure_undo();
				isTextInsert=FALSE;
				isBackSpace=FALSE;
				shift(BFR,eline,ONWARD);
				/* copy the first part of line at the 
				 * beginning of next to the previous */
				i=0;
				while (i<epos) {
					mem[_buffer_data[BFR].start+CPL*eline+i]=
					mem[_buffer_data[BFR].start+CPL*(eline+1)+i];
					i++;
				}
				/* fill the rest with blanks */
				while (i<CPL)
					mem[_buffer_data[BFR].start+CPL*eline+i++]=BL;
				/* copy the second part to the beginning of 
				 * second line */
				i=epos; j=0;
				while (i<CPL) {
					mem[_buffer_data[BFR].start+CPL*(eline+1)+j]=
					mem[_buffer_data[BFR].start+CPL*(eline+1)+i];
					i++; j++;
				}
				/* fill the rest with blanks */
				while (j<CPL)
					mem[_buffer_data[BFR].start+CPL*(eline+1)+j++]=BL;
				_buffer_data[BFR].update=TRUE; 
			}
			/* in overwrite mode, no break is done */
			if (eline<LPS) eline++;
			epos=0;
			break;
		/* CTRL-N new buffer (erase content, fill with blanks) */
		case CTRL_KEY('n'):
			if (isHelpScreen) break;
			isTextInsert=FALSE;
			isBackSpace=FALSE;
			_assure_undo();
			for (i=0; i<B_BUF; i++)
				mem[_buffer_data[BFR].start+i]=BL;
			_buffer_data[BFR].update=FALSE; 
			BUFFERS--; if (BUFFERS<0) BUFFERS=0;
			break;
		/* CTRL-O reopen data file and undo all unsaved changes */
		case CTRL_KEY('o'):
			if (isHelpScreen) break;
			isTextInsert=FALSE;
			isBackSpace=FALSE;
			_assure_undo();
			load_buffer(SCR1+BYTE*SCR,BFR);
			_buffer_data[BFR].number=SCR1+BYTE*SCR;
			_buffer_data[BFR].update=TRUE;
			break;
		/* CTRL-P put header (stamp) */
		case CTRL_KEY('p'):
			if (isHelpScreen) break;
			isTextInsert=FALSE;
			isBackSpace=FALSE;
			_assure_undo();
			if (BFR<1 || BFR>4) break;
			sprintf(stamp,ISTAMP,INITIALS,getDay(),month[getMonth()],getYear());
			put_line(0,stamp);	
			_buffer_data[BFR].update=TRUE; 
			break;
		/* CTRL-Q quit: set exit flag */
		case CTRL_KEY('q'):
			isExitEdit=TRUE;
			break;
		/* CTRL-R replace a string with another */
		case CTRL_KEY('r'):
			if (isHelpScreen) break;
			isTextInsert=FALSE;
			isBackSpace=FALSE;
			_assure_undo();
			
			j=eline; l=epos;
			
			/* find match string if it's void */
			if (!matchString[0]) {
				eline=LPS+1; epos=8;
				strncpy(footer_buffer,FFOOTER,68);
				editorRefreshScreen(TRUE);
				q=matchString;
				editorPrompt(q,48,8,' ',126);
			}
			
			/* find substitute string */
			if (!replaceString[0]) {
				eline=LPS+1; epos=9;
				strncpy(footer_buffer,RFOOTER,68);
				editorRefreshScreen(TRUE);
				q=replaceString;
				editorPrompt(q,48,9,' ',126);
			}

			/* search & change occurrence */
			if (strlen(matchString) && strlen(replaceString)) {
				eline=j; epos=l;
				i=_change_occurrence(BFR,matchString,replaceString,CPL*j+l);
				/* text found: reposition cursor */
				if (i>=0) {
					eline=i/CPL;
					epos=i%CPL;
					strncpy(footer_buffer,EFOOTER,68);
					_buffer_data[BFR].update=TRUE;
				}
				/* text not found */
				else {
					replaceString[0]='\0';
					strncpy(footer_buffer,TFOOTER,68);
				}
			}
			else {
				replaceString[0]='\0';
				strncpy(footer_buffer,TFOOTER,68);
			}
			break;
		/* CTRL-S toggle selection mode */
		case CTRL_KEY('s'):
			if (isHelpScreen) break;
			isSel=isSel?FALSE:TRUE;
			if (isSel) {
				SelStart=CPL*eline+epos;
				SelEnd=CPL*eline+epos;
			}
			else {
				SelStart=SelEnd=0;
			}
			break;
		/* CTRL-T top: move cursor to top left or bottom right */
		case CTRL_KEY('t'):
			if (isHelpScreen) break;
			if (!eline && !epos) eline=LPS, epos=CPL-1;
			else eline=0, epos=0;
			break;
		/* CTRL-U undo */
		case CTRL_KEY('u'):
			if (isHelpScreen) break;
			if (!undo_buffer[0]) break;
			isTextInsert=FALSE;
			isBackSpace=FALSE;
			for (i=0; i<B_BUF; i++)
				mem[_buffer_data[BFR].start+i]=undo_buffer[i];
			_buffer_data[BFR].update=TRUE;
			break;
		/* CTRL-V paste selection */
		case CTRL_KEY('v'):
			if (isHelpScreen) break;
			isTextInsert=FALSE;
			isBackSpace=FALSE;
			_assure_undo();
			l=strlen(sel);
			/* no selection done */
			if (!l) break;
			_change_occurrence(BFR,"",sel,CPL*eline+epos);
			_buffer_data[BFR].update=TRUE; 
			break;
		/* CTRL-W write current buffer to storage mass */
		case CTRL_KEY('w'): {
			char output[70];
			unsigned short scr=SCR1+SCR*BYTE;
			if (isHelpScreen) break;
			if (_buffer_data[BFR].update) {
				save_buffer(scr,BFR);
				sprintf(output,SFOOTER,scr);
				strncpy(footer_buffer,output,67);
			}
			_buffer_data[BFR].update=FALSE;
			break;
		}
		/* CTRL-X extract selection (cut) */
		case CTRL_KEY('x'):
			if (isHelpScreen) break;
			isTextInsert=FALSE;
			isBackSpace=FALSE;
			if (SelStart-SelEnd==0) break;
			_check_selection();
			_assure_undo();
			/* copy extraction area to sel[] */
			for (i=SelStart,j=0; i<=SelEnd; i++,j++) 
				sel[j]=mem[_buffer_data[BFR].start+i];
			sel[j]='\0';
			/* move remaining line of text of last selection line 
			 * copying it to first line after last good char */
			rstart=SelStart/CPL;	pstart=SelStart%CPL;
			rend=SelEnd/CPL;	pend=SelEnd%CPL;
			delta=rend-rstart;
			for (i=pend+1,j=pstart; j<CPL && i<CPL; i++,j++) 
				mem[_buffer_data[BFR].start+rstart*CPL+j]= 
				mem[_buffer_data[BFR].start+rend*CPL+i];
			/* fill the rest of text in last selection line 
			 * with blanks */
			for (i=j;i<CPL;i++) 
				mem[_buffer_data[BFR].start+rstart*CPL+i]=BL;
			/* move up remaining lines */
			if (delta>0) { 
				for (i=rstart+1; i<=LPS-delta; i++) 
					for (j=0; j<CPL; j++)
						mem[_buffer_data[BFR].start+CPL*i+j]=
						mem[_buffer_data[BFR].start+CPL*(i+delta)+j];
				/* fill bottom lines if there's space */
				for (i=0; i<delta; i++)
				       for (j=0; j<CPL; j++)	
						mem[_buffer_data[BFR].start+CPL*(LPS-i)+j]=BL;

			}
			isSel=FALSE;
			SelStart=SelEnd=0;
			epos=pstart;
			eline=rstart;
			_buffer_data[BFR].update=TRUE; 
			break;
		/* CTRL-Y ????????? */
		case CTRL_KEY('y'):
			strncpy(footer_buffer,EASTEREGG,68);
			editorRefreshScreen(TRUE);
			break;
		/* CTRL-Z zero-file: set current buffer not updated and 
		 * remove file from directory */
		case CTRL_KEY('z'): {
			char output[70];
			unsigned short scr=SCR1+SCR*BYTE;
			if (isHelpScreen) break;
			_buffer_data[BFR].update=FALSE;
			eline=LPS+1; epos=50;
			sprintf(output,ZFOOTERQ,scr);
			strncpy(footer_buffer,output,67);
			editorRefreshScreen(TRUE);
			q=matchString;
			editorPrompt(q,1,50,'A','z');
			if (*q=='Y' || *q=='y') {
				remove_block(_buffer_data[BFR].number);
				strncpy(footer_buffer,ZFOOTER,68);
			}
			else strncpy(footer_buffer,EFOOTER,68);
			editorRefreshScreen(TRUE);
			break;
		}
		/* ignore unused (future scopes?) */
		case CTRL_KEY(' '):
		case CTRL_BAR:
		case 29:
		case 30:
		case 31:
			break;
		/* HOME: move to start of line */
		case HOME_KEY:
			if (isHelpScreen) break;
			epos=0;
			if (isSel) SelEnd=CPL*eline+epos;
			break;
		/* END: move to end of line */
		case END_KEY:
			if (isHelpScreen) break;
			epos=CPL-1;
			if (isSel) SelEnd=CPL*eline+epos;
			break;
		/* BACKSPACE: delete previous character and scroll left */
		case BACKSPACE: 
			if (isHelpScreen) break;
			isTextInsert=FALSE;
			/* stop deleting on pos 0,0 */
			if (!eline && !epos) break;
			/* calculate real end line (only the first time)*/
			if (!isBackSpace) {
				_assure_undo();
				TextLine=eline;	TextPos=epos;
				OTextLine=eline; OTextPos=epos;
				i=epos;
				while (i<CPL) {
					if (mem[_buffer_data[BFR].start+CPL*eline+i]!=BL)
						TextPos=i;
					i++;
				}
			}
			isBackSpace=TRUE;
			/* note: the cursor update must be done first */
			editorMoveCursor(ARROW_LEFT);
			/* shift left characters until last real, 
			 * and fill bottom with blank */
			for (i=CPL*eline+epos; i<=CPL*TextLine+TextPos;i++)	
				mem[_buffer_data[BFR].start+i]=
				mem[_buffer_data[BFR].start+i+1];
			mem[_buffer_data[BFR].start+CPL*TextLine+CPL-1]=BL;
			/* update last character's position after shifting */
			if (TextPos==0) {
				shift(BFR,TextLine,BACKWARD);
				if (TextLine>0) TextLine--;
				TextPos=CPL-1;
			}
			else TextPos--;
			_buffer_data[BFR].update=TRUE;
			break;
		/* DELETE: delete current character and scroll left
		 * all characters to the right (not the current) */
		case DEL_KEY:
			if (isHelpScreen) break;
			isTextInsert=FALSE;
			isBackSpace=FALSE;
			_assure_undo();
			/* shift left all characters to the right 
			 * in the same line */
			for (i=epos;i<CPL;i++)
				mem[_buffer_data[BFR].start+CPL*eline+i]=
				mem[_buffer_data[BFR].start+CPL*eline+i+1];
			/* erase last character */
			mem[_buffer_data[BFR].start+CPL*eline+CPL-1]=BL;
			_buffer_data[BFR].update=TRUE;
			break;
		/* arrows */
		case ARROW_UP:
		case ARROW_DOWN:
		case ARROW_LEFT:
		case ARROW_RIGHT:
			if (isHelpScreen) break;
			editorMoveCursor(c);
			break;
		/* character */
		default:
			if (isHelpScreen) break;
			CH=FALSE; // mark if last position is reached
			isBackSpace=FALSE;
			strncpy(footer_buffer,EFOOTER,68);
			_assure_undo();
			/* overwrite mode: simply substitute the character
			 * under the cursor and move right */
			if (!isIns) {
				mem[_buffer_data[BFR].start+CPL*eline+epos]=c;
				editorMoveCursor(ARROW_RIGHT);
				_buffer_data[BFR].update=TRUE;
				break;
			}
			if (TextLine==LPS && TextPos==CPL-1) CH=TRUE;
			/* insert mode: shift text to the right */
			/* if moving outside insertion area, reset parameters */
			if (CH || eline<OTextLine || eline>TextLine) 
				isTextInsert=FALSE;
			/* calculate real end line (only the first time)*/
			if (!isTextInsert) {
				_assure_undo();
				TextLine=eline;	TextPos=epos;
				OTextLine=eline;OTextPos=epos;
				i=epos;
				while (i<CPL) {
					if (mem[_buffer_data[BFR].start+CPL*eline+i]!=BL)
						TextPos=i;
					i++;
				}
			}
			isTextInsert=TRUE;
			/* stop inserting if there's no room */
			if (TextLine==LPS && TextPos==CPL-1) {
				mem[_buffer_data[BFR].start+CPL*LPS+CPL-1]=c;
				break;
			}
			/* update last character's position before shifting */
			else if (TextPos==CPL-1) {
				if (TextLine<LPS) TextLine++;
				TextPos=0;
				shift(BFR,TextLine,ONWARD);
			}
			else TextPos++;
			/* shift right characters included last real */
			for (i=CPL*TextLine+TextPos-1; i>=CPL*eline+epos; i--)	
				mem[_buffer_data[BFR].start+i+1]=
				mem[_buffer_data[BFR].start+i];
			mem[_buffer_data[BFR].start+CPL*eline+epos]=c;
			_buffer_data[BFR].update=TRUE;
			if (CH) break;
			/* Note: the cursor update must be done last */
			editorMoveCursor(ARROW_RIGHT);
			break;
	}
}


/* editorRefreshScreen()
 * repaint the screen 
 * Note: this function maintains the Jeremy Ruten conception, but was
 * adapted to my needs */
void editorRefreshScreen(int do_footer) {
	struct abuf ab=ABUF_INIT;
	char header[ILINE+1];
	int blockn=_buffer_data[BFR].number;
	int updaten=_buffer_data[BFR].update;
	int p,r;
	int st=SelStart, en=SelEnd;
	char k, str[2];

	/* terminal control codes */
	abAppend(&ab, "\x1b[?25l", 6); 	// suspend Cursor
	abAppend(&ab, "\x1b[H", 3);	// move cursor 0,0

	/* print header */
  	abAppend(&ab, INVCOLOR, strlen(INVCOLOR));
	sprintf(header,EHEADER,blockn,BFR,isIns?"INS":"OWR",eline,epos,updaten?"UPD":"   ",getDay(),month[getMonth()],getYear());
	abAppend(&ab, header, 67);
  	abAppend(&ab, RESETCOLOR, strlen(RESETCOLOR));

	/* print buffer() */
	if (st>en) p=st,st=en,en=p;
	str[1]='\0';
	for (r=0;r<=LPS;r++) {
		for (p=0;p<CPL;p++) {
			k=mem[_buffer_data[BFR].start+CPL*r+p];
			str[0]=k;
			if (isSel) {
			       if (CPL*r+p>=st && CPL*r+p<=en) {
					abAppend(&ab,INVCOLOR,strlen(INVCOLOR));
				}
				else abAppend(&ab,RESETCOLOR,strlen(RESETCOLOR));
			}
			abAppend(&ab, str, 2);
		}
		abAppend(&ab, "|\r\n", 3);
	}
  	
	/* print footer */	
	if (do_footer) {
		abAppend(&ab, INVCOLOR, strlen(INVCOLOR));
		sprintf(header,"%s",footer_buffer);
		abAppend(&ab, header, 67);
  		abAppend(&ab, RESETCOLOR, strlen(RESETCOLOR));
	}

	/* hardwire cursor position */
	snprintf(header, sizeof(header), "\x1b[%d;%dH", eline+2, epos+1);
	abAppend(&ab, header, strlen(header));
	abAppend(&ab, "\x1b[?25h", 6);	// enable cursor
	write(STDOUT_FILENO, ab.b, ab.len);
	abFree(&ab);
}

/*     Screen Editor command      */

/* built-in EDIT */
void _edit(void) {
	if (STATE==INTERPRET) {
		unsigned short block;
		struct termios raw;
		struct termios orig_termios;
		unsigned int i; //, bbfr=BFR;
		block=_pop();
	
		/* check if not already in memory */
		i=1;
		while (i<4) {
			if (_buffer_data[i].number==block) {
				BFR=i;
			       	SCR1=block%BYTE;
			       	SCR=block/BYTE;
				last_block_accessed=block;
			}
			i++;
		}
	
		/* reset parameters */
		eline=0; epos=0;
		undo_buffer[0]='\0';
		
	      	/* clear screen */
		write(STDOUT_FILENO, "\x1b[2J", 4);
		
		_push(block); // for BLOCK
		_block(); // perform BLOCK - sets BFR
		_pop(); // consume address left by BLOCK 
		_doerr(!check_buffer(BFR),21);
		last_block_accessed=block;
		SCR1=block%BYTE;
		SCR=block/BYTE;
		_buffer_data[BFR].number=block;
		_buffer_data[BFR].update=FALSE;
	
		/* turn off canonical mode */
		tcgetattr(STDIN_FILENO, &orig_termios);
		raw = orig_termios;
		raw.c_iflag &= ~(BRKINT | ICRNL | INPCK | ISTRIP | IXON);
		raw.c_oflag &= ~(OPOST);
		raw.c_cflag |= (CS8);
		raw.c_lflag &= ~(ECHO | ICANON | IEXTEN | ISIG);
		raw.c_cc[VMIN] = 0;
		raw.c_cc[VTIME] = 1;
		tcsetattr(STDIN_FILENO, TCSAFLUSH, &raw);
	
		/* fill the screen */
		strncpy(footer_buffer,EFOOTER,68);
		editorRefreshScreen(TRUE);
		/* reset all flags */
		isHelpScreen=FALSE;
		isStepToLinear=FALSE;
		TextLine=0;
		TextPos=0;
		isTextInsert=FALSE;
		isSel=FALSE;
		isIns=FALSE;
		matchString[0]='\0';
		replaceString[0]='\0';
		/* begin cycle for reading characters from keyboard */
		while (!isExitEdit) {
			editorRefreshScreen(TRUE);
			editorProcessKeypress();
		}
		isSel=FALSE;
		editorRefreshScreen(TRUE);
		isExitEdit=FALSE;
		for (i=eline;i<=17;i++) write(STDOUT_FILENO, "\r\n",2);
	
		/* turn on again canonical mode */
		tcsetattr(STDIN_FILENO, TCSAFLUSH, &orig_termios);
		
		/* list current block when going to the Line Editor
		 * Note: this command must be put here because 
		 * the screen is restored in canonical mode only just 
		 * before this code block */
		if (isStepToLinear) {
			isStepToLinear=FALSE;
			_l();
		}	
		fputc('\n',stdout);
	}
	else codeup(_EDIT);
}


/************************************
 *          FILE WORDS
 ***********************************/

/* built-in OPEN-FILE  c-addr u fam -- fileid ior
 * Note: ax is R_O, or W_O, or R_O+W_O */
void _open_file(void) {
	if (STATE==INTERPRET) {
		unsigned short addr;
		unsigned short n, ax, filetos=1;
		unsigned int i;
		char fname[FILENAMELEN+1], *fn=fname;
		char mode[4];
	
		ax=_pop();
		n=_pop();
		addr=_pop();

		/* check data */
		_doerr(!n,24);
		if (ax<1) ax=1;
		if (ax>3) ax=3;
		if (ax==1) strcpy(mode,"r"); // reading
		if (ax==2) strcpy(mode,"a"); // writing is appending
		if (ax==3) strcpy(mode,"a+");// appending and reading 
	
		/* retrieve file name */
		i=0;
		while (i<n) *fn++=mem[addr+i++];
		*fn='\0';
		
		/* choose first channel */
		while (filetos<=FILELIMIT) {
			if (!yfile[filetos].state) break;
			filetos++;
		}
		_doerr(filetos>FILELIMIT,46);

		/* open file and connect first channel */
		yfile[filetos].fd=fopen(fname,mode);
		/* no connection */
		if (!yfile[filetos].fd) {
			_push(FALSE);
			_push(TRUE);
		}
		/* connection */
		else {
			yfile[filetos].state=TRUE;
			_push(filetos);
			_push(0);
		}
	}
	else codeup(_OPEN_FILE);
}


/* built-in CLOSE-FILE    fileid -- ior */
void _close_file(void) {
	if (STATE==INTERPRET) {
		unsigned short fid;
		unsigned short ior=FALSE;
		fid=_pop();
		if (yfile[fid].state) {
			if (fclose(yfile[fid].fd)) ior=TRUE;
		}
		yfile[fid].state=FALSE;
		_push(ior);
	}
	else codeup(_CLOSE_FILE);
}


/* built-in READ-LINE c-addr u1 fileid -- u2 flag ior */
void _read_line(void) {
	if (STATE==INTERPRET) {
		unsigned short addr;
		unsigned short n1, fid;
		int c,i;
		FILE *tf;
	
		fid=_pop();
		n1=_pop();
		addr=_pop();
	
		if (fid<=FILELIMIT) {	
			/* check */
			/* connection never established */
			if (!yfile[fid].state) {
				_push(0); _push(0); _push(1);
				return;
			}
			/* channel not open or irregular channel identifier */
			if (!yfile[fid].fd || fid<0 || fid>FILELIMIT) {
				_push(0); _push(0); _push(2);
				return;
			}
		}

		/* assign input */
		if (fid<=FILELIMIT) tf=yfile[fid].fd;
		else if (fid==STDIN) tf=stdin;
		else if (fid==STDERR) {
			_doerr(TRUE,59);
		}
		else if (fid==STDOUT) {
			_doerr(TRUE,59);
		}
		else if (fid>FILELIMIT) {
			_doerr(TRUE,59);
		}

		/* read characters from stream */
		i=0; c=0;
		while (i<n1) {
			c=fgetc(tf);
			if (c==EOF) {
				_push(0); _push(0); _push(0);
				return;
			}
			else if (c==10 || c==13 || c=='\n') {
				mem[addr+i++]='\n';
				break;
			}
			/* store */
			else mem[addr+i++]=c;		
		}
		/* erase the rest */
		_push(i);
		while (i<n1) mem[addr+i++]=0;
		_push(1); _push(0);
	}
	else codeup(_READ_LINE);
}


/* built-in WRITE-LINE c-addr u fileid -- ior */
void _write_line(void) {
	if (STATE==INTERPRET) {
		unsigned short addr;
		unsigned short n1, fid;
		int c,i;
		FILE *tf;
	
		fid=_pop();
		n1=_pop();
		addr=_pop();

		if (fid<=FILELIMIT) {
			/* check */
			/* connection never established */
			if (!yfile[fid].state) {
				_push(1);
				return;
			}
			/* channel not open or irregular channel identifier */
			if (!yfile[fid].fd || fid<0 || fid>FILELIMIT) {
				_push(1);
				return;
			}
		}

		/* assign output channel */
		if (fid<=FILELIMIT) tf=yfile[fid].fd;
		else if (fid==STDIN) {
			_doerr(TRUE,60);
		}
		else if (fid==STDERR) tf=stderr;
		else if (fid==STDOUT) tf=stdout;
		else if (fid>FILELIMIT) {
			_doerr(TRUE,60);
		}

		/* write characters to file */
		i=0;
		while (i<n1) {
			c=mem[addr+i];
			if (!c) break;
			fputc(c,tf);
			i++;
		}
		/* write terminator */
		fputc('\n',tf);
		_push(0);
	}
	else codeup(_WRITE_LINE);
}


/* SAVE-SYSTEM ( - flag )
 * This command (which is not an ANSI word) may be used to save on a file
 * the content of memory, which can be restored on a new execution by
 * LOAD-SYSTEM. The file name follows the command:
 * SAVE-SYSTEM backup.sav
 * no spaces are allowed
 * flag is true if file could not be found or is unwritable. */
void _save_system(void) {
	if (STATE==INTERPRET) {
		char fname[B_BUF+1], *fn=fname;
		FILE *ff=NULL;
		int i=0;
		/* build file name, open it and write memory to it;
		 * Note: memory is saved as this:
		 * first the value of HERE
		 * then the content of MEM */
		while (isBlank(line_buffer(to_in))) _inc_to_in_();
		_doerr(!line_buffer(to_in),36);
		while (line_buffer(to_in) && !isBlank(line_buffer(to_in))) { 
			*fn++=line_buffer(to_in);
			_inc_to_in_();
		}
		*fn='\0';
		if (!(ff=fopen(fname,"w"))) {
			_push(TRUE);
			return;
		}
		/* save HERE */
		fprintf(ff,"%d\n",HERE);
		fprintf(ff,"%d\n",PREVHERE);
		/* save vocabularies data */
		fprintf(ff,"%d\n",vocCount);
		i=0;
		while (i<=vocCount) {
			fprintf(ff,"%d\n",vocChain[i]);
			fprintf(ff,"%d\n",vocs[i].code);
			fprintf(ff,"%s\n",vocs[i].name);
			i++;
		}
		/* save dictionary */
		i=0;
		while (i<HERE) fprintf(ff,"%d\n",mem[i++]);
		if (ff) fclose(ff);
		_push(0);
	}
	else codeup(_SAVE_SYSTEM);
}


/* LOAD-SYSTEM ( - flag ) 
 * This command (which is not an ANSI word) may be used to recover the
 * memory state from a file on which SAVE-SYSTEM was used before.
 * The file name follows the command:
 * LOAD-SYSTEM backup.sav
 * flag is true if file could not be found or read. */
void _load_system(void) {
	if (STATE==INTERPRET) {
		char fname[B_BUF+1], *fn=fname;
		FILE *ff=NULL;
		int i=0, val=0;
		char vname[WORDNAMELEN+1];
		/* build file name, open it and retrieve memory */
		while (isBlank(line_buffer(to_in))) _inc_to_in_();
		_doerr(!line_buffer(to_in),36);
		while (line_buffer(to_in) && !isBlank(line_buffer(to_in))) {
			*fn++=line_buffer(to_in);
			_inc_to_in_();
		}
		*fn='\0';
		if (!(ff=fopen(fname,"r"))) {
			_push(TRUE);
			return;
		}
		/* read HERE */
		fscanf(ff,"%d\n",&val);
		HERE=val;
		fscanf(ff,"%d\n",&val);
		PREVHERE=val;
		LAST_(PREVHERE);
		/* read vocabularies data */
		fscanf(ff,"%d\n",&val);
		vocCount=val;
		i=0;
		while (i<=vocCount) {
			fscanf(ff,"%d\n",&val);
			vocChain[i]=val;
			fscanf(ff,"%d\n",&val);
			vocs[i].code=val;
			fscanf(ff,"%s\n",vname);
			strncpy(vocs[i].name,vname,WORDNAMELEN);
			i++;
		}
		/* read dictionary */
		i=0;
		while (i<HERE) {
			fscanf(ff,"%d\n",&val);
			mem[i++]=val;
		}
		if (ff) fclose(ff);
		_push(0);
	}
	else codeup(_LOAD_SYSTEM);
}


/************************************
 *          TIME WORDS
 ***********************************/

/*     System functions            */

/* getTime()
 * return current time structure */
struct tm *getTime(void) {
	time_t st_lt;
	st_lt=time(NULL);
	return localtime(&st_lt);
}


/* getDay()
 * return current day (1-31) */
int getDay(void) {
	struct tm* vt = getTime();
	return (int) vt->tm_mday;
}


/* getMonth()
 * return current month (1-12) */
int getMonth(void) {
	struct tm* vt = getTime();
	return (int) (vt->tm_mon+1);
}


/* getYear()
 * return current year */
int getYear(void) {
	struct tm* vt = getTime();
	return (int) (vt->tm_year+1900);
}


/* getMin()
 * return current min (0-59) */
int getMin(void) {
	struct tm* vt = getTime();
	return (int) vt->tm_min;
}


/* getSec()
 * return current second (0-59) */
int getSec(void) {
	struct tm* vt = getTime();
	return (int) vt->tm_sec;
}


/* getHour()
 * return current hour (0-23) */
int getHour(void) {
	struct tm* vt = getTime();
	return (int) vt->tm_hour;
}


/* built-in CLOCK - uses microseconds */
void _clock(void) {
	if (STATE==INTERPRET) {
		long long int now, val;
		long long int high, low;
		gettimeofday(&tv, NULL);
		if (!NUMCLOCK) {
	        	now=(long long int)(tv.tv_sec*1000.0+tv.tv_usec/1000.0);
			val=now-beginS;
		}
		else {
	        	now=(long long int)(tv.tv_sec*1000000.+tv.tv_usec);
			val=now-beginM;
		}
		ll2int(val,&high,&low);
		_push(low);
		_push(high);
	}
	else codeup(_CLOCK);
}


/* built-in DATE */
void _date(void) {
	if (STATE==INTERPRET) {
		struct tm* vt = getTime();
		_push((int)(vt->tm_year+1900)); // year
		_push((int)(vt->tm_mon+1));	// month
		_push((int) vt->tm_mday);	// day
	}
	else codeup(_DATE);
}


/* built-in TIME */
void _time(void) {
	if (STATE==INTERPRET) {
		struct tm* vt = getTime();
		_push((int) vt->tm_sec);	// seconds
		_push((int) vt->tm_min);	// minutes
		_push((int) vt->tm_hour);	// hour
	}
	else codeup(_TIME);
}


/************************************
 *       VERSIONING WORDS
 ***********************************/

/* built-in BANNER */
void _banner(void) {
	if (STATE==INTERPRET) {
		fprintf(stdout,"%s",LOGO);
		if (isFigCompatEnabled)
			fprintf(stdout,"%s"," (in fig-Forth mode)\n");
		else 
			fprintf(stdout,"%s","\n");
	}
	else codeup(_BANNER);
}


/* built-in LICENSE */
void _license(void) {
	if (STATE==INTERPRET) {
		_print_license();
	}
	else codeup(_LICENSE);
}


/* built-in VERSION */
void _version(void) {
	if (STATE==INTERPRET) {
		_push(VERSION_MIN);
		_push(VERSION_MAJ);
	}
	else codeup(_VERSION);
}


/************************************
 *           LINE EDITOR 
 ***********************************/

/*     System functions            */

/* check_buffer() 
 * system function for the BLOCK features */
int check_buffer(int buf) {
	if (buf>0 && buf<4) return TRUE;
	return FALSE;
}


/* search_buffer() 
 * system function for the BLOCK features
 * search text in the current buffer 
 * return TRUE if found, FALSE otherwise */
int search_buffer(char *arg) {
	int pos=CPL*ROW_GRID+POS_GRID;
	char *c=arg;
	while (_buffer_data[BFR].start+pos<_buffer_data[BFR].end) {
		while (*c!=mem[_buffer_data[BFR].start+pos] &&\
				pos<B_BUF) pos++;
		while (*c==mem[_buffer_data[BFR].start+pos] &&\
				pos<B_BUF) c++, pos++;
		if (_buffer_data[BFR].start+pos>=_buffer_data[BFR].end) break;
		if (!*c) {
			ROW_GRID=(int)(pos/CPL);
			POS_GRID=pos-CPL*ROW_GRID;
			RSHARP1=(ROW_GRID*CPL+POS_GRID)%BYTE;
			RSHARP=(ROW_GRID*CPL+POS_GRID)/BYTE;
			return TRUE;
		}
		else c=arg;
	}
	/* return false without changing cursor references */
	return FALSE;
}


/* update_buffer_order()
 * system function for the BLOCK features 
 * updated order of buffers */
void update_buffer_order(void) {
	int i=1;
	int lim=3;
	for (i=1;i<=lim;i++) {
		_buffer_data[i].order++;
		if (_buffer_data[i].order==lim+1) _buffer_data[i].order=1;
	}
}


/* choose_available_buffer()
 * system function for the BLOCK features */
int choose_available_buffer(unsigned short block) {
	int i=1;
	int lim=3;
	/* search if block is already in memory in a buffer */
	for (i=1; i<=lim; i++) {
		if (_buffer_data[i].number==block) return i;
	}
	/* search next available buffer */
	for (i=1; i<=lim; i++) {
		/* if buffer is not in use, set it usable */
		if (!_buffer_data[i].in_use) return i;
		if (_buffer_data[i].order==lim) return i;
	}
	return lim;
}


/* shift()
 * system function for the BLOCK features 
 * shift line row in buffer buf:
 * - if mode = ONWARD, insert a void line in position row, losing last 15th 
 * - if mode = BACKWARD, remove line row, shifting the rest and creating a
 *   new 15th empty line */
void shift(int buf, int row, int mode) {
	unsigned int i=0;
	int r=mode==ONWARD?LPS:row;
	if (mode==ONWARD) {
		/* move lines ahead (losing 15th) */
		while (r>row) {
			i=0;
			while (i<CPL && r>row) {
				mem[_buffer_data[BFR].start+CPL*r+i]=mem[_buffer_data[BFR].start+CPL*(r-1)+i];
				i++;
			}
			r--;
		}
		/* write new row filling it with blanks */
		i=0;
		while (i<CPL) mem[_buffer_data[BFR].start+CPL*(row)+i++]=' ';
	}
	else {
		/* move lines backward (losing row-th) */
		while (r<LPS) {
			i=0;
			while (i<CPL && r<LPS) {
				mem[_buffer_data[BFR].start+CPL*r+i]=mem[_buffer_data[BFR].start+CPL*(r+1)+i];
				i++;
			}
			r++;
		}
		/* write new 15th row filling it with blanks */
		i=0;
		while (i<CPL) mem[_buffer_data[BFR].start+CPL*LPS+i++]=' ';
	}
}


/* put_line()
 * system function for the BLOCK features */
void put_line(int row, char *line) {
	unsigned int i=0;
	/* fill with line */
	while (i<CPL && *line) mem[_buffer_data[BFR].start+CPL*(row)+i++]=*line++;
	/* fill the rest with blanks */
	while (i<CPL) mem[_buffer_data[BFR].start+CPL*(row)+i++]=' ';
}


/* save_buffer()
 * system function for the BLOCK features */
void save_buffer(unsigned short blo, int buf) {
	char bname[ILINE+1], nname[ILINE+1];
	FILE *f=NULL;
	unsigned int r=0, i=0;

	/* assure buffer number is good */
	if (buf<1 || buf >3) return;

	/* get file name */
	strcpy(bname,BLOCKS_PATH);
	sprintf(nname, "%d", blo);
	strcat(bname,nname);

	f=fopen(bname,"w+");
	if (!f) {
		_perror(47,to_in);
		return;
	}

	while (r<=LPS) {
		char c=0;
		i=0;
		while (i<CPL) {
			c=mem[_buffer_data[buf].start+CPL*r+i];
			if (!c) c=BL;
			if (c=='\n') {
				while (i<CPL) fputc(BL,f), i++;
			}
			else {
				fputc(c,f);
				i++;
			}
		}
		fputc('\n',f);
		r++;
	}
	fclose(f);
}


/* load_buffer()
 * system function for the BLOCK features */
void load_buffer(unsigned short blo, int buf) {
	char bname[ILINE+1], nname[ILINE+1];
	FILE *f=NULL;
	unsigned int r=0, i=0, c=0, isAnti=FALSE;

	/* assure buffer number is good */
	if (buf<1 || buf >3) return;

	/* get file name and open file */
	strcpy(bname,BLOCKS_PATH);
	sprintf(nname, "%d", blo);
	strcat(bname,nname);
	f=fopen(bname,"r");

	/* if file does not exist, create it empty 
	 * Note: this is not in tune with the Standard. But this
	 * eases the treating of editors, even if in this way
	 * BLOCK is quite like BUFFER */
	if (!f) {
		while (i<B_BUF) mem[_buffer_data[buf].start+i++] = BL;
		return;
	}

	while (r<=LPS) {
		c=fgetc(f);
		if (c==EOF) {
			/* fill the rest with blank lines */
			while (r<=LPS) {
				i=0;
				while (i<CPL) {
				   mem[_buffer_data[buf].start+CPL*r+i]=BL;
				   i++;
				}
				r++;
			}
			break;
		}
		else ungetc(c,f);
		i=0;
		while (i<CPL) {
			c=fgetc(f);
			if (c==EOF) break;
			/* anticipated end of line (external file) */
			if (!c || c=='\n') {
				isAnti=TRUE;
				while (i<CPL) {
				   mem[_buffer_data[buf].start+CPL*r+i]=BL;
				   i++;
				}
				break;
			}
			/* substitute tab with 8 blanks */
			else if (c=='\t') {
				int j=0;
				while (i+j<CPL && j<8) {
				   mem[_buffer_data[buf].start+CPL*r+i+j]=BL;
				   j++;
				}
				i+=8;
			}
			else {
				mem[_buffer_data[buf].start+CPL*r+i] = c;
				i++;
			}
		}
		if (!isAnti) c=fgetc(f); // discard Return Carriage (as 65th)
		/* empty the rest of line */
		while (c!='\n' && c!=13 && c!=10) {
			c=fgetc(f);
		}
		r++;
	}
	fclose(f);
	_buffer_data[buf].in_use=TRUE;
	_buffer_data[buf].update=FALSE;
}


/* remove_block()
 * Remove physically the block from file system
 * system function for the BLOCK features */
void remove_block(unsigned short block) {
	char bname[ILINE+1], nname[ILINE+1];
	/* get file name and open file */
	strcpy(bname,BLOCKS_PATH);
	sprintf(nname, "%d", block);
	strcat(bname,nname);
	remove(bname);
}


/* strsubst()
 * System function for the A word of the EDITOR
 * in the string 'string', look for the first occurrence of 'old',
 * and substitute old with new, returning a the address of the modified string
 * or NULL if something went wrong 
 * NOTE: string, old and new must be strings with the NULL terminator */
char *strsubst(char *string, char *old, char *new) {
	char *y;
	y=strstr(string,old);
	/* the two strings have the same length */
	if (strlen(old)==strlen(new)) {
		while (*old && *new) *y++=*new++, old++; 
	}
	/* the string old is longer than new */
	else if (strlen(old)>strlen(new)) {
		char *z=y+strlen(old);
		while (*new) *y++=*new++;
		while (*z) *y++=*z++;
		while (*y) *y++=' ';
	}
	/* the string old is shorter than new */
	else if (strlen(old)<strlen(new)) {
		int i=(y-string);
		char *z, store[EDITORLINE], *s=store;
		/* store rest after old */
		z=y+strlen(old);
		while (*z) *s++=*z++;
		*s='\0';
		while (*new && i<CPL) *y++=*new++, i++;
		s=store;
		while (*s && i<CPL) *y++=*s++;
	}
	return string;
}


/*     Line Editor commands      */

/* built-in P (editor command) 
 * put chars in the stream buffer into the given line */
void _p(void) {
	if (STATE==INTERPRET) {
		unsigned int row;
		int count=0;
		char temp[EDITORLINE];
		dobfr;
		row=_pop();
		_doerr(!check_buffer(BFR),21);
		_doerr((row<0 || row>LPS),22);
		/* parse buffer for line content */
		b=temp;
		if (isBlank(line_buffer(to_in))) _inc_to_in_();
		while (line_buffer(to_in) && count < CPL) {
			*b++=line_buffer(to_in);
			_inc_to_in_();
		       	count++;
		}
		*b='\0';
		/* discard other characters */
		while (line_buffer(to_in)) _inc_to_in_(); 
		put_line(row,temp);	
		_buffer_data[BFR].update=TRUE;
		ROW_GRID=row;
		POS_GRID=0;
		RSHARP1=(ROW_GRID*CPL+POS_GRID)%BYTE;
		RSHARP=(ROW_GRID*CPL+POS_GRID)/BYTE;
	}
	else codeup(_EP);
}


/* built-in STAMP (editor command) 
 * put a fixed STAMP scheme on row 0 */
void _stamp(void) {
	if (STATE==INTERPRET) {
		char stamp[EDITORLINE];
		if (BFR<1 || BFR>4) return;
		sprintf(stamp,ISTAMP,INITIALS,getDay(),month[getMonth()],getYear());
		put_line(0,stamp);	
		_buffer_data[BFR].update=TRUE;
		ROW_GRID=1;
		POS_GRID=0;
		RSHARP1=(ROW_GRID*CPL+POS_GRID)%BYTE;
		RSHARP=(ROW_GRID*CPL+POS_GRID)/BYTE;
	}
	else codeup(_STAMP);
}


/* built-in D (editor commands)
 * delete given line and also copy it to PAD */
void _d(void) {
	if (STATE==INTERPRET) {
		unsigned int row;
		int ppad=0, i=0;
		dobfr;
		row=_pop();
		_doerr(!check_buffer(BFR),21);
		_doerr((row<0 || row>LPS),22);
		_pad();
		ppad=_pop(); // provisional pad
		while (i<CPL) {
			mem[ppad+i]=mem[_buffer_data[BFR].start+CPL*row+i];
			i++;
		}
		/* erase line by shifting lines */
		shift(BFR,row,BACKWARD);
		_buffer_data[BFR].update=TRUE;
		ROW_GRID=row;
		POS_GRID=0;
		RSHARP1=(ROW_GRID*CPL+POS_GRID)%BYTE;
		RSHARP=(ROW_GRID*CPL+POS_GRID)/BYTE;
	}
	else codeup(_ED);
}


/* built-in E (editor commands)
 * erase given line by filling it with blanks */
void _e(void) {
	if (STATE==INTERPRET) {
		unsigned int row;
		dobfr;
		row=_pop();
		_doerr(!check_buffer(BFR),21);
		_doerr((row<0 || row>LPS),22);
		/* erase line  */
		put_line(row,VOIDLINE);	
		_buffer_data[BFR].update=TRUE;
		ROW_GRID=row;
		POS_GRID=0;
		RSHARP1=(ROW_GRID*CPL+POS_GRID)%BYTE;
		RSHARP=(ROW_GRID*CPL+POS_GRID)/BYTE;
	}
	else codeup(_EE);
}


/* built-in N (editor command)
 * find next occurrence of the text in PAD */
void _n(void) {
	if (STATE==INTERPRET) {
		int count=0, ppad;
		char temp[EDITORLINE];
		dobfr;
		_doerr(!check_buffer(BFR),21);
		_pad();
		ppad=_pop();
		/* parse PAD for line content */
		b=temp;
		while (mem[ppad] && count < CPL) 
			*b++=mem[ppad++], count++;
		*b='\0';
		/* search text in temp over the buffer starting 
                 * at current cursor pos */
		if (!search_buffer(temp)) 
			fprintf(stdout,"\ntext '%s' not found in current buffer screen\n",temp);
		else {
			/* print found line */
			int i=0;
			fprintf(stdout,"\n %2d ",ROW_GRID);
			while (i<CPL) 
				fputc(mem[_buffer_data[BFR].start+CPL*ROW_GRID+i++],stdout);
			fputc('\n',stdout);
		}
	}
	else codeup(_EN);
}


/* built-in B (editor command)
 * move cursor back for the number of characters in PAD */
void _b(void) {
	if (STATE==INTERPRET) {
		int ppad, count=0;
		dobfr;
		_doerr(!check_buffer(BFR),21);
		_pad();
		ppad=_pop();
		while (mem[ppad]>=BL && mem[ppad]<=ESCL && count<PADLEN) {
			count++;
			ppad++; 
		}
		if (count) {
			_push(-count);
			_m();
		}
	}
	else codeup(_EB);
}


/* built-in T (editor command)
 * type line (row from stack) */
void _t(void) {
	if (STATE==INTERPRET) {
		unsigned int row;
		int i=0;
		dobfr;
		row=_pop();
		_doerr(!check_buffer(BFR),21);
		_doerr((row<0 || row>LPS),22);
		fprintf(stdout,"\n %2d ",row);
		while (i<CPL) 
			fputc(mem[_buffer_data[BFR].start+CPL*row+i++],stdout);
		fputc('\n',stdout);
		i=-5;
		if (row==ROW_GRID) {
			while (i<POS_GRID-1) {
				fputc(' ',stdout);
				i++;
			}
			fputc('^',stdout);
			fputc('\n',stdout);
		}
		ROW_GRID=row;
		POS_GRID=0;
		RSHARP1=(ROW_GRID*CPL+POS_GRID)%BYTE;
		RSHARP=(ROW_GRID*CPL+POS_GRID)/BYTE;
	}
	else codeup(_ET);
}


/* built-in L (editor command) 
 * list current buffer */
void _l(void) {
	if (STATE==INTERPRET) {
		unsigned int r=0, i=0;
		dobfr;
		_doerr(!check_buffer(BFR),21);
		fputc('\n',stdout);
		while (r<=LPS) {
			fprintf(stdout," %2d ",r);
			i=0;
			while (i<CPL) {
				fputc(mem[_buffer_data[BFR].start+CPL*r+i],stdout);
				i++;
			}
			fputc('\n',stdout);
			r++;
		}
		fputc('\n',stdout);
	}
	else codeup(_EL);
}


/* built-in S (editor command) 
 * shift line (row from stack) freeing a line (losing 15th) */
void _s(void) {
	if (STATE==INTERPRET) {
		unsigned int row;
		dobfr;
		row=_pop();
		_doerr(!check_buffer(BFR),21);
		_doerr((row<0 || row>LPS),22);
		shift(BFR,row,ONWARD); // frees row-th line
		ROW_GRID=row;
		POS_GRID=0;
		RSHARP1=(ROW_GRID*CPL+POS_GRID)%BYTE;
		RSHARP=(ROW_GRID*CPL+POS_GRID)/BYTE;
	}
	else codeup(_ES);
}


/* built-in I (editor command) 
 * insert line with PAD content (_i is the LOOP variable I) */
void __i(void) {
	if (STATE==INTERPRET) {
		unsigned int i=0, row;
		int ppad=0;
		dobfr;
		row=_pop();
		_doerr(!check_buffer(BFR),21);
		_doerr((row<0 || row>LPS),22);
		_pad();
		ppad=_pop(); // provisional pad
		shift(BFR,row,ONWARD);
		while (i<CPL) {
			mem[_buffer_data[BFR].start+CPL*row+i]=mem[ppad+i];
			i++;
		}
		_buffer_data[BFR].update=TRUE;
		ROW_GRID=row;
		POS_GRID=0;
		RSHARP1=(ROW_GRID*CPL+POS_GRID)%BYTE;
		RSHARP=(ROW_GRID*CPL+POS_GRID)/BYTE;
	}
	else codeup(_EI);
}


/* built-in F (editor command) 
 * search the following string in the current screen buffer */
void _f(void) {
	if (STATE==INTERPRET) {
		int count=0, ppad;
		char temp[EDITORLINE];
		dobfr;
		_doerr(!check_buffer(BFR),21);
		_pad();
		ppad=_pop();
		/* parse buffer for line content and copy to PAD */
		b=temp;
		if (isBlank(line_buffer(to_in))) _inc_to_in_();
		while (line_buffer(to_in) && count < CPL) {
			*b=line_buffer(to_in);
			_inc_to_in_();
		       	count++;
			mem[ppad++]=*b;
			b++;
		}
		*b='\0'; mem[ppad]=0;
		/* discard other characters in the stream */
		while (line_buffer(to_in)) _inc_to_in_();
		/* search text in temp over the buffer starting 
		 * at current cursor pos */
		if (!search_buffer(temp)) 
			fprintf(stdout,"\ntext '%s' not found in current buffer screen\n",temp);
		else {
			/* print line with updated cursor position */
			_push(ROW_GRID);
			_t();
		}
	}
	else codeup(_EF);
}



/* built-in A (editor command) 
 * search the first string in the current screen buffer and substitute
 * with the second string in the line buffer; also copy the old match to PAD */
void _a(void) {
	if (STATE==INTERPRET) {
		int i, ppad, count=0;
		char base[EDITORLINE], tempold[EDITORLINE], tempnew[EDITORLINE];
		dobfr;
		_doerr(!check_buffer(BFR),21);
		_pad();
		ppad=_pop();
		/* parse line buffer for old string */
		b=tempold; count=0;
		while (isBlank(line_buffer(to_in))) _inc_to_in_();
		if (line_buffer(to_in)=='\"') _inc_to_in_();
		while (line_buffer(to_in) && line_buffer(to_in)!='\"' && count<CPL) {
			*b++=line_buffer(to_in);
			_inc_to_in_();
			count++;
		}
		if (line_buffer(to_in)=='\"') _inc_to_in_();
		*b='\0';
		/* parse line buffer for new string */	
		b=tempnew; count=0;
		while (isBlank(line_buffer(to_in))) _inc_to_in_();
		if (line_buffer(to_in)=='\"') _inc_to_in_();
		while (line_buffer(to_in) && line_buffer(to_in)!='\"' && count<CPL) {
			*b++=line_buffer(to_in);
			_inc_to_in_();
			count++;
		}
		if (line_buffer(to_in)=='\"') _inc_to_in_();
		*b='\0';
		/* copy old match to PAD for another search*/
		i=0; b=tempold;
		while (i<CPL) mem[ppad+i++]=*b++;
		/* discard other characters in the stream */
		while (line_buffer(to_in)) _inc_to_in_();
		/* search tempold over the buffer starting at 
		   current cursor pos */
		if (!search_buffer(tempold)) 
			fprintf(stdout,"\ntext '%s' not found in current buffer screen\n",tempold);
		/* substitute new string */
		else {
			/* copy current line */
			b=base; i=0;
			while (i<CPL) *b++=mem[_buffer_data[BFR].start+CPL*ROW_GRID+i++];
			*b='\0';
			/* call strsubst() */
			b=strsubst(base,tempold,tempnew);
			/* refill current line */
			i=0;
			while (i<CPL) mem[_buffer_data[BFR].start+CPL*ROW_GRID+i++]=*b++;
			/* print line with updated cursor position */
			_push(ROW_GRID);
			_t();
		}
		_buffer_data[BFR].update=TRUE;
	}
	else codeup(_EA);
}


/* built-in C (editor command) 
 * insert the following text at the current cursor position */
void _c(void) {
	if (STATE==INTERPRET) {
		int count=0,i=0,ppad;
		char temp[EDITORLINE];
		dobfr;
		_doerr(!check_buffer(BFR),21);
		_pad();
		ppad=_pop();
		if (isBlank(line_buffer(to_in))) _inc_to_in_();
		/* if cccc is added, use line input as source */
		if (line_buffer(to_in)) {
			/* parse buffer for line content and copy to PAD */
			b=temp;
			while (line_buffer(to_in) && count < CPL) {
				*b=line_buffer(to_in);
				_inc_to_in_();
			       	count++;
				mem[ppad++]=*b;
				b++;
			}
			*b='\0'; mem[ppad]=0;
			/* discard other characters in the stream */
			while (line_buffer(to_in)) _inc_to_in_();
		}
		/* if cccc is not added, use PAD as source */
		else {
			b=temp;
			while (mem[ppad] && count < CPL)
				*b++=mem[ppad++], count++;
			*b='\0';
		}
		/* insert occurrence */
		count=strlen(temp); b=temp;
		for (i=CPL-1;i>=POS_GRID;i--) 
			mem[_buffer_data[BFR].start+CPL*ROW_GRID+i]=mem[_buffer_data[BFR].start+CPL*ROW_GRID+i-count];
		i=POS_GRID;
		while (*b) mem[_buffer_data[BFR].start+CPL*ROW_GRID+i++]=*b++;
		POS_GRID+=strlen(temp);
		/* print line with updated cursor position */
		_push(ROW_GRID);
		_t();
	}
	else codeup(_EC);
}


/* built-in M (editor command) 
 * update cursor position by the steps in stack (forward if positive,
 * backward if negative, unmoved if zero) */
void _m(void) {
	if (STATE==INTERPRET) {
		int off;
		dobfr;
		off=_pop();
		_doerr(!check_buffer(BFR),21);
		/* operate offset */
		if (off>0) {
			POS_GRID+=off;
			if (POS_GRID>=CPL) POS_GRID=0, ROW_GRID++;
			if (ROW_GRID>LPS) POS_GRID=CPL-1, ROW_GRID=LPS;
		}
		else if (off<0) {
			POS_GRID+=off;
			if (POS_GRID<0) POS_GRID=CPL-1, ROW_GRID--;
			if (ROW_GRID<0) POS_GRID=0, ROW_GRID=0;
		}
		RSHARP1=(ROW_GRID*CPL+POS_GRID)%BYTE;
		RSHARP=(ROW_GRID*CPL+POS_GRID)/BYTE;
		/* print line with updated cursor position */
		_push(ROW_GRID);
		_t();
	}
	else codeup(_EM);
}


/* built-in X (editor command) 
* find and delete the first occurrence of the given text */
void _x(void) {
	if (STATE==INTERPRET) {
		int count=0, ppad;
		char temp[EDITORLINE];
		dobfr;
		_doerr(!check_buffer(BFR),21);
		_pad();
		ppad=_pop();
		/* parse buffer for line content and copy to PAD */
		b=temp;
		if (isBlank(line_buffer(to_in))) _inc_to_in_();
		while (line_buffer(to_in) && count < CPL) { 
			*b=line_buffer(to_in); 
			_inc_to_in_();
			count++;
			mem[ppad++]=*b;
			b++;
		}
		*b='\0'; mem[ppad]=0;
		/* discard other characters in the stream */
		while (line_buffer(to_in)) _inc_to_in_();
		/* search text in temp over the buffer starting at current cursor pos */
		if (!search_buffer(temp)) 
			fprintf(stdout,"\ntext '%s' not found in current buffer screen\n",temp);
		else {
			int i;
			count=strlen(temp);
			i=POS_GRID-count;
			/* remove occurrence */
			while (i<CPL-count) mem[_buffer_data[BFR].start+CPL*ROW_GRID+i]=mem[_buffer_data[BFR].start+CPL*ROW_GRID+i+count], i++;
			while (i<CPL) mem[_buffer_data[BFR].start+CPL*ROW_GRID+i++]=' ';
			POS_GRID-=strlen(temp);
			/* print line with updated cursor position */
			_push(ROW_GRID);
			_t();
		}
	}
	else codeup(_EX);
}


/* built-in H (editor command) 
 * hold (copy) the content of the line at PAD */
void _h(void) {
	if (STATE==INTERPRET) {
		unsigned int row;
		int ppad=0, i=0;
		dobfr;
		row=_pop();
		_doerr(!check_buffer(BFR),21);
		_doerr((row<0 || row>LPS),22);
		_pad();
		ppad=_pop(); // provisional pad
		while (i<CPL) {
			mem[ppad+i]=mem[_buffer_data[BFR].start+CPL*row+i];
			i++;
		}
	}
	else codeup(_EH);
}


/* built-in R (editor command) 
 * replace the line with PAD content */
void _r(void) {
	if (STATE==INTERPRET) {
		unsigned int row;
		int count=0,ppad;
		char temp[EDITORLINE];
		dobfr;
		row=_pop();
		_doerr(!check_buffer(BFR),21);
		_pad(); ppad=_pop();
		_doerr((row<0 || row>LPS),22);
		/* parse PAD for line content */
		b=temp;
		while (mem[ppad]>=BL && mem[ppad]<=ESCL && count<PADLEN) {
			*b++=mem[ppad++], count++;
		}
		*b='\0'; 
		if (!count) fprintf(stdout,"PAD is empty.\n");
		else {
			put_line(row,temp);	
			_buffer_data[BFR].update=TRUE;
			/* print line with updated cursor position */
			_push(row);
			_t();
			ROW_GRID=row;
			POS_GRID=0;
			RSHARP1=(ROW_GRID*CPL+POS_GRID)%BYTE;
			RSHARP=(ROW_GRID*CPL+POS_GRID)/BYTE;
		}
	}
	else codeup(_ER);
}


/* built-in TILL (editor command) */
void _till(void) {
	if (STATE==INTERPRET) {
		int count=0, ppad;
		char temp[EDITORLINE];
		int bpos=POS_GRID;
		dobfr;
		_doerr(!check_buffer(BFR),21);
		_pad();
		ppad=_pop();
		/* parse buffer for line content and copy to PAD */
		b=temp;
		if (isBlank(line_buffer(to_in))) _inc_to_in_();
		while (line_buffer(to_in) && count < CPL) {
			*b=line_buffer(to_in);
			_inc_to_in_();
		       	count++;
			mem[ppad++]=*b;
			b++;
		}
		*b='\0'; mem[ppad]=0;
		/* discard other characters in the stream */
		while (line_buffer(to_in)) _inc_to_in_();
		/* search text in temp over the buffer starting 
		 * at current cursor pos */
		if (!search_buffer(temp)) {
			fprintf(stdout,"\ntext '%s' not found in current buffer screen\n",temp);
		}
		else {
			int i=0; bpos++;
			/* remove all occurrences from POS_GRID 
			 * till the end of match */
			while (i<CPL-POS_GRID-1) mem[_buffer_data[BFR].start+CPL*ROW_GRID+bpos+i]=mem[_buffer_data[BFR].start+CPL*ROW_GRID+POS_GRID+i], i++;
			while (i<CPL) mem[_buffer_data[BFR].start+CPL*ROW_GRID+i++]=' ';
			/* print line with updated cursor position */
			_push(ROW_GRID);
			_t();
		}
		POS_GRID=bpos-1;
		RSHARP1=(ROW_GRID*CPL+POS_GRID)%BYTE;
		RSHARP=(ROW_GRID*CPL+POS_GRID)/BYTE;
	}
	else codeup(_TILL);
}


/* built-in TEXT (editor command)  
 * expect CPL characters and put them in PAD */
void _etext(void) {
	if (STATE==INTERPRET) {
		int i, len=CPL, ppad;
		unsigned char c;
		_pad();
		i=ppad=_pop();
		fputc('\n',stdout);
		/* get string */
		while (len>0) {
			c=getchar();
			if (c==10 || c==13 || c=='\n') {
				mem[i]='\0';
				fputc('\n',stdout); // echo the Carriage Return
				break;
			}
			/* treat backspace */
			else if (c==127 && i>ppad) {
				mem[--i]='\0';
				fputc('\b',stdout); // echo the backspace
				continue;
			}
			else if (c>31 && c<128) {
				fputc(c,stdout); // forced echo (readline does not)
				mem[i++]=c;
			}
			len--;
		}
		while (len>0) mem[i++]='\0',len--;
		while (len>CPL-PADLEN) mem[i++]='\0',len--;
	}
	else codeup(_ETEXT);
}


/* built-in COPY (editor command)
 * copy one screen onto another */
void _copy(void) {
	if (STATE==INTERPRET) {
		int r=0, i=0, m, n;
		m=_pop();
	        n=_pop();
		_doerr((!check_buffer(m) || !check_buffer(n)),21);
		while (r<=LPS) {
			i=0;
			while (i<CPL) {
				mem[_buffer_data[m].start+CPL*r+i]=mem[_buffer_data[n].start+CPL*r+i];
				i++;
			}
			r++;
		}
		_buffer_data[m].update=TRUE;
	}
	else codeup(_COPY);
}


/* built-in CLEAR (editor command)
 * clear given buffer filling it with blanks */
void _clear(void) {
	if (STATE==INTERPRET) {
		int i, buffer;
		int bbfr=BFR;
		buffer=_pop();
		_doerr(!check_buffer(buffer),21);
		BFR=buffer;
		for (i=0;i<=LPS;i++) put_line(i,VOIDLINE);
		BFR=bbfr;
		_buffer_data[BFR].update=TRUE;
		POS_GRID=0;
		ROW_GRID=0;
		RSHARP1=(ROW_GRID*CPL+POS_GRID)%BYTE;
		RSHARP=(ROW_GRID*CPL+POS_GRID)/BYTE;
	}
	else codeup(_CLEAR);
}


/* built-in TOP (editor command) */
void _top(void) {
	if (STATE==INTERPRET) {
		POS_GRID=ROW_GRID=0;
		RSHARP1=RSHARP=0;
	}
	else codeup(_TOP);
}


/************************************
 *  LIST - LOAD - BLOCK - BUFFER 
 *  and related primitives
 ***********************************/

/* built-in LIST */
void _list(void) {
	if (STATE==INTERPRET) {
		unsigned short block;
		int i; 
		block=_pop();

		fprintf(stdout,"\n\nSCR # %d",block);
		/* check if not already in memory */
		i=1;
		while (i<4) {
			if (_buffer_data[i].number==block) {
				BFR=i;
			       	SCR1=block%BYTE;
			       	SCR=block/BYTE;
				last_block_accessed=block;
				_l();
				return;
			}
			i++;
		}
		
		_push(block); // for BLOCK
		_block(); // perform BLOCK - sets BFR
		_pop(); // consume address left by BLOCK 
		if (!check_buffer(BFR)) {
			_perror(21,to_in);
			return;
		}
		last_block_accessed=block;
		SCR1=block%BYTE;
		SCR=block/BYTE;
		_l();
	}
	else codeup(_LIST);
}


/* built-in LOAD */
void _load(void) {
	if (STATE==INTERPRET) {
		unsigned short block;
		int cw=countWords;
		int i, bblk, btoin, bpc;
		char bib[B_BUF+2], *bb;
		int isFileBk=isFile;
		isLoad=TRUE;
		block=_pop(); 

		/* check number */
		if (block==BLK1+BLK*BYTE) {
			_perror(28,to_in);
			return;
		}

		isBlock=TRUE;
		isFile=FALSE;
		/* back up input buffer and >IN resetting some variables */
		countWords=0; // necessary for DOES> 
		i=0; bb=bib;
		while (i<B_BUF) *bb++=line_buffer(i++);
		bblk=BLK1+BLK*BYTE; btoin=to_in; bpc=_PC;
		
		if (!block || block!=(BLK1+BYTE*BLK)) { 
			_push(block); // necessary for BLOCK
			_block(); // perform BLOCK (which also sets BFR)
			_pop(); // consume address left by BLOCK 
			last_block_accessed=block;
			BLK1=block%BYTE;
			BLK=block/BYTE;
		}
		_evaluate_block_buffer(0);
		
		BLK1=bblk%BYTE;
		BLK=bblk/BYTE;
		/* restore up input buffer and >IN */
		i=0; bb=bib; 
		while (i<B_BUF) line_buffer(i++)=*bb++;
		to_in_(btoin);
		PC_(bpc);
		countWords=cw;
		isBlock=FALSE;
		isFile=isFileBk;
		isLoad=FALSE;
	}

	else codeup(_LOAD);
}


/* built-in --> */
void _minus_to(void) {
	if (STATE==INTERPRET) {
		isMinusTo=TRUE;
	}
	else codeup(_MINUS_TO);
}


/* built-in ;S */
void _semi_s(void) {
	if (STATE==INTERPRET) {
		if (isBlock) isSemiS=TRUE;
		if (isFile) isStopFile=TRUE;
	}
	else {
		codeup(_WORD_ENDER);
		mem[HERE++]=0;
	}

}


/* built-in (BLOCK_READ) */
void _block_read(void) {
	if (STATE==INTERPRET) {
		unsigned short buf, block;
		BFR=buf=_pop();
		block=_pop();
		load_buffer(block,buf);
		_buffer_data[buf].number=block;
		if (buf>=1 && buf<=3) _push(_buffer_data[buf].start);
		else {
			BFR=0;
			_push(FALSE);
		}
	}
	else codeup(_BLOCK_READ);
}


/* built-in (BLOCK_WRITE) */
void _block_write(void) {
	if (STATE==INTERPRET) {
		unsigned short buf, block;
		buf=_pop();
		block=_pop();
		/* set block number in buffer structure */
		_buffer_data[buf].number=block;
		save_buffer(block, buf);
		if (buf>=1 && buf<=3) _push(TRUE);
		else _push(FALSE);

	}
	else codeup(_BLOCK_WRITE);
}


/* _update_buffer_count()
 * System function in BLOCK */
void _update_buffer_count(void) {
	unsigned int i;
	BUFFERS=0;
	/* update #BUFFERS counter */
	for (i=1; i<=3; i++) {
		if (_buffer_data[i].in_use) BUFFERS++;
	}
}


/* built-in BLOCK */
void _block(void) {
	if (STATE==INTERPRET) {
		int i=1;
		unsigned short block;
		block=_pop();
		_doerr(!block,28);
		/* check if a screen is already engaged with block */
		while (i<=3) {
			if (_buffer_data[i].number==block) {
				BFR=i;
				break;
			}
			i++;
		}
		/* block was not already found in screens */
		if (i>3) {
			/* proceed with BLOCK */
			BFR=choose_available_buffer(block); // the least used
			update_buffer_order();
			if (_buffer_data[BFR].in_use&&_buffer_data[BFR].update) 
				save_buffer(_buffer_data[BFR].number,BFR);
			load_buffer(block,BFR);
			_buffer_data[BFR].number=block;
			_update_buffer_count();
		}
		_push(_buffer_data[BFR].start);
	}
	else {
		codeup(_BLOCK);
	}
}


/* built-in BUFFER */
void _buffer(void) {
	if (STATE==INTERPRET) {
		int i=1;
		unsigned short block;
		block=_pop();
		_doerr(!block,28);
		/* check if a screen is already engaged with block */
		while (i<=3) {
			if (_buffer_data[i].number==block) {
				BFR=i;
				break;
			}
			i++;
		}
		/* block was not already found in screens */
		if (i>3) {
			/* proceed with BUFFER */
			BFR=choose_available_buffer(block); // the least used
			update_buffer_order();
			if (_buffer_data[BFR].in_use&&_buffer_data[BFR].update) 
				save_buffer(_buffer_data[BFR].number,BFR);
			_buffer_data[BFR].number=block;
			_update_buffer_count();
		}
		_push(_buffer_data[BFR].start);
	}
	else {
		codeup(_BUFFER);
	}
}


/* built-in EMPTY-BUFFERS */
void _empty_buffers(void) {
	if (STATE==INTERPRET) {
		unsigned int i=0;
		for (i=1; i<=3; i++) {
			_buffer_data[i].in_use=FALSE;
			_buffer_data[i].number=0;
			_buffer_data[i].update=FALSE;
		}
		BUFFERS=0;
	}
	else codeup(_EMPTY_BUFFERS);
}

/* built-in SAVE-BUFFERS */
void _save_buffers(void) {
	if (STATE==INTERPRET) {
		unsigned int i=0;
		/* error is detected by save_buffer()*/
		for (i=1; i<=3; i++) {
			if (_buffer_data[i].update) {
				save_buffer(_buffer_data[i].number,i);
				_buffer_data[i].update=FALSE;
			}
		}
	}
	else codeup(_SAVE_BUFFERS);
}


/* built-in UPDATE/MARK */
void _update(void) {
	if (STATE==INTERPRET) {
		_doerr(!check_buffer(BFR),21);
		_buffer_data[BFR].update=TRUE;
	}
	else codeup(_UPDATE);
}


/**********************************************
 *     INFO and its subsidiary functions   
 *********************************************/

/* _list_peculiar()
 * used by WHERE to list all the words with a precise bit pattern*/
void _list_peculiar(int bit, int lower_addr) {
	int count=0;
	int h=PREVHERE; /* last vocabulary entry */
	while (h>lower_addr) {
		if (count + wordlen(h)>SCREEN) { 
			fputc('\n',stdout); 
			count=0; 
		}
		if ((cbits(h)&bit) || !bit) {
			count+=_print_word_name(h);
			fputc(' ',stdout);
			count++;
		}
		h=prev(h);
	}
	if (count) fputc('\n',stdout);
}


/* _count_peculiar() 
 * used by WHERE to count all the words with a precise bit pattern*/
int _count_peculiar(int bit, int lower_addr) {
	int count=0;
	int h=PREVHERE; /* last vocabulary entry */
	while (h>lower_addr) {
		if ((cbits(h)&bit) || !bit) {
			count++;
		}
		h=prev(h);
	}
	return count;
}


/* where()
 * system function for the INFO word */
void _where(void) {
	if (STATE==INTERPRET) {
		int wcount=1, h=PREVHERE;
		fprintf(stdout,"\n\nqartus status:\n");
		fprintf(stdout,"-------------\n");
		fprintf(stdout,"last compiled word:   ");
		if (compiling[0]) {
			fprintf(stdout,"%s",compiling);
			if (!last_word_with_success) 
				fprintf(stdout," (uncompleted compilation)");
			fputc('\n',stdout);
		}
		else fprintf(stdout,"none\n");
		
		fprintf(stdout,"last accessed block:  ");
		if (last_block_accessed) fprintf(stdout,"%d\n",last_block_accessed);
		else fprintf(stdout,"none\n");
		
		fprintf(stdout,"last accessed buffer: ");
		if (BFR) fprintf(stdout,"%d\n",BFR);
		else fprintf(stdout,"none\n");

		/* count words live */
		while (h>0) wcount++, h=prev(h);
		/* exclude 000 */
		fprintf(stdout,"words compiled so far: %d",wcount-1); 
		if (!compiling[0]) fprintf(stdout," (no user words yet)\n");
		else fputc('\n',stdout);
	}
	else codeup(_WHERE);
}


/* count_vocs()
 * count all vocabularies outside default */
int _count_vocs(void) {
	int count=0; // predefined
	int k,h=PREVHERE; /* last vocabulary entry */
	fputc('\n',stdout);
	while (h>LOWMEM) {
		k=codestart(h)+2; //skip name and reach VOC_MARK
		if (mem[k]==VOC_MARK) {
			count++;
		}
		h=prev(h);
	}
	return count;
}


/* list_vocs()
 * print all vocabularies outside default */
void _list_vocs(void) {
	int pcount=0; // predefined
	int k,h=PREVHERE; /* last vocabulary entry */
	while (h>LOWMEM) {
		if (pcount + wordlen(h)>SCREEN) {
			fputc('\n',stdout);
			pcount=0;
		}
		k=codestart(h)+2; //skip name and reach VOC_MARK
		if (mem[k]==VOC_MARK) {
			pcount+=_print_word_name(h);
			fputc(' ',stdout);
			pcount++;
		}
		h=prev(h);
	}
	if (pcount) fputc('\n',stdout);
}


/* built-in INFO (information center) 
 * depends on value on tos:
 * 0 print in chunks
 * 1 print continuously
 * 2 print run mode and memory settings
 * 3 print interpreter status
 * 4 print limits
 * 5 dictionary system status 
 * 6 dictionary user status
 * 7 system directories and files */
void _info(void) {
	if (STATE==INTERPRET) {
		int val;
		char g1[EDITORLINE],g2[EDITORLINE],g3[EDITORLINE];
		val=_pop(); // print continuously = TRUE
		if (val>=2) {
			switch(val) {
				case 2: goto info2; break;
				case 3: goto info3; break;
				case 4: goto info4; break;
				case 5: goto info5; break;
				case 6: goto info6; break;
				case 7: goto info7; break;
				default: return;
			}
		}
		else if (val<0) return;
		/* 2 INFO */
		/* run mode and memory setting */
info2:		fprintf(stdout,"\n\nqartus running mode:        %s\n\n",isFigCompatEnabled?"fig-forth compatibility":"Forth-79");
		fprintf(stdout,"qartus memory settings (in bytes):\n");
		fprintf(stdout,"Memory sectors              Width [bytes]     Start [addr]\n");
		fprintf(stdout,"----------------------------------------------------------\n");
		if (_buffer_data[1].in_use) sprintf(g1,", in use");
		else g1[0]='\0';
		if (_buffer_data[2].in_use) sprintf(g2,", in use");
		else g2[0]='\0';
		if (_buffer_data[3].in_use) sprintf(g3,", in use");
		else g3[0]='\0';
		fprintf(stdout,"Start of memory                   1               0 (reserved)\n");
		fprintf(stdout,"Program memory              %7d               %d (available %d)\n",HIGHMEM,1,HIGHMEM-HERE+1);
		fprintf(stdout,"Memory in use               %7d\n",HERE-1);
		fprintf(stdout,"Terminal input buffer       %7d           %d\n",TERMINALINPUTBUFFERLIMIT-TERMINALINPUTBUFFERSTART+1, TERMINALINPUTBUFFERSTART);
		fprintf(stdout,"WORD buffer                 %7d           %d\n",WORDBUFFERLIMIT-WORDBUFFERSTART+1, WORDBUFFERSTART);
		fprintf(stdout,"Scratch area                %7d           %d\n",SCRATCHAREALIMIT-SCRATCHAREASTART+1,SCRATCHAREASTART);
		fprintf(stdout,"Screen buffer 1             %7d%-10s %d\n",_buffer_data[1].end-_buffer_data[1].start+1,g1,_buffer_data[1].start);
		fprintf(stdout,"Screen buffer 2             %7d%-10s %d\n",_buffer_data[2].end-_buffer_data[2].start+1,g2,_buffer_data[2].start);
		fprintf(stdout,"Screen buffer 3             %7d%-10s %d\n",_buffer_data[3].end-_buffer_data[3].start+1,g3,_buffer_data[3].start);
		fprintf(stdout,"Data stack                  %7d           %d\n",DATASTACKLIMIT-DATASTACKSTART+1,DATASTACKSTART);
		fprintf(stdout,"Return stack                %7d           %d\n",RETURNSTACKLIMIT-RETURNSTACKSTART+1,RETURNSTACKSTART);
		fprintf(stdout,"----------------------------------------------------------");
		fprintf(stdout,"\nEnd of memory                                 %d\n",ADDRLIMIT-1);
		if (!val) {
			fprintf(stdout,"\n<Press ENTER>");
			fgets(g1,2,stdin);
			if (g1[0]!=10 && g1[0]!=13) return;
		}
		if (val>1) return;
		/* 3 INFO */
info3:		_where();
		if (val>1) return;

info4:		/* 4 INFO */
		fprintf(stdout,"\n\nqartus limits and features:\n");
		fprintf(stdout,"--------------------------\n");
		fprintf(stdout,"Maximum strings size: %d\n",ILINE-1);
		fprintf(stdout,"Size of PAD: %d (%d bytes after HERE)\n",padlen,PADWIDTH);
		fprintf(stdout,"Type of division: symmetric\n");
		fprintf(stdout,"Maximum dimension of a character: 8 bits\n");
		fprintf(stdout,"  Default ASCII range is 0..127 (7 bits)\n");
		fprintf(stdout,"  Range 128-255 is O.S. dependent\n");
		fprintf(stdout,"Maximum string length: 255 characters\n");
		fprintf(stdout,"Screen/Block default size: %d bytes\n",B_BUF);
		fprintf(stdout,"Maximum number of word lists usable in the search order: %d\n",VOCS-1);
		fprintf(stdout,"Largest usable unsigned double number: %s\n",UDOUBLE);
		fprintf(stdout,"Largest usable unsigned number: %d\n",DOUBLE-1);
		fprintf(stdout,"Largest positive usable signed double number: %s\n",LPDOUBLE);
		fprintf(stdout,"Largest negative usable signed double number: %s\n",LNDOUBLE);
		fprintf(stdout,"Largest positive usable signed number: %d\n",SINGLE-1);
		fprintf(stdout,"Largest negative usable signed number: %d\n",-SINGLE);
		if (!val) {
			fprintf(stdout,"\n<Press ENTER>");
			fgets(g1,2,stdin);
			if (g1[0]!=10 && g1[0]!=13) return;
		}
		if (val>1) return;

		/* 5 INFO */
		/* database setting */
info5:		fprintf(stdout,"\n\nsystem dictionary status:\n");
		fprintf(stdout,"------------------------\n");

		fprintf(stdout,"list of immediate words: %d\n",
				_count_peculiar(B_PREC,0));
		_list_peculiar(B_PREC,0);
		
		fprintf(stdout,"\nlist of compile-only words: %d\n",
				_count_peculiar(B_COMPILE,0));
		_list_peculiar(B_COMPILE,0);
		
		fprintf(stdout,"\nlist of defining words: %d\n",
				_count_peculiar(B_DEFINING,0));
		_list_peculiar(B_DEFINING,0);
		
		fprintf(stdout,"\nlist of executable-only words: %d\n",
				_count_peculiar(B_EXECUTION,0));
		_list_peculiar(B_EXECUTION,0);
		
		if (!val) {
			fprintf(stdout,"\n <Press ENTER>");
			fgets(g1,2,stdin);
			if (g1[0]!=10 && g1[0]!=13) return;
		}
		if (val>1) return;

		/* 6 INFO */
info6:		fprintf(stdout,"\n\nuser dictionary status:\n");
		fprintf(stdout,"----------------------\n");
		fprintf(stdout,"list of user words: %d",
				_count_peculiar(0,LOWMEM-1));
		_list_peculiar(0,LOWMEM-1);
		fprintf(stdout,"list of user vocabularies: %d\n", _count_vocs());
		_list_vocs();
		if (val>1) return;

		/* 7 DIRECTORIES AND SYSTEM FILES */
info7:		fprintf(stdout,"\n\nsystem directories and files:\n");
		fprintf(stdout,"----------------------------\n");
		fprintf(stdout, "system directory: %s\n",HID_HOME);
		fprintf(stdout, "blocks directory: %s\n",BLOCKS_PATH);
		fprintf(stdout, "current linked library: %s\n",LIBS_PATH);
		fprintf(stdout, "resource file: %s %s\n",RES_PATH,isRes?"(enabled)":"(disabled)");

		fputc('\n',stdout);
	}
	else codeup(_INFO);
}


/************************************
 *           Forth WORDS
 ***********************************/

/* built-in .S */
void _dot_s(void) {
	if (STATE==INTERPRET) {
		unsigned int i;
		/* Stack with values */
		fprintf(stdout,"<%d> ",(ds-DATASTACKSTART)/2);
		for (i=DATASTACKSTART; i<ds; i+=2) {
			_print(mem[i]+BYTE*mem[i+1],BASE,NULL);
			fprintf(stdout,"%s"," ");
		}
	}
	else codeup(_DOT_S);
}

/* _print_cf()
 * auxiliary function for printing the control flow stack in
 * signed form */
void _print_cf(void) {
	int i;
	fprintf(stdout,"<CF:%d> ",cfs-CFSTACKSTART);
	for (i=CFSTACKSTART; i<cfs; i++) {
		_print(cfmem[i],BASE,NULL);
		fputc(' ',stdout);
	}
}


/* _uprint_s()
 * auxiliary function for printing the stack in unsigned form */
void _uprint_s(void) {
	int i; 
	unsigned short val;
	fprintf(stdout,"<DU:%d> ",(ds-DATASTACKSTART)/2);
	for (i=DATASTACKSTART; i<ds; i+=2) {
		val=mem[i]+BYTE*mem[i+1];
		_print_unsigned(val,BASE,NULL);
		fputc(' ',stdout);
	}
}


/* _uprint_r()
 * auxiliary function for printing the return stack in unsigned form */
void _uprint_r(void) {
	int i;
	unsigned short val;
	fprintf(stdout,"<RU:%d> ",(rs-RETURNSTACKSTART)/2);
	for (i=RETURNSTACKSTART; i<rs; i+=2) {
		val=mem[i]+BYTE*mem[i+1];
		_print_unsigned(val,BASE,NULL);
		fputc(' ',stdout);
	}
	return;
}


/* built-in DECIMAL */
void _decimal(void) {
	if (STATE==INTERPRET) BASE=10;
	else codeup(_DECIMAL);
}


/* built-in HEX */
void _hex(void) {
	if (STATE==INTERPRET) BASE=16;
	else codeup(_HEX);
}


/* built-in ?TERMINAL */
void _question_terminal(void) {
	if (STATE==INTERPRET) {
		int n;
		struct termios initial_settings, new_settings;	
 		tcgetattr(0,&initial_settings);
 		new_settings = initial_settings;
  		new_settings.c_lflag &= ~ICANON;
  		new_settings.c_lflag &= ~ECHO;
  		new_settings.c_lflag &= ~ISIG;
  		new_settings.c_cc[VMIN] = 0;
  		new_settings.c_cc[VTIME] = 0;
  		tcsetattr(STDIN_FILENO, TCSANOW, &new_settings);
    		n = getchar();
 		if (n != EOF) _push(n);
		else _push(0);
  		tcsetattr(STDIN_FILENO, TCSANOW, &initial_settings);
	}
	else codeup(_QUESTION_TERMINAL);
}


/* built-in ASCII */
void _ascii(void) {
	unsigned int c=0;
	while (isBlank(line_buffer(to_in))) _inc_to_in_();
	_doerr(!line_buffer(to_in),36);
	c=line_buffer(to_in)&CHAR;
	_inc_to_in_();
	if (STATE==INTERPRET) {
		_push(c);
	}
	else if (STATE==COMPILE) {
		mem[HERE++]='I';
		mem[HERE++]=(int)(c%BYTE);
		mem[HERE++]=c/BYTE;
	}
}

/* built-in ( 
 * Note: this word is immediate, not intended to be compiled, so
 * it has an immediate effect even during compilation */
void _paren(void) {
	unsigned int ch;
	while ((ch=line_buffer(to_in))) {
		if (ch==')') {
			ch=line_buffer(to_in+1);
			if (!ch || isBlank(ch)) {
				isOpen=FALSE;
				_inc_to_in_(); // skip )
				break;
			}
		}
		_inc_to_in_();
	}
}


/* built-in ABORT */
void _abort(void) {
	if (STATE==INTERPRET) {
		/* Note: by setting isQuit execution mode is set and
		 * control is passed to the terminal when _exec() performs
		 * the next istruction. */
		isQuit=TRUE;
		_erase_stacks(ALL);
		/* the fig-Forth requires that ABORT prints a message 
		 * "appropriate to the installation" */
		if (isFigCompatEnabled) fprintf(stdout,"\n\nqartus %d.%d %s aborted.\nReturning control to the operator.\n", VERSION_MAJ,VERSION_MIN, "(fig mode)");
	}
	else codeup(_ABORT);
}


/* built-in (ABORT) (Fig-forth) */
void _paren_abort(void) {
	_abort();
}


/* built-in COLD */
void _cold(void) {
	unsigned int i=0;
	/* hard reset */
	for (i=0;i<=HERE;i++) mem[i]=0;
	_erase_stacks(ALL);
	HERE=1;
	PREVHERE=0;
	LAST_(PREVHERE);
	BASE=10;
	STATE=INTERPRET;
	for (i=2;i<VOCS;i++) vocChain[i]=0;
	vocCount=1;
	CURRENT_(0);
	CONTEXT_(0);
	NVOCS=2;
	compiling[0]='\0';
	_add();
	print_long_banner();
	/* important: light on all semaphores! */
	isInterrupt=isQuit=TRUE;
}


/* built-in KEY */
void _key(void) {
	if (STATE==INTERPRET) {
		unsigned short c=0;
		/* before blocking, let's flush 
		 * the possible output on stdout */
		fflush(stdout);
		while (!c || c>ESCL) c=editorReadKey();
		_push(c);
	}
	else codeup(_KEY);
}


/* built-in INTERPRET */
void _interpret(void) {
	if (STATE==INTERPRET) {
		unsigned short block=BLK1+BYTE*BLK;
		/* interpret from current >IN position from current BLK 
		 * Note: get sure that BFR is instantiated */
		if (block>0) {
			if (BFR<1 || BFR>3) return;
			_evaluate_block_buffer(to_in);
		}
		/* interpret current input buffer (resetting >IN)
		 * Note: only one occurrence of INTERPRET is executed,
		 * to avoid circularity */
		else {
			if (isInterpret) return;
			isInterpret=TRUE;
			to_in_(0);
			_evaluate_input_buffer();
			isInterpret=FALSE;
		}
	}
	else codeup(_INTERPRET);
}


/* built-in QUERY */
void _query(void) {
	if (STATE==INTERPRET) {
		int i=0;
		char c;
		/* get string */
		while (i<QUERYLEN) {
			c=getchar();
			if (c==10 || c==13 || c=='\n') {
				line_buffer(i)='\0';
				fputc('\n',stdout); // echo the Carriage Return
				break;
			}
			else if (c==127) {
				i--;
				fputc('\b',stdout); // echo the backspace
				continue;
			}
			else {
				fputc(c,stdout); // force echo (readline do not)
				line_buffer(i)=c;
			}
			i++;
		}
		/* reset >IN */
		to_in_(0); 	// required
		BLK1=BLK=0;	// required
		OUT=0;
	}
	else codeup(_QUERY);
}


/* built-in EXPECT */
void _expect(void) {
	if (STATE==INTERPRET) {
		unsigned short addr;
		short len;
		int i=0;
		int c;
		len=_pop();
		addr=_pop();
		if (len<=0) return;
		_doerr((len>ILINE-1),19);
		/* get string */
		i=len; // back up length
		while (len>0) {
			c=getchar();
			if (c==10 || c==13 || c=='\n') {
				mem[addr++]='\0';
				mem[addr]='\0';
				fputc('\n',stdout); // echo the Carriage Return
				break;
			}
			/* treat backspace */
			else if (c==127) {
				mem[--addr]='\0';
				fputc('\b',stdout); // echo the backspace
				continue;
			}
			else if (c>=BL && c<=CHAR) {
				fputc(c,stdout); // echo char (readline doesn't)
				mem[addr++]=c;
			}
			len--;
		}
		SPAN=i-len;
		/* blank the rest of characters in the addr buffer;
		 * Note: the Standard required that one or two nulls are
		 * added at the end of text. 
		 * Note: gforth does not reset the input buffer but 
		 * I like it this way...*/
		while (len>0) mem[addr++]='\0',len--;
		OUT=0;
	}
	else codeup(_EXPECT);
}


/* built-in [ - immediate 
 * sets interpretation mode*/
void _open_square(void) {
	if (STATE==INTERPRET) 
		provokeSTATE=TRUE;
	else STATE=INTERPRET;
}


/* built-in ] */
void _closed_square(void) {
	if (STATE==COMPILE) {
		codeup(_CLOSED_SQUARE);
	}
	else STATE=COMPILE;
}


/* built-in LITERAL */
void _literal(void) {
	if (STATE==COMPILE) {
		short val;
		val=_pop();
		mem[HERE++]='I';
		mem[HERE++]=val%BYTE;
		mem[HERE++]=val/BYTE;
	}
}


/* built-in DLITERAL */
void _dliteral(void) {
	if (STATE==COMPILE) {
		int v1, v2;
		v2=_pop();
		v1=_pop();
		mem[HERE++]='D';
		mem[HERE++]=(int)(v2%BYTE);
		mem[HERE++]=v2/BYTE;
		mem[HERE++]=(int)(v1%BYTE);
		mem[HERE++]=v1/BYTE;
	}
}


/* built-in CONVERT */
void _convert(void) {
	if (STATE==INTERPRET) {
		unsigned short addr, len=1;
		int i=0, sign=1;
		long long int high, low;
		long long int val=0, num=0;
		char str[ILINE+1], *s=str;
		addr=_pop();
		high=_pop();
		low=_pop();
		len=mem[addr++];
		_doerr(len>=ILINE,24);
		val=int2ll(high,low);
		/* get string 
		 * Note: CONVERT doesn't specify if + or - are rejected 
		 * So I accept them */
		if (mem[addr]=='-') sign=-1, addr++, len--;
		else if (mem[addr]=='+') addr++, len--;
		while (i<len) {
			if (mem[addr+i]!='.') *s++=mem[addr+i];
			i++;
		}
		*s='\0';
		/* get number according to BASE */
		s=str;
		num=strtol(s,&s,BASE);
		_doerr((*s),23);
		/* sum values and push results */
		val+=sign*num;
		ll2int(val,&high,&low);
		_push(low);
		_push(high);
		_push(addr+strlen(str));
	}
	else codeup(_CONVERT);
}


/* built-in FILL */
void _fill(void) {
	if (STATE==INTERPRET) {
		unsigned short addr;
		short num;
		unsigned char byte;
		int i;
		byte=_pop(); // filling byte
		num=_pop();
		addr=_pop();
		if (num<=0) return;
		for (i=0;i<num;i++) mem[addr+i]=byte;
	}
	else codeup(_FILL);
}


/* built-in MOVE 
 * Note: actually a COPY */
void _move(void) {
	if (STATE==INTERPRET) {
		unsigned short addr2, addr1;
		short num;
		unsigned int i;
		num=_pop();
		num*=2; // copying numbers
		addr2=_pop();
		addr1=_pop();
		if (num>0) {
			/* copy source to dest */
			for (i=0;i<num;i++) 
				mem[addr2+i]=mem[addr1+i];
		}
	}
	else codeup(_MOVE);
}


/* built-in CMOVE 
 * Note; actually it's a COPY */
void _cmove(void) {
	if (STATE==INTERPRET) {
		unsigned short addr2, addr1;
		short num;
		unsigned int i;
		num=_pop(); // copying characters
		addr2=_pop();
		addr1=_pop();
		if (num>0) {	
			/* copy source to dest */
			for (i=0;i<num;i++) 
				mem[addr2+i]=mem[addr1+i]&FFCHAR;
		}
	}
	else codeup(_CMOVE);
}


/* built-in -TRAILING */
void _trailing(void) {
	if (STATE==INTERPRET) {
		unsigned short addr;
		short num;
		int i;
		num=_pop();
		addr=_pop();
		_doerr((!addr || addr+num>=ADDRLIMIT),20);
		_doerr(num<0,24);
		i=addr+num-1;
		while (isblank(mem[i]) && i>=addr) i--;
		_push(addr);
		_push(i-addr+1);
	}
	else codeup(_TRAILING);
}


/* built-in WORD */
void _word(unsigned short addr) {
	if (STATE==INTERPRET) {
		_act_word(addr);
	}
	else if (STATE==COMPILE) {
		codeup(_ACT_WORD);
	}
}


/* built-in runtime WORD 
 * Note: the argument addr is at present unused */
void _act_word(unsigned short addr) {
	/* direct execution: interpret a keyboard command parsing text 
	 * and storing characters to be printed */
	static int offset=0;
	int ch, b;
	int fullen=FALSE;
	int destination;

	/* decide WORD chunk */
	if (offset>3) offset=0;
	destination=WORDBUFFERLIMIT+offset*ILINE;
	/* in fig-Forth mode set destination to HERE */
	if (isFigCompatEnabled) destination=HERE;
	/* clear temporary buffer */
	b=destination;
	while (b<destination) mem[b++]=0;
	/* read terminator; if terminator is 1, read text until the end
	 * of the line buffer */ 
	ch=_pop();
	ch=((char)ch)%CHAR; // delimiter
	if (ch==1) fullen=TRUE;
	/* error in case the delimiter is the ASCII null or negative*/
	_doerr(ch<=0,55);
	b=destination+1;
	/* remove first trailing blank, the necessary word separator */
	if (isBlank(line_buffer(to_in))) _inc_to_in_();
	/* Note: the Standard explicitly requires that leading delimiters
	 * should be ignored (page 31); I deliberately ignore leading 
	 * delimiters only when they are blanks, because otherwise this would
	 * prevent the treating of empty strings 
	 * (see also FD Vol. III N. 3) */
	if (isblank(ch)) {
		while (line_buffer(to_in)==ch && to_in<B_BUF) _inc_to_in_();
	}
	/* read until terminator */
	if (!fullen) {
		while (line_buffer(to_in) && line_buffer(to_in)!=ch &&\
			       (b-destination<ILINE) && to_in<B_BUF) {
			mem[b++]=line_buffer(to_in);
			_inc_to_in_();
		}
		/* skip terminator */
		if (line_buffer(to_in)==ch) _inc_to_in_();
		/* add terminator, but if the input stream is exhausted 
		 * before encountering the delimiter, put a zero, else put ch */
		if (!line_buffer(to_in)) mem[b]='\0';
		else mem[b]=ch;
		mem[b+1]='\0';
	}
	/* read until end of buffer */
	else {
		while(line_buffer(to_in) && (b-destination<ILINE) &&\
			       to_in<B_BUF) { 
			mem[b++]=line_buffer(to_in);
			_inc_to_in_();
		}
		mem[b]='\0';
		mem[b+1]='\0';
	}
	/* push counted string without counting the final terminator 
	 * (whichever it is) */
	mem[destination]=b-(destination+1); // len
	/* push address only in Forth-79 mode */
	if (!isFigCompatEnabled) _push(destination);
	offset++;
}


/* built-in compile-time part of BEGIN */
void _begin(void) {
	_cfpush(HERE); // position of BEGIN
	codeup(_RT_BEGIN);
}

/* built-in run-time part of BEGIN*/
void _rt_begin(void) {
	/* Note: run-time part of BEGIN is a noop, since it serves as a
	 * hook for back-jumping from REPEAT/UNTIL/AGAIN */
}


/* built-in compile-time part of AGAIN */
void _again(void) {
	codeup(_RT_AGAIN);
	_doerr(cfs<1,32);
	mem[HERE++]=_cfpop();
}


/* built-in run-time part of AGAIN */
void _rt_again(unsigned short addr) {
	PC_(mem[addr+1]);
}

		
/* built-in compile-time part of UNTIL */
void _until(void) {
	codeup(_RT_UNTIL);
	_doerr(cfs<1,35);
	mem[HERE++]=_cfpop();
}


/* built-in run-time part of UNTIL */
void _rt_until(unsigned short addr) {
	int val;
	val=_pop();
	if (val) PC_(addr); 
	else PC_(mem[addr+1]);
}


/* built-in compile-time part of WHILE */
void _while(void) {
	codeup(_RT_WHILE);
	_cfpush(HERE);
	mem[HERE++]=0; // address of REPEAT
}


/* built-in run-time part of WHILE */
void _rt_while(unsigned short addr) {
	int val;
	val=_pop();
	/* jump to REPEAT to stop the cycle */
	if (!val) PC_(mem[addr+1]); // WHILE has one more cell
	else {
		int v=_PC+1;
		PC_(v); // discard the +1 address cell of WHILE
	}
}


/* built-in compile-time part of REPEAT */
void _repeat(void) {
	int val=0;
	codeup(_RT_REPEAT);
	_doerr(cfs<1,34);
	val=_cfpop(); // pushed by WHILE
	mem[val]=HERE-1;
	_doerr(cfs<1,34);
	val=_cfpop(); // pushed by BEGIN
	mem[HERE++]=val;
}


/* built-in run-time part of REPEAT */
void _rt_repeat(unsigned short addr) {
	PC_(mem[addr+1]);
}


/* built-in compile-time part of DO */
void _do(void) {
	_cfpush(HERE); // DO's position
	codeup(_RT_DO);
	loops++;
}


/* built-in run-time part of DO */
void _rt_do(void) {
	short start, end;
	start=_pop();
	end=_pop();
	_rspush(end);
	_rspush(start);	// index
	loops++;
}


/* built-in compile-time part of LOOP */
void _loop(void) {
	codeup(_RT_LOOP);
	_doerr(cfs<1,33);
	mem[HERE++]=_cfpop();
	loops--;
}


/* built-in LOOP */
void _rt_loop(void) {
	short index, end;
	_doerr(loops<=0,10);
	index=_rspop();
	index+=1;
	end=_rspop();
	if (index>=end) {
		int v=_PC+1;
		loops--;
		PC_(v);
	}
	else {
		int v=mem[_PC+2];
		_rspush(end);
		_rspush(index);
		PC_(v); // address of DO
	}
}


/* built-in compile-time part of +LOOP */
void _plus_loop(void) {
	codeup(_RT_PLUS_LOOP);
	_doerr(cfs<1,31);
	mem[HERE++]=_cfpop();
	loops--;
}


/* built-in run-time part of +LOOP */
void _rt_plus_loop(void) {
	short index, end;
	short add;
	_doerr(loops<=0,3);
	add=_pop();
	index=_rspop();
	index+=add;
	end=_rspop();
	/* fig-Forth compatibility control */
	if (isFigCompatEnabled &&\
		     ((add>0 && index>=end) || (add<0 && index<=end) || !add)) {
		int v=_PC+1;
		loops--;
		PC_(v);
	}
	/* Forth-79 control */
	else if ((add>0 && index>=end) || (add<0 && index<end) || !add) {
		int v=_PC+1;
		loops--;
		PC_(v);
	}
	else {
		int v=mem[_PC+2];
		_rspush(end);
		_rspush(index);
		PC_(v); // address of DO
	}
}


/* built-in I */
void _i(void) {
	if (STATE==INTERPRET) {
		short v;
		if (!isFigCompatEnabled) {
			_doerr(loops<1,7);
		}
		_doerr((rs-RETURNSTACKSTART<2*1),25);
		v=_rspop();
		_rspush(v);
		_push(v);
	}
	else codeup(_I);
}


/* builtin J */
void _j(void) {
	if (STATE==INTERPRET) {
		short vi1, vi2, vj1;
		_doerr(loops<2,8);
		_doerr((rs-RETURNSTACKSTART<2*2),25);
		vi1=_rspop();
		vi2=_rspop();
		vj1=_rspop();
		_rspush(vj1);
		_rspush(vi2);
		_rspush(vi1);
		_push(vj1);
	}
	else codeup(_J);
}


/* built-in LEAVE */
void _leave(void) {
	if (STATE==INTERPRET) {
		short index;
		_doerr(loops<=0,9);
		index=_rspop(); // index
		_rspop(); // end
		_rspush(index);
		_rspush(index);
	}
	/* Note: also gforth lets compile LEAVE outside a DO..LOOP, and issues
	 * the "Invalid memory address" in run-time */
	else codeup(_LEAVE);
}


/* built-in compile-time part of IF */
void _if(void) {
	codeup(_RT_IF);
	_cfpush(HERE);
	mem[HERE++]=0;
}


/* built-in run-time part of IF */
void _rt_if(void) {
	short val;
	val=_pop();
	/* if true, execute next word */
	if (val) {
		int v=_PC+1;
		PC_(v);
		_cfpush(TRUE);
	}
	/* if false, execute ELSE clause */
	else {
		int v=mem[_PC+2]-2;
		PC_(v); // to ELSE
		_cfpush(FALSE);
	}
}


/* built-in compile-part of ELSE */
void _else(void) {
	unsigned short thaddr=(unsigned short)_cfpop();
	codeup(_RT_ELSE);
	mem[thaddr]=HERE-2;
	_cfpush(HERE);
	mem[HERE++]=0;
}


/* built-in run-time part of ELSE */
void _rt_else(void) {
	int flag;
	flag=_cfpop();
	/* if false, execute next word (else clause) */
	if (!flag) {
		int v=_PC+1;
		PC_(v);
	}
	/* else, if clause has already been executed, jump to THEN */
	else {
		int v=mem[_PC+2]-2;
		PC_(v);
	}
	_cfpush(flag);
}


/* built-in compile-time part of THEN */
void _then(void) {
	unsigned short thaddr=(unsigned short)_cfpop();
	codeup(_RT_THEN);
	mem[thaddr]=HERE-2;
}


/* built-in run-time parto of THEN */
void _rt_then(void) {
	_cfpop(); // unload the flag put by IF (and ELSE)
}


/* builtin compile-time part of CASE */
void _case(void) {
	codeup(_RT_CASE);
	_cfpush(-1); // mark the end of ENDCASE search 
}


/* built-in run-time part of CASE*/
void _rt_case(void) {
	/* Note: run-time part of CASE is a noop */
}


/* builtin compile-time part of OF */
void _of(void) {
	codeup(_RT_OF);
	_cfpush(HERE);
	mem[HERE++]=0;
}


/* builtin run-time part of OF */
void _rt_of(void) {
	int v, val, myval;
	_over(); // original value is preserved
	val=_pop();
	myval=_pop();
	if (val==myval) {
		_cfpush(TRUE);
		v=_PC+1;
		PC_(v); 
	}
	else {
		_cfpush(FALSE);
		v=mem[_PC+2]-2;
		PC_(v);// jump to its proper ENDOF
	}
}


/* builtin compile-time part of ENDOF */
void _endof(void) {
	int ofpos=_cfpop();
	mem[ofpos]=HERE;
	codeup(_RT_ENDOF);
	_cfpush(HERE); // gathered by ENDCASE
	mem[HERE++]=0;
}


/* builtin run-time part of ENDOF */
void _rt_endof(void) {
	int v, test;
	test=_cfpop(); // left by OF
	if (test) {
		v=mem[_PC+2]-2;
		PC_(v); // jump to ENDCASE
	}
	/* in case test is false, go on with next instruction */
	else {
		v=_PC+1;
		PC_(v);
	}
}


/* builtin compile-time part of ENDCASE */
void _endcase(void) {
	int thaddr, myCC=HERE;
	codeup(_RT_ENDCASE);
	thaddr=_cfpop();
	while (thaddr>0) {
		mem[thaddr]=myCC; // compile ENDCASE into each ENDOF
		thaddr=_cfpop();
	}
}


/* builtin ENDCASE */
void _rt_endcase(void) {
	_pop(); // pop original value
}


/* defer_compile() - run-time part of COMPILE
 * executed when a COMPILE is compiled within a word  
 * Note: the Forth-79 standard asserts, in the two-page handy reference, that 
 * COMPILE should compile the address of the next non-IMMEDIATE word into the 
 * dictionary when the word containing COMPILE is executed (not when it's
 * compiled) 
 t Note: the different treatment for (SLITERAL) is due to the fact that is
 * this specific case, the code for _PUSH_STRING_ADDR must be compiled, rather
 * than its address, because the _push_string_addr() routine needs the proper
 * address in order to find the string to print. */
void _defer_compile(void) {
	unsigned short v;
	if (_find_word("(sliteral)",TRUE)==mem[_PC+2]) {
		mem[HERE++]=BWD;
		mem[HERE++]=_PUSH_STRING_ADDR;
	}
	else{
		mem[HERE++]=UWD;
		mem[HERE++]=mem[_PC+2];
	}
	v=_PC+1;
	PC_(v);
}


/* built-in compile-time part of COMPILE
 * Note: at page 22 where COMPILE is explained, no mention is made about 
 * immediacy. Rather, it's explicitly written that COMPILE DUP will copy
 * the compilation address of DUP.  */
void _compile(void) {
	unsigned int i=to_in;
	while (isBlank(line_buffer(i))) i++;
	if (line_buffer(i)=='[' && (isBlank(line_buffer(i+1)) || !line_buffer(i+1))) {
		/* do nothing: the [ ] token is supposed to compile directly
		 * with comma or C,
		 * See COMPILE in the Standard where it's stated: 
		 *    COMPILE  [ 0 , ]   will copy zero. */
		return;
	}
	mem[HERE++]=BWD;
	mem[HERE++]=_DEFERCOMPILE;
	compile_ON=TRUE;
}


/* built-in [COMPILE] 
 * Note: the Forth-79 standard asserts, in the two-page handy reference, that 
 * [COMPILE] should compile the address of the next word into the dictionary
 * even if immediate, but at page 31 where [COMPILE] is explained, no mention 
 * is made about strict requirement of an immediate word. Thus, no warning 
 * is issued in case a not immediate word is given as argument */
void _brackets_compile(void) {
	unsigned short addr=0;
	char bib[B_BUF+1];
	unsigned int len=1;
	b=bib;
	while (isBlank(line_buffer(to_in))) _inc_to_in_(); 
	_doerr(!line_buffer(to_in),36);
	while (line_buffer(to_in) && !isBlank(line_buffer(to_in)) &&\
			len<WORDNAMELEN) { 
		*b++=line_buffer(to_in);
		_inc_to_in_();
		len++;
	}
	*b='\0';
	addr=_find_word(bib,TRUE);
	_doerr(!addr,14);
	/* since addr is relative, it's an user word (use UWD, not BWD!) */
	mem[HERE++]=UWD;
	mem[HERE++]=addr;
}


/* built-in IMMEDIATE */
void _immediate(void) {
	cbits(LAST)|=B_PREC;
}


/* built-in COMPILE-ONLY */
void _compile_only(void) {
	cbits(LAST)|=B_COMPILE;
}


/* built-in EXEC-ONLY */
void _exec_only(void) {
	cbits(LAST)|=B_EXECUTION;
}


/* built-in sharp # */
void _sharp(void) {
	if (STATE==INTERPRET) {
		long long int high, low;
		unsigned int num;
		long long int quot, rem;
		_doerr(!isPict,0);
		high=_pop();
		low=_pop();
		num=int2ll(high,low);
		quot=num/BASE;
		rem=num-quot*BASE;
		mem[picp++]=digits[rem],FLD++;
		ll2int(quot,&high,&low);
		_push(low);
		_push(high);
	}
	else codeup(_SHARP);
}


/* built-in lesser_sharp <# */
void _lesser_sharp(void) {
	if (STATE==INTERPRET) {
		picp=SC;
		FLD=0;
		isPict=TRUE;
	}
	else codeup(_LESSER_SHARP);
}


/* built-in sharp_greater #> */
void _sharp_greater(void) {
	if (STATE==INTERPRET) {
		char pictured[TEMPBUFFER+1], *pp=pictured;
		unsigned char i=0;
		_two_drop();
		_doerr(!isPict,1);
		while (picp>=SC) *pp++=mem[(picp--)-1];
		*pp='\0';
		pp=pictured;
		while (*pp) mem[SC+i++]=*pp++;
		_push(SC);
		_push(i);
		HLD1=(SC+i-1)%BYTE;
		HLD=(SC+i-1)/BYTE;
		isPict=FALSE;
	}
	else codeup(_SHARP_GREATER);
}


/* built-in sharp_s #S */
void _sharp_s(void) {
	if (STATE==INTERPRET) {
		long long int high, low;
		unsigned int num;
		long long int val, quot, rem;
		_doerr(!isPict,2);
		high=_pop();
		low=_pop();
		num=int2ll(high,low);
		if (num==0) mem[picp++]='0', FLD++;
		else {
			val=num;
			while (val>0) {
				quot=val/BASE;
				rem=val-quot*BASE;
				mem[picp++]=digits[rem], FLD++;
				val=quot;
			}
		}
		_push(0); // required by Forth-79 Standard
		_push(0);
	}
	else codeup(_SHARP_S);
}


/* built-in HOLD */
void _hold(void) {
	if (STATE==INTERPRET) {
		unsigned short ch;
		_doerr(!isPict,6);
		ch=_pop();
		ch=ch&CHAR;
		mem[picp++]=ch, FLD++;
	}
	else codeup(_HOLD);
}


/* built-in SIGN */
void _sign(void) {
	if (STATE==INTERPRET) {
		short val; 
		_doerr(!isPict,11);
		val=_pop(); 
		if (val<0) mem[picp++]='-',FLD++;
	}
	else codeup(_SIGN);
}


/* built-in QUIT */
void _quit(void) {
	if (STATE==INTERPRET) {
		_erase_stacks(RETURN_STACK);
		isQuit=TRUE;
	}
	else codeup(_QUIT);
}


/* built-in CR */
void _cr(void) {
	if (STATE==INTERPRET) {
		fputc('\n',stdout);
		OUT=0;
	}
	else codeup(_CR);
}


/* built-in R@ */
void _r_fetch(void) {
	if (STATE==INTERPRET) {
		short val;
		val=_rspop();
		_push(val);
		_rspush(val);
	}
	else codeup(_R_FETCH);
}


/* built-in R> */
void _r_to(void) {
	if (STATE==INTERPRET) {
		short val;
		val=_rspop();
		_push(val);
	}
	else codeup(_R_TO);
}


/* built-in >R */
void _to_r(void) {
	if (STATE==INTERPRET) {
		short val;
		val=_pop();
		_rspush(val);
	}
	else codeup(_TO_R);
}


/* built-in 0> */
void _zero_greater(void) {
	if (STATE==INTERPRET) {
		short val;
		val=_pop();
		_push(val>0);
	}
	else codeup(_ZERO_GREATER);
}


/* built-in < */
void _lesser(void) {
	if (STATE==INTERPRET) {
		short val2, val1;
		val2=_pop();
		val1=_pop();
		_push(val1<val2);
	}
	else codeup(_LESSER);
}


/* built-in > */
void _greater(void) {
	if (STATE==INTERPRET) {
		short val2, val1;
		val2=_pop();
		val1=_pop();
		_push(val1>val2);
	}
	else codeup(_GREATER);
}


/* built-in 0< */
void _zero_lesser(void) {
	if (STATE==INTERPRET) {
		short val;
		val=_pop();
		_push(val<0);
	}
	else codeup(_ZERO_LESSER);
}


/* built-in ?DUP */
void _question_dup(void) {
	if (STATE==INTERPRET) {
		short val;
		val=_pop();
		if (val) {
			_push(val);
			_push(val);
		}
		else _push(0);
	}
	else codeup(_QUESTION_DUP);
}


/* built-in AND */
void _and(void) {
	if (STATE==INTERPRET) {
		short val1, val2;
		val1=_pop();
		val2=_pop();
		_push(val1&val2);
	}
	else codeup(_AND);
}


/* built-in OR */
void _or(void) {
	if (STATE==INTERPRET) {
		short val1, val2;
		val1=_pop();
		val2=_pop();
		_push(val1|val2);
	}
	else codeup(_OR);
}


/* built-in XOR */
void _xor(void) {
	if (STATE==INTERPRET) {
		short val1, val2;
		val1=_pop();
		val2=_pop();
		_push(val1^val2);
	}
	else codeup(_XOR);
}


/* built-in 0=/NOT */
void _zero_equal(void) {
	if (STATE==INTERPRET) {
		short val;
		val=_pop();
		_push(!val);
	}
	else codeup(_ZERO_EQUAL);
}


/* built-in = */
void _equal(void) {
	if (STATE==INTERPRET) {
		short val1, val2;
		val1=_pop();
		val2=_pop();
		_push(val1==val2);
	}
	else codeup(_EQUAL);
}


/* built-in .R */
void _dot_r(void) {
	if (STATE==INTERPRET) {
		int field, n=0, ival;
		short val;
		char str[ILINE+1],*s=str;
		field=_pop();
		val=_pop();
		ival=val;
		n=_print(ival,BASE,s);;
		while (n<field) {
			fputc(' ',stdout);
			n++;
			OUT++;
		}
		fprintf(stdout,"%s",str);
		OUT+=strlen(str);
	}
	else codeup(_DOT_R);
}


/* built-in U.R */
void _u_dot_r(void) {
	if (STATE==INTERPRET) {
		int field, n=0, ival;
		unsigned short val;
		char str[ILINE+1],*s=str;
		field=_pop();
		val=_pop();
		ival=val;
		n=_print(ival,BASE,s);;
		while (n<field) {
			fputc(' ',stdout);
			n++;
			OUT++;
		}
		fprintf(stdout,"%s",str);
		OUT+=strlen(str);
	}
	else codeup(_U_DOT_R);
}


/* built-in D.R */
void _d_dot_r(void) {
	if (STATE==INTERPRET) {
		short field;
		int n=0;
		long long int high, low;
		char str[ILINE+1],*s=str;
		field=_pop();
		high=_pop();
	        low=_pop();
		n=_print(int2ll(high,low),BASE,s);
		while (n<field) {
			fputc(' ',stdout);
			n++;
			OUT++;
		}
		fprintf(stdout,"%s",str);
		OUT+=strlen(str);
	}
	else codeup(_D_DOT_R);
}


/* built-in D0= */
void _d_zero_equal(void) {
	if (STATE==INTERPRET) {
		long long int high, low;
		high=_pop();
		low=_pop();
		if (!high && !low) _push(TRUE);
		else _push(FALSE);
	}
	else codeup(_D_ZERO_EQUAL);
}


/* built-in D= */
void _d_equal(void) {
	if (STATE==INTERPRET) {
		short high1, low1, high2, low2;
		high1=_pop();
		low1=_pop();
		high2=_pop();
		low2=_pop();
		if (high1==high2 && low1==low2) _push(TRUE);
		else _push(FALSE);
	}
	else codeup(_D_EQUAL);
}


/* built-in DMAX */
void _dmax(void) {
	if (STATE==INTERPRET) {
		long long int high2, low2, high1, low1;
		long long int val1,val2,val3;
		high2=_pop();
		low2=_pop();
		high1=_pop();
		low1=_pop();
		val1=int2ll(high1,low1);
		val2=int2ll(high2,low2);
		if (val1>=val2) val3=val1;
		else val3=val2;
		ll2int(val3,&high1,&low1);
		_push(low1);
		_push(high1);
	}
	else codeup(_DMAX);
}


/* built-in DMIN */
void _dmin(void) {
	if (STATE==INTERPRET) {
		long long int high2, low2, high1, low1;
		long long int val1,val2,val3;
		high2=_pop();
		low2=_pop();
		high1=_pop();
		low1=_pop();
		val1=int2ll(high1,low1);
		val2=int2ll(high2,low2);
		if (val1<=val2) val3=val1;
		else val3=val2;
		ll2int(val3,&high1,&low1);
		_push(low1);
		_push(high1);
	}
	else codeup(_DMIN);
}


/* built-in DU< */
void _d_u_lesser(void) {
	if (STATE==INTERPRET) {
		long long int high2, low2, high1, low1;
		unsigned int val1, val2;
		high2=_pop();
		low2=_pop();
		high1=_pop();
		low1=_pop();
		val1=int2ll(high1,low1);
		val2=int2ll(high2,low2);
		if (val1<val2) _push(TRUE);
		else _push(FALSE);
	}
	else codeup(_D_U_LESSER);
}


/* built-in integer division - default: rounding toward zero 
 * Note: the ANSI 1994 division (starting from Forth-83) performs a rounding
 * toward ceiling in case all operands have the same sign and toward floor
 * in case they differ. This means that -5 / 4 is -2.
 * But Forth-79 used a more logical approach, simply rounding toward zero the
 * integer value, so that (as natural mind expects) -5 / 4 is -1.*/
void _slash(void) {
	if (STATE==INTERPRET) {
		short n, val;
		n=_pop();
		val=_pop();
		_doerr(!n,18);
		_push(val/n);
	}
	else codeup(_SLASH);
}


/* built-in D. */
void _d_dot(void) {
	if (STATE==INTERPRET) {
		long long int high, low;
		long int res;
		high=_pop();
		low =_pop();
		res=int2ll(high,low);
		OUT+=_print(res,BASE,NULL);
	}
	else codeup(_D_DOT);
}


/* built-in U. */
void _u_dot(void) {
	if (STATE==INTERPRET) {
		unsigned short val;
		val=_pop();
		OUT+=_print_unsigned(val,BASE,NULL);
	}
	else codeup(_U_DOT);
}


/* built-in UD. */
void _u_d_dot(void) {
	if (STATE==INTERPRET) {
		long long int high, low;
		unsigned int val;
		high=_pop();
		low =_pop();
		val=int2ll(high,low);
		OUT+=_print_double_unsigned(val,BASE,NULL);
	}
	else codeup(_U_D_DOT);
}


/* built-in dot command */
void _dot(void) {
	if (STATE==INTERPRET) {
		short val;
		val=_pop();
		OUT+=_print(val,BASE,NULL);
	}
	else codeup(_DOT);
}


/* built-in dot command */
void _h_dot(void) {
	if (STATE==INTERPRET) {
		unsigned short val;
		val=_pop();
		OUT+=_print_unsigned(val,16,NULL);
	}
	else codeup(_H_DOT);
}


/* built-in C! */
void _c_store(void) {
	if (STATE==INTERPRET) {
		unsigned short addr;
		signed char n;
		addr=_pop();
		_doerr(((addr<LOWMEM&&mem[addr-1]!='d')||addr>=DATASTACKSTART),20);
		n=_pop();
		n=n&FFCHAR;
		mem[addr]=n;
	}
	else codeup(_C_STORE);
}


/* built-in C@ */
void _c_fetch(void) {
	if (STATE==INTERPRET) {
		unsigned short addr;
		unsigned char val;	
		addr=_pop(); 
		val=mem[addr];
		_push(val);
	}
	else codeup(_C_FETCH);
}


/* built-in C, */
void _c_comma(void) {
	if (STATE==INTERPRET) {
		short val;
		val=_pop();
		_doerr(HERE+1>=HIGHMEM,30);
		mem[HERE++]=val&FFCHAR;
	}
	else codeup(_C_COMMA);
}


/* built-in D+ */
void _d_plus(void) {
	if (STATE==INTERPRET) {
		long long int high1, low1, high2, low2;
		long long int res,num1,num2,high,low;
		high1=_pop();
		low1=_pop();
		high2=_pop();
		low2=_pop();
		num1=int2ll(high1,low1);
		num2=int2ll(high2,low2);
		res=num1+num2;
		ll2int(res,&high,&low);
		_push(low);
		_push(high);
	}
	else codeup(_D_PLUS);
}


/* built-in D< */
void _d_lesser(void) {
	if (STATE==INTERPRET) {
		long long int high1, low1, high2, low2;
		long long int num1,num2;
		high1=_pop();
		low1=_pop();
		high2=_pop();
		low2=_pop();
		num1=int2ll(high1,low1);
		num2=int2ll(high2,low2);
		if (num1>num2) _push(TRUE);
		else _push(FALSE);
	}
	else codeup(_D_LESSER);
}


/* built-in D- */
void _d_minus(void) {
	if (STATE==INTERPRET) {
		long long int high1, low1, high2, low2;
		long long int res,num1,num2,high,low;
		high1=_pop();
		low1=_pop();
		high2=_pop();
		low2=_pop();
		num1=int2ll(high1,low1);
		num2=int2ll(high2,low2);
		res=num2-num1;
		ll2int(res,&high,&low);
		_push(low);
		_push(high);
	}
	else codeup(_D_MINUS);
}


/* built-in D* */
void _d_star(void) {
	if (STATE==INTERPRET) {
		long long int high1, low1, high2, low2;
		long long int res,num1,num2,high,low;
		high1=_pop();
		low1=_pop();
		high2=_pop();
		low2=_pop();
		num1=int2ll(high1,low1);
		num2=int2ll(high2,low2);
		res=num1*num2;
		ll2int(res,&high,&low);
		_push(low);
		_push(high);
	}
	else codeup(_D_STAR);
}


/* built-in D/ */
void _d_slash(void) {
	if (STATE==INTERPRET) {
		long long int high1, low1, high2, low2;
		long long int res,num1,num2,high,low;
		high1=_pop();
		low1=_pop();
		high2=_pop();
		low2=_pop();
		num1=int2ll(high1,low1);
		num2=int2ll(high2,low2);
		res=num2/num1;
		ll2int(res,&high,&low);
		_push(low);
		_push(high);
	}
	else codeup(_D_SLASH);
}


/* built-in UD/ */
void _u_d_slash(void) {
	if (STATE==INTERPRET) {
		long long int high1, low1, high2, low2;
		unsigned int num1, num2, res;
		long long int high, low;
		high1=_pop();
		low1=_pop();
		high2=_pop();
		low2=_pop();
		num1=int2ll(high1,low1);
		num2=int2ll(high2,low2);
		res=num2/num1;
		ll2int(res,&high,&low);
		_push(low);
		_push(high);
	}
	else codeup(_U_D_SLASH);
}


/* built-in SHIFT */
void _shift(void) {
	if (STATE==INTERPRET) {
		short places, val;
		places=_pop();
		val=_pop();
		if (places>0) _push(val<<places);
		else _push(val>>(-places));
	}
	else codeup(_SHIFT);
}


/* built-in MOD */
void _mod(void) {
	if (STATE==INTERPRET) {
		short n1, n2;
		short quot;
		n2=_pop();
		n1=_pop();
		_doerr(!n2,18);
		quot=n1/n2;
		_push(n1-quot*n2);
	}
	else codeup(_MOD);
}


/* built-in /MOD */
void _slash_mod(void) {
	if (STATE==INTERPRET) {
		short n1, n2;
		short quot;
		n2=_pop();
		n1=_pop();
		_doerr(!n2,18);
		quot=n1/n2;
		_push(n1-quot*n2);
		_push(quot);
	}
	else codeup(_SLASH_MOD);
}


/* built-in °/MOD star_slash_mod */
void _star_slash_mod(void) {
	if (STATE==INTERPRET) {
		short n3, n2, n1, quot;
		int val;
		n3=_pop();
		n2=_pop();
		n1=_pop();
		_doerr(!n3,18);
		val=n1*n2;
		quot=val/n3;
		_push(val-quot*n3);
		_push(quot);
	}
	else codeup(_STAR_SLASH_MOD);
}


/* built-in U/MOD */
void _u_slash_mod(void) {
	if (STATE==INTERPRET) {
		unsigned short n;
		long long int high, low;
		int quot, val;
		n=_pop();
		high=_pop();
		low=_pop();
		_doerr(!n,18);
		val=int2ll(high,low);
		quot=val/n;
		_push(val-quot*n);
		_push(quot);
	}
	else codeup(_U_SLASH_MOD);
}


/* built-in U* */
void _u_star(void) {
	if (STATE==INTERPRET) {
		long long int high,low;
		unsigned short v2, v1;
		int val;
		v2=_pop();
		v1=_pop();
		val=v1*v2;
		ll2int(val,&high,&low);
		_push(low);
		_push(high);
	}
	else codeup(_U_STAR);
}


/* built-in U< */
void _u_lesser(void) {
	if (STATE==INTERPRET) {
		unsigned short n2, n1;
		n2=_pop();
		n1=_pop();
		_push(n1<n2);
	}
	else codeup(_U_LESSER);
}


/* built-in ABS */
void _abs(void) {
	if (STATE==INTERPRET) {
		short val;
		val=_pop();
		if (val<0) val*=-1;
		_push(val);
	}
	else codeup(_ABS);
}


/* built-in DABS */
void _dabs(void) {
	if (STATE==INTERPRET) {
		long long int high, low, dnum;
		high=_pop();
		low =_pop();
		dnum=int2ll(high,low);
		if (dnum<0) dnum=-dnum;
		ll2int(dnum,&high,&low);
		_push(low);
		_push(high);
	}
	else codeup(_DABS);
}


/* built-in MAX */
void _max(void) {
	if (STATE==INTERPRET) {
		short v1, v2;
		v1=_pop();
		v2=_pop();
		if (v1>v2) _push(v1);
		else _push(v2);
	}
	else codeup(_MAX);
}


/* built-in MIN */
void _min(void) {
	if (STATE==INTERPRET) {
		short v1, v2;
		v1=_pop();
		v2=_pop();
		if (v2>v1) _push(v1);
		else _push(v2);
	}
	else codeup(_MIN);
}


/* built-in 1+ */
void _one_plus(void) {
	if (STATE==INTERPRET) {
		short val;
		val=_pop();
		_push(++val);
	}
	else codeup(_ONE_PLUS);
}


/* built-in 1- */
void _one_minus(void) {
	if (STATE==INTERPRET) {
		short val;
		val=_pop();
		_push(--val);
	}
	else codeup(_ONE_MINUS);
}



/* built-in 2+ */
void _two_plus(void) {
	if (STATE==INTERPRET) {
		short val;
		val=_pop();
		_push(val+2);
	}
	else codeup(_TWO_PLUS);
}


/* built-in 2- */
void _two_minus(void) {
	if (STATE==INTERPRET) {
		short val;
		val=_pop();
		_push(val-2);
	}
	else codeup(_TWO_MINUS);
}


/* built-in ALLOT */
void _allot(void) {
	if (STATE==INTERPRET) {
		short val;
		val=_pop();
		_doerr(HERE+val>=HIGHMEM,30);
		_doerr(HERE+val<LOWMEM,30);
		HERE+=val;
	}
	else codeup(_ALLOT);
}


/* built-in VARIABLE */
void _variable(void) {
	if (STATE==INTERPRET) {
		int hb;
		int val=0;
		/* fig-Forth compatibility */
		if (isFigCompatEnabled) {
			val=_pop();
		}
		/* parse buffer for name */
		_doerr(!_fill_buff(backbuffer),38);
		hb=HERE;
		LAST_(HERE);
		/* check if redefined */
		if (_find_word(backbuffer,TRUE)) {
			if (!isFile && !isBlock && !isLoad) 
				fprintf(stdout," %s isn't unique.",backbuffer);
		}
		HERE=_add_var('v', HERE,PREVHERE,strlen(backbuffer),0,0,backbuffer,val);
		PREVHERE=hb;
		LAST_(PREVHERE);
	}
	else codeup(_VARIABLE);
}

/* built-in CONSTANT */
void _constant(void) {
	if (STATE==INTERPRET) {
		int hb, val;
		val=_pop();
		_doerr(!_fill_buff(backbuffer),51);
		hb=HERE;
		LAST_(HERE);
		/* check if redefined */
		if (_find_word(backbuffer,TRUE)) {
			if (!isFile && !isBlock && !isLoad) 
				fprintf(stdout," %s isn't unique.",backbuffer);
		}
		HERE=_add_var('c',HERE,PREVHERE,strlen(backbuffer),0,0,backbuffer,val);
		PREVHERE=hb;
		LAST_(PREVHERE);
	}
	else codeup(_CONSTANT);
}


/* built-in 2VARIABLE */
void _two_variable(void) {
	if (STATE==INTERPRET) {
		int hb;
		int val1=0, val2=0;
		/* fig-Forth compatibility */
		if (isFigCompatEnabled) {
			val1=_pop();
			val2=_pop();
		}
		/* parse buffer for name */
		_doerr(!_fill_buff(backbuffer),38);
		hb=HERE;
		LAST_(HERE);
		/* check if redefined */
		if (_find_word(backbuffer,TRUE)) {
			if (!isFile && !isBlock && !isLoad) 
				fprintf(stdout," %s isn't unique.",backbuffer);
		}
		HERE=_add_2_var('V',HERE,PREVHERE,strlen(backbuffer),
				0,0,backbuffer,val1,val2);
		PREVHERE=hb;
		LAST_(PREVHERE);
	}
	else codeup(_TWO_VARIABLE);
}


/* built-in 2CONSTANT */
void _two_constant(void) {
	if (STATE==INTERPRET) {
		int hb, val2, val1;
		val1=_pop();
		val2=_pop();
		/* parse buffer for name */
		_doerr(!_fill_buff(backbuffer),51);
		hb=HERE;
		LAST_(HERE);
		/* check if redefined */
		if (_find_word(backbuffer,TRUE)) {
			if (!isFile && !isBlock && !isLoad) 
				fprintf(stdout," %s isn't unique.",backbuffer);
		}
		HERE=_add_2_var('C',HERE,PREVHERE,strlen(backbuffer),
				0,0,backbuffer,val1,val2);
		PREVHERE=hb;
		LAST_(PREVHERE);
	}
	else codeup(_TWO_CONSTANT);
}


/* _question() - built-in ? */
void _question(void) {
	if (STATE==INTERPRET) {
		unsigned short addr;
		short val;
		addr=_pop();
		val=(mem[addr]%BYTE)+BYTE*(mem[addr+1]%BYTE);
		_print(val,BASE,NULL);
	}
	else codeup(_QUESTION);
}


/* built-in 32, */
void _32_comma(void) {
	if (STATE==INTERPRET) {
		long long int val, high,low;
		_doerr(HERE+1>=HIGHMEM,30);
		high=_pop();
		low=_pop();
		val=int2ll(high,low);
		mem[HERE++]=val;
	}
	else codeup(_32_COMMA);
}


/* built-in comma */
void _comma(void) {
	if (STATE==INTERPRET) {
		short val;
		val=_pop();
		_doerr(HERE+1>=HIGHMEM,30);
		mem[HERE++]=val%BYTE;
		mem[HERE++]=val/BYTE;
	}
	else codeup(_COMMA);
}


/* built-in ! */
void _store(void) {
	if (STATE==INTERPRET) {
		unsigned short addr;
		short val;
		addr=_pop();
		_doerr(((addr<LOWMEM&&mem[addr-1]!='d')||addr>=DATASTACKSTART),20);
		val=_pop();
		mem[addr]=(int)(val%BYTE);
		mem[addr+1]=val/BYTE;
	}
	else codeup(_STORE);
}


/* built-in +! */
void _plus_store(void) {
	if (STATE==INTERPRET) {
		unsigned short addr;
		short val,orig;
		addr=_pop();
		_doerr(((addr<LOWMEM&&mem[addr-1]!='d')||addr>=DATASTACKSTART),20);
		val=_pop();
		orig=(mem[addr]%BYTE)+BYTE*(mem[addr+1]%BYTE);
		mem[addr]=(val+orig)%BYTE;
		mem[addr+1]=(val+orig)/BYTE;
	}
	else codeup(_PLUS_STORE);
}


/* built-in 2! */
void _two_store(void) {
	if (STATE==INTERPRET) {
		unsigned short addr;
		long long int high, low;
		addr=_pop();
		_doerr((addr<LOWMEM || addr>=DATASTACKSTART),20);
		high=_pop();
		low=_pop();
		mem[addr]=(int)(high%BYTE);
		mem[addr+1]=high/BYTE;
		mem[addr+2]=(int)(low%BYTE);
		mem[addr+3]=low/BYTE;
	}
	else codeup(_TWO_STORE);
}


/* built-in @ */
void _fetch(void) {
	if (STATE==INTERPRET) {
		unsigned short addr;
		addr=_pop();
		_push((mem[addr]%BYTE)+BYTE*(mem[addr+1]%BYTE));
	}
	else codeup(_FETCH);
}


/* built-in 32! */
void _32_store(void) {
	if (STATE==INTERPRET) {
		unsigned short addr;
		long long int val, high,low;
		addr=_pop();
		_doerr((addr<LOWMEM || addr>=DATASTACKSTART),20);
		high=_pop();
		low=_pop();
		val=int2ll(high,low);
		mem[addr]=val;
	}
	else codeup(_32_STORE);
}


/* built-in 32@ */
void _32_fetch(void) {
	if (STATE==INTERPRET) {
		long long int val,high,low;
		unsigned short addr;
		addr=_pop();
		val=mem[addr];
		ll2int(val,&high,&low);
		_push(low);
		_push(high);
	}
	else codeup(_32_FETCH);
}


/* built-in 2@ */
void _two_fetch(void) {
	if (STATE==INTERPRET) {
		unsigned short addr; 
		addr=_pop(); 
		_push((mem[addr+2]%BYTE)+BYTE*(mem[addr+3]%BYTE));
		_push((mem[addr]%BYTE)+BYTE*(mem[addr+1]%BYTE));
	}
	else codeup(_TWO_FETCH);
}


/* built-in ." 
 * Note: strings are stored as counted strings, that is
 * the length is stored immediately before the string.*/
void _dot_quotes(unsigned short addr) {
	char bib[B_BUF+1];
	if (STATE==INTERPRET) {
		/* direct execution: interpret a command parsing text 
		 * and storing characters in bib, to be later printed */
		if (!strcmp(word,".\""))  {
			b=bib;
			if (isBlank(line_buffer(to_in))) _inc_to_in_();
			while (line_buffer(to_in) && line_buffer(to_in)!='\"'){ 
				*b++=line_buffer(to_in);
				_inc_to_in_();
			}
			_doerr(line_buffer(to_in)!='\"',57);
			_inc_to_in_(); // skip closing quotes
			*b='\0';
			fprintf(stdout,"%s",bib);
			OUT+=strlen(bib);
		}
		/* write compiled string in case of indirect execution */
		else {
			addr++; // points to len
			int i=0, len=mem[addr++];
			while (i<len) {
				fputc(mem[addr+i++],stdout);
				OUT++;
			}
			i=addr+len-2;
			PC_(i); // skip string
		}
	}
	else {
		mem[HERE++]=BWD;
		mem[HERE++]=_DOT_QUOTES;
		b=bib; 
		if (isBlank(line_buffer(to_in))) _inc_to_in_();
		while (line_buffer(to_in) && line_buffer(to_in)!='\"') {
			*b++=line_buffer(to_in);
			_inc_to_in_();
		}
		_inc_to_in_(); // skip closing quotes
		*b='\0';
		b=bib;
		mem[HERE++]=strlen(bib);
		while (*b) mem[HERE++]=*b++;
	}
}


/* built-in NEGATE 
 * Note: while NOT is 0= and simply reverse the truth value, 
 * NEGATE is the two's complement of the argument number, that is the 
 * difference of 0 less n.*/
void _negate(void) {
	if (STATE==INTERPRET) {
		short v1;
		v1=_pop();
		_push(0-v1);
	}
	else codeup(_NEGATE);
}


/* built-in DNEGATE */
void _d_negate(void) {
	if (STATE==INTERPRET) {
		long long int val, high, low;
		high=_pop(); 
		low=_pop();
		val=int2ll(high,low);
		val=0-val;
		ll2int(val,&high,&low);
		_push(low);
		_push(high);
	}
	else codeup(_DNEGATE);
}


/* built-in DEPTH */
void _depth(void) {
	if (STATE==INTERPRET) {
		_push((ds-DATASTACKSTART)/2);
	}
	else codeup(_DEPTH);
}


/* built-in 2OVER */
void _two_over(void) {
	if (STATE==INTERPRET) {
		short v4, v3, v2, v1;
		v4=_pop();
		v3=_pop();
		v2=_pop();
		v1=_pop();
		_push(v1);
		_push(v2);
		_push(v3);
		_push(v4);
		_push(v1);
		_push(v2);
	}
	else codeup(_TWO_OVER);
}


/* built-in OVER */
void _over(void) {
	if (STATE==INTERPRET) {
		short v1, v2;
		v1=_pop();
		v2=_pop();
		_push(v2);
		_push(v1);
		_push(v2);
	}
	else codeup(_OVER);
}


/* built-in ROLL */
void _roll(void) {
	if (STATE==INTERPRET) {
		short n;
		int back1, back2, i;
		n=_pop();
		n*=2; // n in 2 cells
		_doerr(n<1,39);
		_doerr(((ds-DATASTACKSTART)<n),27);
		back1=mem[ds-n]; //xul
		back2=mem[ds-n+1]; //xuh
		for (i=0; i<n-1; i+=2) 
			mem[ds-n+i]=mem[ds-n+i+2],
			mem[ds-n+i+1]=mem[ds-n+i+3];
		mem[ds-2]=back1;
		mem[ds-1]=back2;
	}
	else codeup(_ROLL);
}


/* built-in 2ROT */
void _two_rot(void) {
	if (STATE==INTERPRET) {
		short high1,low1, high2, low2, high3, low3;
		high1=_pop();
		low1=_pop();
		high2=_pop();
		low2=_pop();
		high3=_pop();
		low3=_pop();
		_push(low2); 
		_push(high2);
		_push(low1); 
		_push(high1);
		_push(low3); 
		_push(high3);
	}
	else codeup(_TWO_ROT);
}


/* built-in ROT */
void _rot(void) {
	if (STATE==INTERPRET) {
		short v1, v2, v3;
		v1=_pop();
		v2=_pop();
		v3=_pop();
		_push(v2);
		_push(v1);
		_push(v3);
	}
	else codeup(_ROT);
}


/* built-in PICK */
void _pick(void) {
	if (STATE==INTERPRET) {
		short n;
		n=_pop();
		n*=2; // n in 2 cells
		_doerr(n<1,40);
		_doerr(((ds-DATASTACKSTART)<n),26);
		_push(mem[ds-n]+BYTE*mem[ds-n+1]);
		
	}
	else codeup(_PICK);
}


/* built-in 2DROP */
void _two_drop(void) {
	if (STATE==INTERPRET) {
		_pop();
		_pop();
	}
	else codeup(_TWO_DROP);
}


/* built-in DROP */
void _drop(void) {
	if (STATE==INTERPRET) {
		_pop();
	}
	else codeup(_DROP);
}


/* built-in 2DUP */
void _two_dup(void) {
	if (STATE==INTERPRET) {
		short val2, val1;
		val2=_pop();
		val1=_pop();
		_push(val1);
		_push(val2);
		_push(val1);
		_push(val2);
	}
	else codeup(_TWO_DUP);
}


/* built-in DUP */
void _dup(void) {
	if (STATE==INTERPRET) {
		short val;
		val=_pop();
		_push(val);
		_push(val);
	}
	else codeup(_DUP);
}


/* built-in 2SWAP */
void _two_swap(void) {
	if (STATE==INTERPRET) {
		short val4, val3, val2, val1;
		val4=_pop();
		val3=_pop();
		val2=_pop();
		val1=_pop();
		_push(val3);
		_push(val4);
		_push(val1);
		_push(val2);
	}
	else codeup(_TWO_SWAP);
}


/* built-in SWAP */
void _swap(void) {
	if (STATE==INTERPRET) {
		short val1, val2;
		val1=_pop();
		val2=_pop();
		_push(val1);
		_push(val2);
	}
	else codeup(_SWAP);
}


/* built-in M* / m-star-slash */
void _m_star_slash(void) {
	if (STATE==INTERPRET) {
		long long int res, val, high, low;
		short val1, val2;
		val1=_pop(); // n1
		val2=_pop(); // n2
		high=_pop();
		low=_pop();
		val=int2ll(high,low)*val2;
		res=val/val1;
		ll2int(res,&high,&low);
		_push(low);
		_push(high);
	}
	else codeup(_M_STAR_SLASH);
}


/* built-in * / star-slash */
void _star_slash(void) {
	if (STATE==INTERPRET) {
		int interm;
		short val1, val2, val3, res;
		val1=_pop();
		val2=_pop();
		val3=_pop();
		interm=val3*val2;
		res=interm/val1;
		_push(res);
	}
	else codeup(_STAR_SLASH);
}


/* built-in * (multiplication) */
void _star(void) {
	if (STATE==INTERPRET) {
		short val1, val2;
		val1=_pop();
		val2=_pop();
		_push(val1*val2);
	}
	else codeup(_STAR);
}


/* built-in + (addition) */
void _plus(void) {
	if (STATE==INTERPRET) {
		short val1,val2;
		val1=_pop();
		val2=_pop();
		_push(val1+val2);
	}
	else codeup(_PLUS);
}


/* built-in - (subtraction) */
void _minus(void) {
	if (STATE==INTERPRET) {
		short val1, val2;
		val1=_pop();
		val2=_pop();
		_push(val2-val1);
	}
	else codeup(_MINUS);
}


/* built-in BYE command 
 * Note: BYE is ignored in case it appears in a file belonging to a list of 
 * files typed as argument in the console invocation, when it appears in any 
 * but the last file. */
void _bye(void) {
	if (fr!=fc) return;
	isBye=TRUE;
}


/* built-in WORDS */
void _words(void) { 
	if (STATE==INTERPRET) {
		int count=0, val, v=30;
		int h=0; 
		fputc('\n',stdout);
		/* proceed from */
		while (v>=0) {
			val=(int)pow(2,v);
			if (vocChain[(CONTEXT)] & val) {
				h=PREVHERE; /* last vocabulary entry */
				while (h>0) {
					/* control Carriage Return */
					if (count + wordlen(h)>SCREEN) { 
						fputc('\n',stdout); 
						count=0; 
					}
					/* check smudge bit & vocabulary code */
					if (!isSmudge(h) && voccode(h)==v) {
						count+=_print_word_name(h);
						fputc(' ',stdout);
						count++;
					}
					h=prev(h);
				}
			}
			v--;
		}
	}
	else codeup(_WORDS);
}


/* built-in SEE */
void _see(void) {
	if (STATE==INTERPRET) {
		unsigned short addr=0;
		
		/* parse buffer for word name */
		_doerr(!_fill_buff(backbuffer),36);
		addr=_find_word(backbuffer,TRUE);
		_doerr(!addr,49);
		/* print name of word under SEE */
		_do_see(backbuffer, addr);
	}
	else codeup(_SEE);
}


/* _do_see */
void _do_see(char *backbuffer, unsigned short addr) {
	unsigned short c=0;
	int totalcount=0, i=0, localcount=0;
	int offset=0, exitNow=FALSE;
	
	/* print name of word under SEE */
	fprintf(stdout,"\n: %s",backbuffer);
	CRS;
	offset=0;
	c=codestart(addr);
	if ((mem[c]=='d' || mem[c]=='v') && !mem[c+3]) {
		if (mem[c+4]==addr) exitNow=TRUE;
		else if (c+4==HERE) exitNow=TRUE;
	}
	else if ((mem[c]=='k' || mem[c]=='c') && !mem[c+3]) {
		if (mem[c+4]==addr) exitNow=TRUE;
		else if (c+4==HERE) exitNow=TRUE;
	}
	else if ((mem[c]=='V' || mem[c]=='C') && !mem[c+5]) {
		if (mem[c+6]==addr) exitNow=TRUE;
		else if (c+6==HERE) exitNow=TRUE;
	}
	else if (mem[c]==BWD && !mem[c+2]) {
		if (mem[c+3]==addr) exitNow=TRUE;
		else if (c+3==HERE) exitNow=TRUE;
	}

	/* PROGRAM ANALYSIS */
	/* print compiled program or variables data */
	while (c<HERE) {
		/* shorten output line in case of 'breakers'
		 * (that is structures-closures words) */
		if (totalcount>SCREEN/2 && mem[c+1]!=_RT_THEN
					&& mem[c+1]!=_RT_REPEAT
					&& mem[c+1]!=_RT_UNTIL
					&& mem[c+1]!=_RT_AGAIN
					&& mem[c+1]!=_RT_LOOP
					&& mem[c+1]!=_RT_PLUS_LOOP) { 
			CRS;
			totalcount=localcount=0;
			for (i=0;i<offset;i++) fputc(' ',stdout);
		}
		/* skip DOESMARK */
		if (mem[c]==DOESMARK) {
			c+=2;
			continue;
		}
		/* vocabulary */
		else if (mem[c]==VOC_MARK) {
			fprintf(stdout, "(#%d)", mem[c+1]);
			c+=2;
			continue;
		}
		/* user word */
		else if (mem[c]==UWD) {
			localcount=_print_word_name(mem[c+1]);
		}
		/* user single cell constant */
		else if (mem[c]=='c') {
			if (codestart(addr)==c) {
				fprintf(stdout,"%s","(const) ");
				localcount=8;
				localcount+=_print(mem[c+1]+BYTE*mem[c+2],BASE,NULL);
				c++;
			}
			else localcount=_print_word_name(mem[c+1]);
		}
		/* system single cell constant */
		else if (mem[c]=='k') {
			if (codestart(addr)==c) {
				fprintf(stdout,"%s","(sysconst) ");
				localcount=11;
				localcount+=_print(mem[c+1]+BYTE*mem[c+2],BASE,NULL);
				c++;
			}
			else localcount=_print_word_name(mem[c+1]);
		}
		/* user double cell constant */
		else if (mem[c]=='C') {
			if (codestart(addr)==c) {
				fprintf(stdout,"%s","(2const) ");
				localcount=9;
				localcount+=_print(int2ll(mem[c+1]+BYTE*mem[c+2],mem[c+3]+BYTE*mem[c+4]),BASE,NULL);
				fputc('\b',stdout);
				fputc('.',stdout);
				fputc(' ',stdout);
				c+=3;
			}
			else localcount=_print_word_name(mem[c+1]);
		}
		/* user single cell variable */
		else if (mem[c]=='v') {
			/* seeing inner variable state */
			if (codestart(addr)==c) {
				fprintf(stdout,"%s","(var) ");
				localcount=6;
				localcount+=_print(mem[c+1]+BYTE*mem[c+2],BASE,NULL);
				c++;
			}
			else localcount=_print_word_name(mem[c+1]);
		}
		/* user double cell variable */
		else if (mem[c]=='V') {
			if (codestart(addr)==c) {
				long long int v1, v2;
				fprintf(stdout,"%s","(2var) ");
				v1=mem[c+1]+BYTE*mem[c+2];
				v2=mem[c+3]+BYTE*mem[c+4];
				localcount=7;
				localcount+=_print(int2ll(v1,v2),BASE,NULL);
				fputc('\b',stdout);
				fputc('.',stdout);
				fputc(' ',stdout);
				c+=3;
			}
			else localcount=_print_word_name(mem[c+1]);
		}
		/* system single cell variable */
		else if (mem[c]=='d') {
			if (addr<LOWVAR) {
				fprintf(stdout,"%s","(sysvar) ");
				localcount=9;
				localcount+=_print(mem[c+1]+BYTE*mem[c+2],BASE,NULL);
				c++;
			}
			else localcount=_print_word_name(mem[c+1]);
		}
		/* integer value */
		else if (mem[c]=='I') {
			localcount=_print(mem[c+1]+BYTE*mem[c+2],BASE,NULL);
			c++;
			fputc('\b',stdout);
		}
		/* double value */
		else if (mem[c]=='D') {
			localcount=_print(int2ll(mem[c+1]+BYTE*mem[c+2],mem[c+3]+BYTE*mem[c+4]),BASE,NULL);
			c+=3;
			fputc('\b',stdout);
			fputc('.',stdout);
			fputc(' ',stdout);
		}
		/* print builtin words */
		else if (mem[c]==BWD) {
			/* SHORT CIRCUITS */
			/* break on end of word */
			if (mem[c+1]==_WORD_ENDER) {
				break;
			}
			/* skip noop */
			else if (mem[c+1]==_NOOP) {
				c+=2;
				continue;
			}
			/* PRE-PRINT */
			/* CASE/IF/BEGIN/DO pre print (do nothing) */
			else if (mem[c+1]==_RT_CASE
					|| mem[c+1]==_RT_IF
					|| mem[c+1]==_RT_BEGIN
					|| mem[c+1]==_RT_DO) {
				;
			}
			/* ELSE/WHILE pre print */
			else if (mem[c+1]==_RT_ELSE 
					|| mem[c+1]==_RT_WHILE) {
				offset-=3;
				CRS;
				for (i=0;i<offset;i++) 
					fputc(' ',stdout);
			}
			/* THEN/REPEAT/UNTIL/AGAIN/LOOP/+LOOP 
			 * pre print */
			else if (mem[c+1]==_RT_THEN
					|| mem[c+1]==_RT_REPEAT
					|| mem[c+1]==_RT_UNTIL
					|| mem[c+1]==_RT_AGAIN
					|| mem[c+1]==_RT_LOOP
					|| mem[c+1]==_RT_PLUS_LOOP) {
				offset-=3;
				CRS;
				totalcount=0;
				for (i=0;i<offset;i++) 
					fputc(' ',stdout);
			}
			/* PRINT program in word */
			if (mem[c+2]) 
				localcount+=_print_builtin_name(mem[c+1]);
			/* POST-PRINT */
			/* OF simply advances one totalcount */
			if (mem[c+1]==_RT_OF) {
				c++;
			}
			/* CASE/BEGIN/DO post print*/
			else if (mem[c+1]==_RT_CASE
					|| mem[c+1]==_RT_BEGIN
					|| mem[c+1]==_RT_DO) {
				offset+=3;
				if (offset) totalcount=SCREEN;
			}
			/* IF/ELSE/WHILE post print*/
			else if (mem[c+1]==_RT_IF
					|| mem[c+1]==_RT_ELSE
					|| mem[c+1]==_RT_WHILE) {
				offset+=3;
				if (offset) totalcount=SCREEN;
				c++;
			}
			/* THEN post print*/
			else if (mem[c+1]==_RT_THEN) {
				if (offset) totalcount=SCREEN;
			}
			/* REPEAT/UNTIL/AGAIN/LOOP/+LOOP post print*/
			else if (mem[c+1]==_RT_REPEAT
					|| mem[c+1]==_RT_UNTIL
					|| mem[c+1]==_RT_AGAIN
					|| mem[c+1]==_RT_LOOP
					|| mem[c+1]==_RT_PLUS_LOOP) {
				if (offset) totalcount=SCREEN;
				c++;
			}
			/* ENDCASE post print*/
			else if (mem[c+1]==_RT_ENDCASE) {
				if (offset) totalcount=SCREEN;
				offset-=3;
			}
			/* ENDOF post print*/
			else if (mem[c+1]==_RT_ENDOF) {
				if (offset) totalcount=SCREEN;
				c++;
			}
			/* COMPILE and (COMPILE) */
			else if (mem[c+1]==_DEFERCOMPILE) {
				fputc(' ',stdout); 
				localcount=1;
				localcount+=_print(mem[c+2],BASE,NULL);
				c++;
			}
			/* words with counted strings */
			else if (mem[c+1]==_DOT_QUOTES
					|| mem[c+1]==_PUSH_STRING_ADDR){
				int i=0, len=mem[c+2];
				fputc(' ',stdout); 
				localcount=1;
				if (mem[c+1]==_PUSH_STRING_ADDR) {
					fputc('\"',stdout); 
					localcount++;
				}
				c+=3; // points to string
				while (i<len) {
					localcount++;
					fprintf(stdout,"%c",mem[c+i++]);
				}
				fputc('\"',stdout); 
				localcount++;
				fprintf(stdout,"%s","  "); 
				localcount+=2;
				c=c+len-2;
			}
			if (!mem[c+2]) 
				fprintf(stdout,"%s","(built-in) ");
		}
		/* print isolated units 
		 * Note: check that the value is not the final 
		 * zero that separates two words 
		 * NOTE: this should result in an error in compilation! */
		else if (mem[c+1]!=addr && HERE!=c+1) {
			int cc=mem[c];
			if (cc>=BYTE) cc=BYTE-1;
			localcount=_print(cc,BASE,NULL);
			if (cc>32) {
				fprintf(stdout,"%s {%c} %s",REDCOLOR,(char)cc,RESETCOLOR);
				localcount+=6;
			}
			c--;
		}
		
		/* update counters and print a separator blank */
		c+=2;
		totalcount+=localcount+1;
		fputc(' ',stdout);
		
		/* in case of variables and built-ins, there is no
		 * _WORD_ENDER, so finish it here */
		if (exitNow) break;
	}

	/* setup for closure */
	if (totalcount) fputc('\n',stdout);
	else NOCRS;
	fputc(';',stdout);
	if (isComp(addr)) fprintf(stdout,"%s"," compile-only");
	if (isExec(addr)) fprintf(stdout,"%s"," exec-only");
	if (isImm(addr)) fprintf(stdout,"%s"," immediate");
	fputc('\n',stdout);
}


/* built-in ' (tick) 
 * Note: while in ANSI 1994 the ' (tick) returns the execution token,
 * in Forth-79 it returns the parameters field address of the word that
 * follows.  */
void _tick(unsigned short addr) {
	unsigned short val=0;
	if (STATE==INTERPRET) {
		_doerr(!_fill_buff(backbuffer),36);
		addr=_find_word(backbuffer,TRUE);
		_doerr(!addr,49);
		/* in case of a CREATEd word, skip to PFA */
		val=codestart(addr);
		if (mem[val]=='v' || mem[val]=='V' || mem[val]=='d' ||
				mem[val]=='k' || mem[val]=='c' || mem[val]=='C')
			val++;
		else if (mem[val]==BWD && mem[val+1]==_NOOP)
			val=mem[val+3]+BYTE*mem[val+4];
		_push(val);
	}
	/* ' (tick) does not compile itself but the PFA address 
	 * of the word that follows */
	else if (STATE==COMPILE) {
		_doerr(!_fill_buff(backbuffer),36);
		addr=_find_word(backbuffer,TRUE);
		_doerr(!addr,49);
		/* compile literal */
		val=codestart(addr);
		if (mem[val]=='v' || mem[val]=='V' || mem[val]=='d' ||
				mem[val]=='k' || mem[val]=='c' || mem[val]=='C')
			val++;
		else if (mem[val]==BWD && mem[val+1]==_NOOP)
			val=mem[val+3]+BYTE*mem[val+4];
		mem[HERE++]='I';
		mem[HERE++]=(int)(val%BYTE);
		mem[HERE++]=val/BYTE;
	}
}


/* built-in FIND 
 * in Forth-79 it uses the text that follows, just like the tick.
 * FIND returns the CFA or zero if word was not found */
void _find(void) {
	if (STATE==INTERPRET) {
		unsigned short addr=0;
		_doerr(!_fill_buff(backbuffer),36);
		addr=_find_word(backbuffer,TRUE);
		if (!addr) _push(0);
		else _push(codestart(addr));
	}
	else codeup(_FIND); 
}


/* built-in EXECUTE */
void _execute(void) {
	if (STATE==INTERPRET) {
		unsigned short addr;
		unsigned short oldPC=_PC;
		addr=_pop();
		_doerr(!addr,20);
		isExecute=TRUE;
		_exec_flow(addr);
		isExecute=FALSE;
		PC_(oldPC);

	}
	else codeup(_EXECUTE);
}


/* built-in COUNT */
void _count(void) {
	if (STATE==INTERPRET) {
		unsigned short addr;
		short len=0;
		addr=_pop();
		len=mem[addr];
		_push(addr+1);
		_push(len);
	}
	else codeup(_COUNT);
}


/* built-in TYPE */
void _type(void) {
	if (STATE==INTERPRET) {
		unsigned int i=0;
		unsigned short addr;
		short len;
		len=_pop();
		addr=_pop();
		/* This is required by the standard */
		if (len<=0) return;
		/* checks */
		_doerr(len>=ILINE,24);
		_doerr(!addr,20);
		_doerr((addr+len>=ADDRLIMIT),41);
		while(i<len && mem[addr+i]) {
			fputc(mem[addr+i],stdout);
			i++;
			OUT++;
		}
	}
	else codeup(_TYPE);
}


/* built-in SPACE/SPACES
 * if arg is FALSE perform SPACES and pop the number of spaces
 * else perform SPACE */
void _spaces(int arg) {
	if (STATE==INTERPRET) {
		short ch=0;
		if (!arg) {
			ch=_pop();
		}
		else ch=1;
		while (ch>0) {
			fputc(' ',stdout);
			OUT++;
			ch--;
		}
	}
	else {
		if (arg) codeup(_SPACE);
		else codeup(_SPACES);
	}
}


/* built-in PAD */
void _pad(void) {
	if (STATE==INTERPRET) {
		/* PAD is always PADWIDTH cells after HERE, 
		 * to prevent immediate overwriting */
		_doerr((HERE+PADWIDTH+padlen>=HIGHMEM),30);
		_push(HERE+PADWIDTH);
	}
	else codeup(_PAD);
}


/* built-in PAGE 
 * Note: [2J clears the entire screen (arg.=2), and the [H command 
 * reposition the cursor in the upper left corner of the screen. 
 * Note: the  character is the ESC code (CTRL-C + ESC in vim/gvim) */ 
void _page(void) {
	if (STATE==INTERPRET) {
		fprintf(stdout,"%s","[2J");	// clear screen
		fprintf(stdout,"%s"," [H");	// reposition cursor
		isPage=TRUE;
		OUT=0;
	}
	else codeup(_PAGE);
}

void _qsound() {
	if (STATE==INTERPRET) {
		unsigned short freq=0, dur=0;
		double rdur=0.0;
		char comm[ILINE+1];
		freq=_pop();
		dur=_pop();
		rdur=(double)dur/1000.;
		SOUND;
		qr=system(comm);
	}
	else codeup(_SOUND);
}

/* built-in BELL */
void _bell(void) {
	if (STATE==INTERPRET) {
		SPEAKER;
	}
	else codeup(_BELL);
}


/* built-in EMIT */
void _emit(void) {
	if (STATE==INTERPRET) {
		unsigned short c;
		c=_pop();
		c=c&FFCHAR;
		fputc(c,stdout);
		OUT++;
	}
	else codeup(_EMIT);
}


/* _set_dump()
 * set parameters for DUMP based on BASE */
void _set_dump(int *al, int *sl, int *step) {
	switch(BASE) {
		case 2:	*al=16, *sl=32, *step=1; break;
		case 3: *al=11, *sl=21, *step=2; break;
		case 4: *al=8,  *sl=16, *step=3; break;
		case 5: *al=7,  *sl=14, *step=4; break;
		case 6: *al=7,  *sl=13, *step=4; break;
		case 7: *al=6,  *sl=12, *step=5; break;
		case 8: 
		case 9:	*al=6,  *sl=11, *step=5; break;
		case 10:*al=5,  *sl=10, *step=5; break;
		case 11:*al=5,  *sl=10, *step=6; break;
		case 12:
		case 13:
		case 14: 
		case 15:*al=5,  *sl=9,  *step=6; break;
		case 16:
		case 17:
		case 18:
		case 19:
		case 20: 
		case 21:
		case 22:
		case 23:*al=4,  *sl=8,  *step=7; break;
		case 24:
		case 25:
		case 26:
		case 27:
		case 28:
		case 29:
		case 30:
		case 31:
		case 32:
		case 33:
		case 34:
		case 35:
		case 36:*al=4,  *sl=7,  *step=8; break;
		default:*al=4,  *sl=8,  *step=7; break;
	}
}


/* _out_dump() 
 * print number val according to BASE, with leading zeroes
 * sl = string length
 * sign = if TRUE, the value in sl is signed
 * rev = if TRUE, print in reverse
 * System routine for DUMP */
void _out_dump(int val, int strl, int sign, int rev) {
	char output[ILINE+1];
	char temp[ILINE+1];
	int i, s=1;
	if (strl>=ILINE-1) strl=ILINE-1;
	
	/* erase output */
	i=0; while (i<ILINE-1) output[i++]='0';
	output[i]='\0';

	/* fill with zeroes */
	output[strl+sign]='\0';
	
	/* print number according to BASE */
	if (sign && val<0) val*=-1, s=-1;
	if (sign) _print(val,BASE,temp);
	else _print_double_unsigned(val,BASE,temp);
	i=strlen(output)-strlen(temp);
	if (i<0) i=0; 
	strcpy(output+i,temp);
	if (sign) {
		if (s<0) output[0]='-';
		else output[0]=' ';
	}
	if (rev) fprintf(stdout,"%s%s%s",INVCOLOR,output,RESETCOLOR);
	else fprintf(stdout," %s",output);
}


/* built-in DUMP */
void _dump(void) {
	if (STATE==INTERPRET) {
		int n1, addr;
		int addrf, i, sign=FALSE;
		int addrlen, steplen, step;
		n1=_pop();
		n1=n1-1;
		addr=_pop();
		if (addr<0) addr+=ADDRLIMIT;
		_doerr((addr>ADDRLIMIT-1),20);
		addrf=addr+n1;
		if (addrf>ADDRLIMIT-1) addrf=ADDRLIMIT-1;

		/* get DUMP data based on BASE */
		_set_dump(&addrlen, &steplen, &step);
		if (BASE==10) sign=TRUE;

		/* DUMP! */
		putchar('\n');
		i=0;
		_out_dump(addr,addrlen,FALSE,TRUE);
		while (addr<=addrf) {
			char text[ILINE+1];
			while (i<step) {
				if (addr+i<=addrf) {
					_out_dump(mem[addr+i],steplen,sign,FALSE);
					text[i]=_dottize(mem[addr+i]);
					text[i+1]='\0';
				}
				else {
					int k=0;
					putchar(' ');
					if (sign) putchar(' ');
					while (k<steplen) putchar(' '),k++;
				}
				i++;
			}
			printf(" %s\n",text);
			addr+=step;
			if (addr>addrf) break; 
			_out_dump(addr,addrlen,FALSE,TRUE);
			i=0;
		}
		putchar('\n');
	}
	else codeup(_DUMP);
}


/* built-in HERE */
void _here(void) {
	if (STATE==INTERPRET) {
		_push(HERE);
	}
	else codeup(_HERE);
}


/* built-in colon */
void _colon(void) {
	if (STATE==INTERPRET) {
		/* parse buffer for name */
		currentDef=HERE;
		backcfs=cfs; // record current status
		_doerr(!_fill_buff(backbuffer),37);
		LAST_(HERE);
		/* save name for later (serves to semicolon) */
		strcpy(compiling,backbuffer);
		/* check if word name already exists */
		if (_find_word(compiling,TRUE)) isExists=TRUE;
		/* set CONTEXT to CURRENT (pag. 19 of Standard) */
		CONTEXT_(CURRENT);
		/* compile link address, MARKVAR and name */
		mem[HERE++]=PREVHERE;
		mem[HERE++]=B_SMUDGE; // control bits
		mem[HERE++]=CURRENT;
		mem[HERE++]=MARKVAR;
		mem[HERE++]=strlen(compiling);
		b=compiling;
		while (*b) mem[HERE++]=*b++;
		/* set state */
		STATE=COMPILE; 
	}
	else codeup(_COLON);
}


/* built-in semicolon */
void _semicolon(void) {
	/* close compiling word */
	mem[HERE++]=BWD;
	mem[HERE++]=_WORD_ENDER;
	mem[HERE++]=0;
	PREVHERE=currentDef;
	mem[currentDef+1]^=B_SMUDGE;
	currentDef=0;
	last_word_with_success=TRUE;
	if (isExists) {
		if (!isFile && !isBlock && !isLoad) 
			fprintf(stdout," %s isn't unique.",compiling);
	}
	isExists=FALSE;
	STATE=INTERPRET;
	doDoes=FALSE;
	WCOUNT++;
	if (backcfs!=cfs) fprintf(stdout,"\n%s\n",error_list[4]);
}


/* _push_string_addr
 * used to push the string address and count associated to the address
 * Note: DON'T REMOVE! This is not associated to any direct command but
 * serves to SLITERAL and the like*/ 
void _push_string_addr(unsigned short addr) {
	unsigned short len;
	addr++; 
	len=mem[addr];
	_push(addr+1);
	_push(len);
	PC_(addr+(len+1-2));
}


/* built-in DOES> */
void _does_to(void) {
	/* execute normal immediate procedure */
	if (!doDoes) {
		int i=currentDef;
		mem[HERE++]=BWD;
		mem[HERE++]=_PUSH_DOES;
		doDoes=TRUE;
		/* search for DOESMARK looking backwards, to put DOES> addr 
		 * (which is HERE) in the address that follows.
		 * Note: to let use the CREATE DOES> CREATE DOES> sequence
		 * the process stop if another DOES> was already compiled. */
		while (i<HERE) {
			if (mem[i]==BWD && mem[i+1]==_PUSH_DOES) break;
			if (mem[i]==DOESMARK) mem[i+1]=HERE;
			i++;
		}
	}
	/* in case DOES> belongs to another previous DOES>, it does not
	 * compile the address, but replicates the _DOES_TO code to let
	 * the system interpret the correct DOES> when due 
	 * (typically on defining words that are defining words.
	 * See "Forth Dimensions" vol. IV n. 1-2-3, "Defining Words",
	 * parts I, II and III by Henry Laxen) */
	else {
		mem[HERE++]=BWD;
		mem[HERE++]=_DOES_TO;
	}
}


/* _get_does_addr()
 * Search forward for DOESMARK and check address that follows.
 * If not null, return the address, otherwise get to the origin.
 * System function for create() */
unsigned short _get_does_addr(unsigned short addr)  {
	while (mem[addr]!=DOESMARK && addr<HERE) addr++;
	if (addr>HERE) return 0;
	if (mem[addr+1]) return mem[addr+1];
	if (!mem[addr+1] && mem[addr-1]==_CREATE) return 0;
	if (!mem[addr+1] && !mem[addr-1] && !mem[addr+1]) return 0;
	return _get_does_addr(mem[addr-1]);
}


/* built-in CREATE */
void _create(unsigned short addr) {
	if (STATE==INTERPRET) {
		int ph=HERE, pos=0, tpos=0;
		char *pw=backbuffer;
		int c=0, i=0, e=0, d=0;
		/* get word name */
		_doerr(!_fill_buff(backbuffer),36);
		/* check if redefined */
		if (_find_word(backbuffer,TRUE)) {
			if (!isFile && !isBlock && !isLoad) 
				fprintf(stdout," %s isn't unique.",backbuffer);
		}
		/* start definition of target word */
		LAST_(HERE);
		currentDef=HERE;
		mem[HERE++]=PREVHERE; tpos=HERE; // store for more
		mem[HERE++]=B_NULL; 
		mem[HERE++]=CURRENT; // voc code
		mem[HERE++]=MARKVAR;
		mem[HERE++]=strlen(backbuffer);
		/* copy name */
		while (*pw) mem[HERE++]=*pw++;
		/* copy default code */
		mem[HERE++]=BWD;
		mem[HERE++]=_NOOP;
		mem[HERE++]='I'; // this will push address naturally
		pos=HERE;
		mem[HERE++]=0;  // low
		mem[HERE++]=0;  // high
		/* compile DOES> code if found 
		 * AM - November 2017
		 * Note: this process passes through many phases: the scope
		 * is finding the i value (the address where to copy the DOES>
		 * istructions of the defining word or of its direct
		 * ancestor). Steps: */
		/* 1. find in c the address of the defining word under 
		 * execution and get in e its proper compiled DOES> address 
		 * (which may be zero or non-zero) */
		d=0; 
		c=_cfcopy();	// address of defining word under execution
		e=mem[c+3]; 	// content of DOES> address
		/* 2a. if currPC is lower than LOWMEM, is a direct call from
		 * console, so set i value to null because no DOES> can be
		 * attached in this case - AM 27 December 2017 */
		if (currPC<LOWMEM) i=0;
		/* 2b. if c is -1, it's the direct case of a CREATE, with or
		 * without DOES>, so use currPC as the original seed to 
		 * retrieve the i value */
		else if (c<0) i=_get_does_addr(currPC);
		/* 3. proceed with the other cases to determine i */
		else {
			/* 3a. in the current defining word, find MARKVAR; the
			 * scope is counting the defining words in pair 
			 * positions (e.g. CREATE..CREATE..[DOES>]) and not in
			 * pair couples (e.g CREATE..DOES>..CREATE..DOES>) and
			 * we need to find the start of the definition */ 
			while (c>0 && mem[c]!=MARKVAR) c--;
			/* 3b. count in d the CREATE or defining words in the
			 * definitions */
			while (!(mem[c]==BWD&& mem[c+1]==_WORD_ENDER) &&c<HERE){
				/* stop if a DOES> was already compiled */
				if (mem[c]==BWD && mem[c+1]==_PUSH_DOES) break;
				/* add a CREATE */
				if (mem[c]==BWD && mem[c+1]==_CREATE) d++;
				/* add a defining word */
				else if (mem[c]==UWD && isDef(mem[c+1])) d++;
				c++;
			}
			/* 3c. if there are more than one defining word and
			 * the original DOES> address is null, use addr
			 * as the original seed */
			if (d>1 && !e) i=_get_does_addr(addr);
			/* 3d. in all other remaining cases, currPC is good
			 * as the original seed */
			else i=_get_does_addr(currPC);
		}
		/* 4. i the DOES> address returned is not zero (that is a 
		 * DOES> address was found at least in one of the available 
		 * ancestors) copy all DOES> instructions */
		if (i) {
			int doneOne=FALSE;
			while (!(mem[i]==BWD && mem[i+1]==_WORD_ENDER)) {
				/* copy all DOES> section but, if an uncompiled
				 * DOES> is found, force a compile, but only 
				 * for the first item in the list; the rest
				 * is supposedly a series of DOES> of other
				 * CREATEs or other defining words. */
				if (!doneOne && mem[i]==BWD &&\
					       mem[i+1]==_DOES_TO) {
					int b=doDoes;
					doDoes=FALSE;
					_does_to(); // DOES> compile itself
					i+=2;
					doDoes=b;
					doneOne=TRUE;
				}
				/* set a defining word in case DOES>
				 * instructions contain a CREATE */
				else if (mem[i]==BWD && mem[i+1]==_CREATE) {
					mem[HERE++]=mem[i++];
					mem[tpos]|=B_DEFINING;
				}
				/* store the DOES> instructions */
				else {
					mem[HERE++]=mem[i++];
				}
			}
		}
		/* compile regular word ender */
		mem[HERE++]=BWD;
		mem[HERE++]=_WORD_ENDER;
		/* compile return address of CREATEd word */
		mem[pos]=(HERE+1)%BYTE;
		mem[pos+1]=(HERE+1)/BYTE;
		mem[HERE++]=0; // terminator or address returned
		PREVHERE=ph;
		LAST_(PREVHERE);
		WCOUNT++;
	}
	else {
		/* The same procedure here written is also added to _exec() 
		 * in case of words which have the B_DEFINING bit set 
		 * (that is all those which inherit the CREATE property) */
		mem[HERE++]=BWD;
		mem[HERE++]=_CREATE;
		mem[HERE++]=DOESMARK;
		mem[HERE++]=0;
		mem[currentDef+1]|=B_DEFINING;
	}
}


/* built-in ORDER (not Forth 79) */
void _order(void) {
	if (STATE==INTERPRET) {
		int i=vocCount, val;
		/* print CONTEXT vocabularies */
		while (i>=0) {
			val=(int)pow(2,i);
			if (vocChain[(CONTEXT)] & val)
				fprintf(stdout,"%s ",vocs[i].name);
			i--;
		}
		/* print CURRENT vocabulary */
		fprintf(stdout,"  %s",vocs[(CURRENT)].name);
	}
	else codeup(_ORDER);
}


/* built-in FORTH */
void _forth(void) {
	CONTEXT_(VFORTH);
}


/* built-in EDITOR */
void _editor(void) {
	CONTEXT_(VEDITOR);
}


/* built-in DEFINITIONS */
void _definitions(void) {
	if (STATE==INTERPRET) {
		CURRENT_(CONTEXT);
	}
	else codeup(_DEFINITIONS);
}


/* built-in FORGET
 * Note: while in Fig-forth it was required that CONTEXT and CURRENT should
 * be equal, for the FORGET command to work, the Forth-79 Standard does not
 * require this, but only that the word can be found in the CURRENT
 * vocabulary */
void _forget(void) {
	if (STATE==INTERPRET) {
		int i, h=0, HIGH=HERE;
		unsigned short fence=FENCE1+BYTE*FENCE;
		/* parse buffer for word name */
		_doerr(!_fill_buff(backbuffer),36);
		/* search word */
		h=PREVHERE;
		while (h>0) {
			if (!isSmudge(h) && voccode(h)==(CURRENT)) {
				if (wordlen(h)==strlen(backbuffer) && 
					 _compare(wordstart(h),wordlen(h),backbuffer)){ 
					break;
				}
			}
			h=prev(h);
		}
		/* check limits */
		if (h<=0) return;
		_doerr(h<LOWMEM,12);
		_doerr((fence && h<fence),50);
		/* rebuild the vocabularies count by removing the vocabularies
		 * created after fence */
		i=h;
		while (i<HIGH) {
			if (mem[i]==VOC_MARK) NVOCS=NVOCS-1;
			i++;
		}
		/* if found, delete this word and all following words 
		 * regardless of their vocabulary, as required 
		 * by the Standard */
		HERE=h;
		PREVHERE=prev(HERE);
		LAST_(PREVHERE);
		i=HIGH;
		while (i>=HERE) {
			mem[i]=0;
			i--;
		}
	}
	else codeup(_FORGET);
}


/* _set_voc_number
 * used by vocabulary words in INTEPRET state 
 * to set as CONTEXT the proper number */ 
void _set_voc_number(unsigned short addr) {
	CONTEXT_(mem[addr+2]);
}


/* built-in VOCABULARY */
void _vocabulary(unsigned short addr) {
	if (STATE==INTERPRET) {
		int where=HERE, ph=0;
		char *pw=backbuffer;
		/* get voc name */
		_doerr(!_fill_buff(backbuffer),36);
		/* setup vocabulary data VOCS = 32 by default */
		/* last vocabulary = 31 is not used, so if current 
		 * vocCount is 29, the new vocabulary can be set to 30, 
		 * but if it is 30, the new vocabulary can't be set to 31 */
		_doerr((vocCount>=VOCS-2),29);
		LAST_(HERE);
		vocCount++;
		strncpy(vocs[vocCount].name,backbuffer,WORDNAMELEN+1);
		vocChain[vocCount]=vocChain[(CONTEXT)]+pow(2,vocCount);
		/* start definition of target word */
		ph=where;
		mem[where++]=PREVHERE;
		mem[where++]=B_NULL; 
		mem[where++]=CURRENT; // voc code
		mem[where++]=MARKVAR;
		mem[where++]=strlen(backbuffer);
		/* copy name */
		while (*pw) mem[where++]=*pw++;
		/* copy default code */
		mem[where++]=BWD;
		mem[where++]=_SET_VOC_NUM;
		mem[where++]=VOC_MARK;
		mem[where++]=vocCount;
		mem[where++]=BWD;
		mem[where++]=_WORD_ENDER;
		mem[where++]=0;
		HERE=where;
		PREVHERE=ph;
		LAST_(PREVHERE);
		NVOCS=NVOCS+1;
		WCOUNT++;
		while(isBlank(line_buffer(to_in))) _inc_to_in_();
	}
	else codeup(_VOCABULARY);
}


/* built-in EXIT 
 * Note: I cannot use _exit() here because it's already defined in unistd.h 
 * addr is currently unused */
void __exit(unsigned short addr) {
	if (STATE==COMPILE) {
		mem[HERE++]=BWD;
		mem[HERE++]=_EXIT;
	}
	else _doerr(loops>0,5);
}


/* built-in MESSAGE */
void _message(void) {
	if (STATE==INTERPRET) {
		int val=_pop();
		/* if WARNING == 1, use current screen if loaded */
		if (WARNING==1 && BFR) {
			int i=0;
			_doerr((val<0 || val>LPS),23);
			while (i<CPL) { 
				fputc(mem[_buffer_data[BFR].start+CPL*val+i],stdout);
				i++;
			}
		}
		/* if WARNING == 0 or -1 (or in case no screen is loaded)
		 * print system message */
		else {
			/* Note: this is a real error condition! */
			_doerr((val<0 || val>ERROR_NUM),23);
			fprintf(stdout,"%s",error_list[val]);
		}
	}
	else codeup(_MESSAGE);
}


/* built-in ERROR */
void _error(void) {
	if (STATE==INTERPRET) {
		short val=_pop();
		/* update ERR */
		ERR1=val%BYTE;
		ERR=val/BYTE;
		/* if WARNING == -1, call (ABORT) 
		 * Note: since this is an error condition, the (ABORT)
		 * instruction is followed by a restore procedure
		 * the reset process in _perror() */
		if (WARNING==-1) {
			int h=_find_word("(ABORT)",TRUE);
			_exec(h);
			_perror(-1,-1);
		}
		/* if WARNING == 1, use current screen if loaded */
		else if (WARNING==1) {
			int i=0;
			/* check value */
			if (val<0 || val>LPS) {
				ERR1=-1%BYTE;
				ERR=-1/BYTE;
				_perror(23,to_in);
				return;
			}
			/* check screen existence */
			if (BFR<1 || BFR>3) {
				ERR1=-1%BYTE;
				ERR=-1/BYTE;
				_perror(21,to_in);
				return;
			}
			/* check screen in use */
			if (!_buffer_data[BFR].in_use ||\
					!_buffer_data[BFR].number) {
				ERR1=-1%BYTE;
				ERR=-1/BYTE;
				_perror(21,to_in);
				return;
			}
			fputc('\n',stdout);
			while (i<CPL) { 
				fputc(mem[_buffer_data[BFR].start+CPL*val+i],stdout);
				i++;
			}
			fputc('\n',stdout);
			_perror(-1,-1);
		}
		/* if WARNING == 0 (or in case no screen is loaded)
		 * call _perror() */
		else {
			/* Note: this is a real error condition! */
			if (val<0 || val>ERROR_NUM) {
				ERR1=-1%BYTE;
				ERR=-1/BYTE;
				_perror(23,to_in);
				return;
			}
			_perror(val,to_in);
		}
	}
	else codeup(_ERROR);
}


/* built-in ?ERROR */
void _question_error(void) {
	if (STATE==INTERPRET) {
		int f=_pop();
		if (f) _error();
		else _drop();
	}
	else codeup(_QUESTION_ERROR);
}


/* _perror()
 * print error and set flags.
 * code is the error code of the string to print
 * pos is the position of the error in the input buffer: if pos is null, 
 * position printing is disabled */
void _perror(int code, int pos) {
	int i=0;
	/* flush current output */
	fflush(stdout);
	/* avoid errors cascade */
	if (RunningErrorInstance) return;
	/* default error management */
	if (code>=0) fprintf(stderr,"\n? %s\n",error_list[code]);
	if (pos>=0 && !isInterrupt) {
		/* procedure out of LOAD */
		if (!isLoad) {
			int lim=0;
			/* writing input buffer for the printing of position
			 * of the arrow indicating pos */
			while (i<B_BUF && line_buffer(i)) 
				fputc(line_buffer(i++),stderr);
			fputc('\n',stderr);
			/* get back to start of incriminated word */
			i=pos;
			while (i>=0 && isBlank(line_buffer(--i)));
			/* print indicator ^ */
			lim=i-strlen(word)+1; 
			if (lim<0) lim=-1;
			i=0;
			while (i<=lim) {
				fputc(' ',stderr);
			       i++;
			}	       
			fprintf(stderr,"%s","^\n");
		}
		/* procedure for LOAD */
		else{
			int i=0, lim=0;
			int startpos=(pos/CPL)*CPL;
			i=startpos;
			/* writing input buffer 
			 * (only the line where error occurred) */
			while (i<startpos+CPL && line_buffer(i)) 
				fputc(line_buffer(i++),stderr);
			fputc('\n',stderr);
			/* print indicator ^ */
			lim=pos-startpos-strlen(word)+1; 
			if (lim<0) lim=-1;
			i=0;
			while (i<=lim) {
				fputc(' ',stderr);
			       i++;
			}	       
			fprintf(stderr,"%s","^\n");
		}

	}
	/* reset operations */
	if (STATE==COMPILE) {
		HERE=currentDef;
		PREVHERE=prev(HERE);
		isExists=FALSE;
		last_word_with_success=FALSE;
	}
	STATE=INTERPRET; // better leave it here...
	LAST_(PREVHERE); // this is necessary...
	_erase_stacks(ALL);
	/* erase input buffer */
	i=0; // to_in;
	while (i<B_BUF) line_buffer(i++)='\0';
	isQuit=TRUE;
	loops=0;
	isError=TRUE;
	/* set BYE in case a file was under compilation */
	if (exitOnFile) isBye=isFile;
	RunningErrorInstance=TRUE;
}

/* SYSTEM FUNCTIONS (not available to user) */

/* wrapper for _quit, to be called by sigint() 
 * Note: at present it merely calls _quit. */
void qquit(int sig) {
	_quit();
}



/* _dottize()
 * printing purposes for DUMP: return a dot for each non-printable char 
 * or return the char as a 7-bit item */
char _dottize(int val) {
	if (val<BL || val>ESCL) return '.';
	else return (char) val;
}


/* _isdigit()
 * checks if the digit is into the range of current base */
int _isdigit(char w) {
	if (w>=digits[0] && w<=digits[BASE-1]) return TRUE;
	else return FALSE;
}


/* print_word_name()
 * print word name, given its starting LFA address
 * returns number of printed characters */
int _print_word_name(unsigned short h) {
	register unsigned int i;
	/* print word's name using the proper macros */
	for (i=0;i<wordlen(h) && i<WORDNAMELEN;i++) 
		fputc(mem[wordstart(h)+i],stdout);
	return i;
}


/* print_builtin_name()
 * print built-in word name, searching it in the range 0-LAST_BUILTIN
 * returns number of printed characters */
int _print_builtin_name(unsigned short code) {
	int k,h=LAST_BUILTIN;
	while (h>0) {
		k=codestart(h);
		if (mem[k]==BWD && mem[k+1]==code) 
			return _print_word_name(h);
		h=prev(h);
	}
	return 0;
}


/* _print()
 * print val in any base 'base', val being at most a signed double precision 
 * number; base is confined between 2 and BASELIMIT. 
 * return length of printed string 
 * dest, if not NULL, is the destination string (suitable for D.R and .R) 
 * Note: ival must be a long long int because it must deal the special case
 * of 2147483648, which is the negative int -2147483648 */
int _print(int val, int base, char *dest) {
	int s=1,c=0;
	long long int ival=val;
	char res[ILINE+1];
	char output[ILINE+1], *o=output;
	if (base < BASELOW) base=BASELOW;
	if (base > BASELIMIT) base=BASELIMIT;
	if (ival<0) s=-1, ival=-ival;
	if (!ival) res[c++]='0';
	/* print in any base building a reverse string and printing it then */
	else while (ival>0) res[c++]=digits[ival % base], ival = ival/base;
	res[c]='\0';
	if (s==-1) *o++='-';
	for (c=strlen(res)-1;c>=0;c--) *o++=toupper(res[c]);
	if (!dest) *o++=' '; // D.R does not require a blank after 
	*o='\0';
	if (!dest) fprintf(stdout,"%s",output);
	else strcpy(dest,output); 
	return strlen(output);
}


/* _print_unsigned()
 * print val in any base 'base', val being an unsigned number;  
 * base is confined between 2 and BASELIMIT. 
 * return length of printed string 
 * dest, if not NULL, is the destination string (suitable for U.R) */
int _print_unsigned(int val, int base, char *dest) {
	unsigned short uval;
	char res[ILINE+1];
	char str[ILINE+1], *s=str;
	int c=0;
	uval=val;
	if (base < BASELOW) base=BASELOW;
	if (base > BASELIMIT) base=BASELIMIT;
	/* treat the case of uval=0 */
	if (!uval) res[c++]='0';
	/* print in any base building a reverse string and printing it then */
	else while (uval>0) res[c++]=digits[uval % base], uval = uval/base;
	res[c]='\0';
	for (c=strlen(res)-1;c>=0;c--) *s++=toupper(res[c]);
	if (!dest) *s++=' ';
	*s='\0';
	if (!dest) fprintf(stdout,"%s",str);
	else strcpy(dest,str);
	return strlen(str);
}


/* _print_double_unsigned()
 * print val in any base 'base', val being a double unsigned number;  
 * base is confined between 2 and BASELIMIT. 
 * return length of printed string 
 * dest, if not NULL, is the destination string */
int _print_double_unsigned(int val, int base, char *dest) {
	unsigned int uval;
	char res[ILINE+1];
	char str[ILINE+1], *s=str;
	int c=0;
	uval=val;
	if (base < BASELOW) base=BASELOW;
	if (base > BASELIMIT) base=BASELIMIT;
	/* treat the case of uval=0 */
	if (!uval) res[c++]='0';
	/* print in any base building a reverse string and printing it then */
	else while (uval>0) res[c++]=digits[uval % base], uval = uval/base;
	res[c]='\0';
	for (c=strlen(res)-1;c>=0;c--) *s++=toupper(res[c]);
	if (!dest) *s++=' ';
	*s='\0';
	if (!dest) fprintf(stdout,"%s",str);
	else strcpy(dest,str);
	return strlen(str);
}


/* _push()
 * push argument arg on two cell on tos */
void _push(int arg) {
	if (ds>DATASTACKLIMIT) {
		_perror(45,to_in);
		return;
	}
	arg=(short)arg;
	mem[ds++]=(arg%BYTE);
	mem[ds++]=arg/BYTE;
}


/* _POP() 
 * pop value from two cells of tos */
int _POP(void) {
	if (ds<=DATASTACKSTART+1) {
		_perror(44,to_in);
		return 0;
	}
	ds-=2;
	return mem[ds]+BYTE*mem[ds+1];
}


/* _rspush()
 * push argument arg on return stack tos */
void _rspush(int arg) {
	if (rs>RETURNSTACKLIMIT) {
		_perror(43,to_in);
		return;
	}
	arg=(short)arg;
	mem[rs++]=(arg%BYTE);
	mem[rs++]=arg/BYTE;
}


/* _RSPOP() 
 * pop value from return stack tos */
int _RSPOP(void) {
	if (rs<=RETURNSTACKSTART+1) {
		_perror(42,to_in);
		return 0;
	}
	rs-=2;
	return mem[rs]+BYTE*mem[rs+1];
}


/* _cfpush()
 * push argument arg on tos of control flow stack */
void _cfpush(int arg) {
	if (cfs>CFSTACKLIMIT) {
		_perror(17,to_in);
		return;
	}
	cfmem[cfs++]=arg;
}


/* _cfpop() 
 * pop value from control flow tos */
int _CFPOP(void) {
	if (cfs<=CFSTACKSTART) {
		_perror(16,to_in);
		return 0;
	}
	return cfmem[--cfs];
}


/* built-in CF-GET*/
void _cf_get(void) {
	if (STATE==INTERPRET) {
		int val=_cfpop();
		_push(val);
	}
	else codeup(_CF_GET);
}


/* built-in CF-PUT */
void _cf_put(void) {
	if (STATE==INTERPRET) {
		int val=_pop();
		_cfpush(val);
	}
	else codeup(_CF_PUT);
}


/* _cfcopy() 
 * copy value from control flow tos. No error if stack is empty */
int _cfcopy(void) {
	if (cfs<=CFSTACKSTART) {
		return -1;
	}
	return cfmem[cfs-1];
}


/* int2ll()
 * convert the integer pair high and low to a double precision number */
long long int int2ll(long long int high, long long int low) {
	if (low<0) low+=DOUBLE;
	return (high*DOUBLE+low);
}


/* ll2int()
 * convert a double precision number into its integer pair high and low */
void ll2int(long long int num, long long int* high, long long int* low) {
	*high=num/DOUBLE;
	*low=num-*high*DOUBLE;
	if (*low<0) *high-=1;
}


/* _noop()
 * dummy function to perform nothing */
void _noop(void) {
	;
}


/* _79_standard() */
void _79_standard(void) {
	if (STATE==INTERPRET) {
		isFigCompatEnabled=FALSE;
	}
	else codeup(_79_STANDARD);
}


/* _erase_stacks 
 * if flag=ALL clear all stacks 
 * if flag=DATA_STACK, clear return stack only
 * if flag=RETURN_STACK, clear return stack only
 * if flag=any other value, clear only control flow stack */
void _erase_stacks(int flag) {
	int i;
	if (flag==ALL) {
		for (i=DATASTACKSTART; i<RETURNSTACKLIMIT; i++) {
			mem[i]=0;
		}
		for (i=CFSTACKSTART; i<CFSTACKLIMIT; i++) {
			cfmem[i]=0;
		}
		ds=DATASTACKSTART;
		rs=RETURNSTACKSTART;
		cfs=CFSTACKSTART;
	}
	else if (flag==DATA_STACK) {
		for (i=DATASTACKSTART; i<DATASTACKLIMIT; i++) {
			mem[i]=0;
		}
		ds=DATASTACKSTART;
	}
	else if (flag==RETURN_STACK) {
		for (i=RETURNSTACKSTART; i<RETURNSTACKLIMIT; i++) {
			mem[i]=0;
		}
		rs=RETURNSTACKSTART;
	}
	else  {
		for (i=CFSTACKSTART; i<CFSTACKLIMIT; i++) {
			cfmem[i]=0;
		}
		cfs=CFSTACKSTART;
	}
}


/* _add()
 * _add all built-in words with code and system variables
 * insertvar & insertbui are macros defined in qartus.h with the following
 * data; for insertbui:
 * arg1 is the type ('d' for system single variable and BWD for built-in word
 * arg2 is the length in characters of the identifier word
 * arg3 is the cbits (IMM/SMUDGE/BUFFER/COMPILE/DEFINING/ADDCOMP)
 * arg4 is the vocabulary code
 * arg5 is a string with the identifier name
 * arg6 is the word code (for its execution)
 * the first inserted is the last shown by WORDS and is the deeper in memory */
void _add(void) {
	int hb=0;
	PREVHERE=0;
	/* system variables: don't move or invert! */
	insertvar('d',4,0,VFORTH,"base",10);
	insertvar('d',5,0,VFORTH,"state",INTERPRET);
	insertvar('d',3,0,VFORTH,">in",0);
	insertvar('d',3,0,VFORTH,"csp",0);
	insertvar('d',4,0,VFORTH,"span",0); 
	insertvar('d',3,0,VFORTH,"blk",0);
	insertvar('d',3,0,VFORTH,"scr",0);
	insertvar('d',3,0,VFORTH,"bfr",0);
	insertvar('d',7,0,VFORTH,"context",0);
	insertvar('d',7,0,VFORTH,"current",0);
	insertvar('d',2,0,VFORTH,"#l",0);
	insertvar('d',2,0,VFORTH,"#p",0);
	insertvar('d',3,0,VFORTH,"dpl",0);
	insertvar('d',3,0,VFORTH,"fld",0);
	insertvar('d',5,0,VFORTH,"fence",0);
	insertvar('d',4,0,VFORTH,"last",0);
	insertvar('d',4,0,VFORTH,"#tib",0);
	insertvar('d',2,0,VFORTH,"r#",0);
	insertvar('d',2,0,VFORTH,"pc",0);
	insertvar('d',6,0,VFORTH,"offset",0);
	insertvar('d',3,0,VFORTH,"out",0);
	insertvar('d',7,0,VFORTH,"warning",0);
	insertvar('d',8,0,VFORTH,"#buffers",0);
	insertvar('d',5,0,VFORTH,"#vocs",2);
	insertvar('d',3,0,VFORTH,"err",-1);
	insertvar('d',5,0,VFORTH,"width",WORDNAMELENBASE);
	insertvar('d',3,0,VFORTH,"hld",0);
	insertvar('d',9,0,VFORTH,"fig-state",isFigCompatEnabled);
	insertvar('d',11,0,VFORTH,"debug-state",isDebug);
	insertvar('d',3,0,VFORTH,"ip1",127);
	insertvar('d',3,0,VFORTH,"ip2",0);
	insertvar('d',3,0,VFORTH,"ip3",0);
	insertvar('d',3,0,VFORTH,"ip4",1);
	insertvar('d',6,0,VFORTH,"#clock",NUMCLOCK);
	LOWVAR=HERE;
	/* built-in words */
	insertbui(BWD,11,32,VFORTH,"79-standard",_79_STANDARD);
	insertconst(5,0,VFORTH,"false",FALSE);
	insertconst(4,0,VFORTH,"true",TRUE);
	insertconst(5,0,VFORTH,"ernum",ERROR_NUM);
	insertconst(2,0,VFORTH,"s0",DATASTACKSTART);
	insertconst(2,0,VFORTH,"r0",RETURNSTACKSTART);
	insertconst(2,0,VFORTH,"w0",WORDBUFFERSTART);
	insertconst(5,0,VFORTH,"limit",_buffer_data[3].end+1);
	insertbui(BWD,1,33,VFORTH,"(",_PAREN);
	insertbui(BWD,2,33,VFORTH,"(s",_PAREN);
	insertbui(BWD,1,33,VFORTH,")",_CLOSE_PAREN);
	insertconst(3,0,VFORTH,"tib",TERMINALINPUTBUFFERSTART);
	insertbui(BWD,4,32,VFORTH,"word",_WORD);
	insertbui(BWD,6,8,VFORTH,"(word)",_ACT_WORD);
	insertbui(BWD,5,32,VFORTH,"abort",_ABORT);
	insertbui(BWD,7,32,VFORTH,"(abort)",_PAREN_ABORT);
	insertbui(BWD,12,32,VFORTH,"(block-read)",_BLOCK_READ);
	insertbui(BWD,13,32,VFORTH,"(block-write)",_BLOCK_WRITE);
	insertbui(BWD,9,32,VFORTH,"interpret",_INTERPRET);
	insertbui(BWD,5,33,VFORTH,"forth",_FORTH);
	insertbui(BWD,6,33,VFORTH,"editor",_EDITOR);
	insertbui(BWD,5,32,VFORTH,"order",_ORDER);
	insertbui(BWD,1,32,VFORTH,"#",_SHARP);
	insertbui(BWD,2,32,VFORTH,"<#",_LESSER_SHARP);
	insertbui(BWD,2,32,VFORTH,"#>",_SHARP_GREATER);
	insertbui(BWD,2,32,VFORTH,"#s",_SHARP_S);
	insertbui(BWD,4,32,VFORTH,"hold",_HOLD);
	insertbui(BWD,4,32,VFORTH,"sign",_SIGN);
	insertbui(BWD,5,32,VFORTH,"space",_SPACE);
	insertbui(BWD,6,32,VFORTH,"spaces",_SPACES);
	insertbui(BWD,6,52,VFORTH,"create",_CREATE);
	insertbui(BWD,5,41,VFORTH,"does>",_DOES_TO);
	insertbui(BWD,7,8,VFORTH,"(does>)",_PUSH_DOES);
	insertbui(BWD,4,32,VFORTH,"here",_HERE);
	insertbui(BWD,1,32,VEDITOR,"a",_EA);
	insertbui(BWD,1,32,VEDITOR,"l",_EL);
	insertbui(BWD,1,32,VEDITOR,"p",_EP);
	insertbui(BWD,1,32,VEDITOR,"c",_EC);
	insertbui(BWD,1,32,VEDITOR,"t",_ET);
	insertbui(BWD,1,32,VEDITOR,"d",_ED);
	insertbui(BWD,1,32,VEDITOR,"e",_EE);
	insertbui(BWD,1,32,VEDITOR,"r",_ER);
	insertbui(BWD,1,32,VEDITOR,"n",_EN);
	insertbui(BWD,1,32,VEDITOR,"b",_EB);
	insertbui(BWD,1,32,VEDITOR,"s",_ES);
	insertbui(BWD,1,32,VEDITOR,"i",_EI);
	insertbui(BWD,1,32,VEDITOR,"h",_EH);
	insertbui(BWD,1,32,VEDITOR,"f",_EF);
	insertbui(BWD,1,32,VEDITOR,"x",_EX);
	insertbui(BWD,1,32,VEDITOR,"m",_EM);
	insertbui(BWD,4,32,VEDITOR,"mark",_UPDATE);
	insertbui(BWD,3,32,VEDITOR,"top",_TOP);
	insertbui(BWD,5,32,VEDITOR,"stamp",_STAMP);
	insertbui(BWD,4,32,VEDITOR,"text",_ETEXT);
	insertbui(BWD,5,32,VEDITOR,"clear",_CLEAR);
	insertbui(BWD,4,32,VEDITOR,"till",_TILL);
	insertbui(BWD,4,32,VEDITOR,"copy",_COPY);
	insertconst(5,0,VFORTH,"b/buf",B_BUF);
	insertconst(5,0,VFORTH,"b/scr",1);
	insertconst(5,0,VFORTH,"l/scr",LPS+1);
	insertconst(3,0,VFORTH,"c/l",CPL);
	insertbui(BWD,5,32,VFORTH,"2over",_TWO_OVER);
	insertbui(BWD,4,32,VFORTH,"2rot",_TWO_ROT);
	insertbui(BWD,5,32,VFORTH,"2drop",_TWO_DROP);
	insertbui(BWD,5,32,VFORTH,"2swap",_TWO_SWAP);
	insertbui(BWD,4,32,VFORTH,"2dup",_TWO_DUP);
	insertbui(BWD,5,32,VFORTH,"depth",_DEPTH);
	insertbui(BWD,2,32,VFORTH,"c,",_C_COMMA);
	insertbui(BWD,1,32,VFORTH,",",_COMMA);
	insertbui(BWD,5,32,VFORTH,"allot",_ALLOT);
	insertbui(BWD,2,40,VFORTH,"r@",_R_FETCH);
	insertbui(BWD,2,40,VFORTH,"r>",_R_TO);
	insertbui(BWD,2,40,VFORTH,">r",_TO_R);
	insertbui(BWD,2,32,VFORTH,"0>",_ZERO_GREATER);
	insertbui(BWD,2,32,VFORTH,"0<",_ZERO_LESSER);
	insertbui(BWD,1,32,VFORTH,"<",_LESSER);
	insertbui(BWD,1,32,VFORTH,">",_GREATER);
	insertbui(BWD,2,32,VFORTH,"0=",_ZERO_EQUAL);
	insertbui(BWD,3,32,VFORTH,"not",_ZERO_EQUAL);
	insertbui(BWD,3,32,VFORTH,"xor",_XOR);
	insertbui(BWD,3,32,VFORTH,"and",_AND);
	insertbui(BWD,2,32,VFORTH,"or",_OR);
	insertbui(BWD,1,32,VFORTH,"=",_EQUAL);
	insertbui(BWD,2,32,VFORTH,"d=",_D_EQUAL);
	insertbui(BWD,2,32,VFORTH,"d*",_D_STAR);
	insertbui(BWD,2,32,VFORTH,"d/",_D_SLASH);
	insertbui(BWD,3,32,VFORTH,"d0=",_D_ZERO_EQUAL);
	insertbui(BWD,1,32,VFORTH,"+",_PLUS);
	insertbui(BWD,1,32,VFORTH,"-",_MINUS);
	insertbui(BWD,4,32,VFORTH,"emit",_EMIT);
	insertconst(2,0,VFORTH,"bl",BL);
	insertbui(BWD,4,32,VFORTH,"dump",_DUMP);
	insertbui(BWD,6,32,VFORTH,"negate",_NEGATE);
	insertbui(BWD,7,32,VFORTH,"dnegate",_DNEGATE);
	insertbui(BWD,2,32,VFORTH,"1+",_ONE_PLUS);
	insertbui(BWD,2,32,VFORTH,"1-",_ONE_MINUS);
	insertbui(BWD,2,32,VFORTH,"2+",_TWO_PLUS);
	insertbui(BWD,2,32,VFORTH,"2-",_TWO_MINUS);
	insertbui(BWD,2,32,VFORTH,"+!",_PLUS_STORE);
	insertbui(BWD,4,32,VFORTH,"dmax",_DMAX);
	insertbui(BWD,4,32,VFORTH,"dmin",_DMIN);
	insertbui(BWD,3,32,VFORTH,"max",_MAX);
	insertbui(BWD,3,32,VFORTH,"min",_MIN);
	insertbui(BWD,3,32,VFORTH,"abs",_ABS);
	insertbui(BWD,4,32,VFORTH,"dabs",_DABS);
	insertbui(BWD,1,32,VFORTH,"*",_STAR);
	insertbui(BWD,3,32,VFORTH,"m*/",_M_STAR_SLASH);
	insertbui(BWD,2,32,VFORTH,"u*",_U_STAR);
	insertbui(BWD,5,32,VFORTH,"u/mod",_U_SLASH_MOD);
	insertbui(BWD,2,32,VFORTH,"u<",_U_LESSER);
	insertbui(BWD,3,32,VFORTH,"du<",_D_U_LESSER);
	insertbui(BWD,2,32,VFORTH,"d<",_D_LESSER);
	insertbui(BWD,2,32,VFORTH,"d+",_D_PLUS);
	insertbui(BWD,2,32,VFORTH,"d-",_D_MINUS);
	insertbui(BWD,1,32,VFORTH,"/",_SLASH);
	insertbui(BWD,3,32,VFORTH,"mod",_MOD);
	insertbui(BWD,4,32,VFORTH,"/mod",_SLASH_MOD);
	insertbui(BWD,5,32,VFORTH,"*/mod",_STAR_SLASH_MOD);
	insertbui(BWD,2,32,VFORTH,"*/",_STAR_SLASH);
	insertbui(BWD,5,32,VFORTH,"shift",_SHIFT);
	insertbui(BWD,8,52,VFORTH,"variable",_VARIABLE);
	insertbui(BWD,8,52,VFORTH,"constant",_CONSTANT);
	insertbui(BWD,9,52,VFORTH,"2variable",_TWO_VARIABLE);
	insertbui(BWD,9,52,VFORTH,"2constant",_TWO_CONSTANT);
	insertbui(BWD,5,32,VFORTH,"query",_QUERY);
	insertbui(BWD,3,32,VFORTH,"key",_KEY);
	insertbui(BWD,9,32,VFORTH,"?terminal",_QUESTION_TERMINAL);
	insertbui(BWD,4,32,VFORTH,"quit",_QUIT);
	insertbui(BWD,7,41,VFORTH,"literal",_LITERAL);
	insertbui(BWD,8,41,VFORTH,"dliteral",_DLITERAL);
	insertbui(BWD,6,32,VFORTH,"expect",_EXPECT);
	insertbui(BWD,9,32,VFORTH,"-trailing",_TRAILING);
	insertbui(BWD,4,32,VFORTH,"fill",_FILL);
	insertbui(BWD,4,32,VFORTH,"move",_MOVE);
	insertbui(BWD,5,32,VFORTH,"cmove",_CMOVE);
	insertbui(BWD,7,32,VFORTH,"convert",_CONVERT);
	EXIT_CODE=HERE; // don't remove
	insertbui(BWD,4,40,VFORTH,"exit",_EXIT);
	insertbui(BWD,7,32,VFORTH,"decimal",_DECIMAL);
	insertbui(BWD,3,32,VFORTH,"hex",_HEX);
	insertbui(BWD,1,32,VFORTH,"?",_QUESTION);
	insertbui(BWD,1,32,VFORTH,"@",_FETCH);
	insertbui(BWD,2,32,VFORTH,"c@",_C_FETCH);
	insertbui(BWD,2,32,VFORTH,"2@",_TWO_FETCH);
	insertbui(BWD,1,32,VFORTH,"!",_STORE);
	insertbui(BWD,2,32,VFORTH,"c!",_C_STORE);
	insertbui(BWD,2,32,VFORTH,"2!",_TWO_STORE);
	insertbui(BWD,2,36,VFORTH,".\"",_DOT_QUOTES);
	insertbui(BWD,2,32,VFORTH,"d.",_D_DOT);
	insertbui(BWD,3,32,VFORTH,"ud.",_U_D_DOT);
	insertbui(BWD,3,32,VFORTH,"ud/",_U_D_SLASH);
	insertbui(BWD,7,44,VFORTH,"compile",_COMPILE);
	insertbui(BWD,9,8,VFORTH,"(compile)",_DEFERCOMPILE);
	insertbui(BWD,9,44,VFORTH,"[compile]",_BRACKETS_COMPILE);
	insertbui(BWD,10,9,VFORTH,"(sliteral)",_PUSH_STRING_ADDR);
	insertbui(BWD,7,32,VFORTH,"(sound)",_SOUND);
	insertbui(BWD,9,64,VFORTH,"immediate",_IMMEDIATE);
	insertbui(BWD,2,32,VFORTH,".r",_DOT_R);
	insertbui(BWD,2,32,VFORTH,"u.",_U_DOT);
	insertbui(BWD,3,32,VFORTH,"d.r",_D_DOT_R);
	insertbui(BWD,3,32,VFORTH,"u.r",_U_DOT_R);
	insertbui(BWD,1,32,VFORTH,"\\",_NOOP);
	insertbui(BWD,3,32,VFORTH,"pad",_PAD);
	insertbui(BWD,2,32,VFORTH,"cr",_CR);
	insertbui(BWD,1,41,VFORTH,";",_SEMICOLON);
	insertbui(BWD,1,32,VFORTH,":",_COLON);
	insertbui(BWD,4,32,VFORTH,"?dup",_QUESTION_DUP);
	insertbui(BWD,6,68,VFORTH,"forget",_FORGET);
	insertbui(BWD,10,52,VFORTH,"vocabulary",_VOCABULARY);
	insertbui(BWD,12,8,VFORTH,"(vocabulary)",_SET_VOC_NUM);
	insertbui(BWD,11,32,VFORTH,"definitions",_DEFINITIONS);
	insertbui(BWD,4,32,VFORTH,"roll",_ROLL);
	insertbui(BWD,4,32,VFORTH,"over",_OVER);
	insertbui(BWD,3,32,VFORTH,"rot",_ROT);
	insertbui(BWD,4,32,VFORTH,"pick",_PICK);
	insertbui(BWD,3,32,VFORTH,"dup",_DUP);
	insertbui(BWD,4,32,VFORTH,"drop",_DROP);
	insertbui(BWD,4,32,VFORTH,"swap",_SWAP);
	insertbui(BWD,1,32,VFORTH,"]",_CLOSED_SQUARE);
	insertbui(BWD,1,1,VFORTH,"[",_OPEN_SQUARE);
	insertbui(BWD,1,32,VFORTH,".",_DOT);
	insertbui(BWD,2,32,VFORTH,"h.",_H_DOT);
	insertbui(BWD,4,32,VFORTH,"type",_TYPE);
	insertbui(BWD,5,32,VFORTH,"count",_COUNT);
	insertbui(BWD,5,32,VFORTH,"words",_WORDS);
	insertbui(BWD,3,36,VFORTH,"see",_SEE);
	insertbui(BWD,7,32,VFORTH,"execute",_EXECUTE);
	insertbui(BWD,1,37,VFORTH,"\'",_TICK);
	insertbui(BWD,5,37,VFORTH,"ascii",_ASCII);
	insertbui(BWD,4,36,VFORTH,"find",_FIND);
	insertbui(BWD,2,41,VFORTH,"if",_IF);
	insertbui(BWD,4,8,VFORTH,"(if)",_RT_IF);
	insertbui(BWD,4,41,VFORTH,"else",_ELSE);
	insertbui(BWD,6,8,VFORTH,"(else)",_RT_ELSE);
	insertbui(BWD,5,41,VFORTH,"endif",_THEN);
	insertbui(BWD,4,41,VFORTH,"then",_THEN);
	insertbui(BWD,6,8,VFORTH,"(then)",_RT_THEN);
	insertbui(BWD,1,40,VFORTH,"i",_I);
	insertbui(BWD,1,40,VFORTH,"j",_J);
	insertbui(BWD,5,40,VFORTH,"leave",_LEAVE);
	insertbui(BWD,5,41,VFORTH,"begin",_BEGIN);
	insertbui(BWD,7,1,VFORTH,"(begin)",_RT_BEGIN);
	insertbui(BWD,6,41,VFORTH,"repeat",_REPEAT);
	insertbui(BWD,8,8,VFORTH,"(repeat)",_RT_REPEAT);
	insertbui(BWD,5,41,VFORTH,"while",_WHILE);
	insertbui(BWD,7,8,VFORTH,"(while)",_RT_WHILE);
	insertbui(BWD,5,41,VFORTH,"again",_AGAIN);
	insertbui(BWD,7,1,VFORTH,"(again)",_RT_AGAIN);
	insertbui(BWD,3,41,VFORTH,"end",_UNTIL);
	insertbui(BWD,5,41,VFORTH,"until",_UNTIL);
	insertbui(BWD,7,8,VFORTH,"(until)",_RT_UNTIL);
	insertbui(BWD,2,41,VFORTH,"do",_DO);
	insertbui(BWD,4,8,VFORTH,"(do)",_RT_DO);
	insertbui(BWD,4,41,VFORTH,"loop",_LOOP);
	insertbui(BWD,6,8,VFORTH,"(loop)",_RT_LOOP);
	insertbui(BWD,5,41,VFORTH,"+loop",_PLUS_LOOP);
	insertbui(BWD,7,8,VFORTH,"(+loop)",_RT_PLUS_LOOP);
	insertbui(BWD,4,41,VFORTH,"case",_CASE);
	insertbui(BWD,6,8,VFORTH,"(case)",_RT_CASE);
	insertbui(BWD,2,41,VFORTH,"of",_OF);
	insertbui(BWD,4,8,VFORTH,"(of)",_RT_OF);
	insertbui(BWD,5,41,VFORTH,"endof",_ENDOF);
	insertbui(BWD,7,8,VFORTH,"(endof)",_RT_ENDOF);
	insertbui(BWD,7,41,VFORTH,"endcase",_ENDCASE);
	insertbui(BWD,9,8,VFORTH,"(endcase)",_RT_ENDCASE);
	insertbui(BWD,5,32,VFORTH,"block",_BLOCK);
	insertbui(BWD,3,33,VFORTH,"-->",_MINUS_TO);
	insertbui(BWD,2,32,VFORTH,";s",_SEMI_S);
	insertbui(BWD,6,32,VFORTH,"buffer",_BUFFER);
	insertbui(BWD,12,32,VFORTH,"save-buffers",_SAVE_BUFFERS);
	insertbui(BWD,13,32,VFORTH,"empty-buffers",_EMPTY_BUFFERS);
	insertbui(BWD,4,32,VFORTH,"list",_LIST);
	insertbui(BWD,4,32,VFORTH,"load",_LOAD);
	insertbui(BWD,6,32,VFORTH,"update",_UPDATE);
	insertbui(BWD,2,32,VFORTH,".s",_DOT_S);
	insertbui(BWD,3,32,VFORTH,"32@",_32_FETCH);
	insertbui(BWD,3,32,VFORTH,"32!",_32_STORE);
	insertbui(BWD,3,32,VFORTH,"32,",_32_COMMA);
	insertbui(BWD,4,32,VFORTH,"?csp",_QUESTION_CSP);
	insertbui(BWD,4,32,VFORTH,"!csp",_STORE_CSP);
	insertbui(BWD,4,32,VFORTH,"date",_DATE);
	insertbui(BWD,5,32,VFORTH,"clock",_CLOCK);
	insertbui(BWD,4,32,VFORTH,"time",_TIME);
	insertbui(BWD,7,32,VFORTH,"version",_VERSION);
	insertbui(BWD,7,32,VFORTH,"license",_LICENSE);
	insertbui(BWD,6,32,VFORTH,"banner",_BANNER);
	insertbui(BWD,5,32,VFORTH,"where",_WHERE);
	insertbui(BWD,4,32,VFORTH,"page",_PAGE);
	insertbui(BWD,4,32,VFORTH,"bell",_BELL);
	insertbui(BWD,4,32,VFORTH,"beep",_BELL);
	insertbui(BWD,4,32,VFORTH,"task",_TASK);
	insertconst(2,0,VFORTH,"on",1);
	insertconst(3,0,VFORTH,"off",0);
	insertconst(5,0,VFORTH,"first",SCRATCHAREALIMIT+1);
	insertconst(5,0,VFORTH,"stdin",STDIN);
	insertconst(6,0,VFORTH,"stdout",STDOUT);
	insertconst(6,0,VFORTH,"stderr",STDERR);
	insertbui(BWD,7,32,VFORTH,"message",_MESSAGE);
	insertbui(BWD,5,32,VFORTH,"error",_ERROR);
	insertbui(BWD,6,32,VFORTH,"?error",_QUESTION_ERROR);
	insertconst(4,0,VFORTH,"noop",_NOOP);
	insertconst(3,0,VFORTH,"r/o",R_O);
	insertconst(3,0,VFORTH,"w/o",W_O);
	insertbui(BWD,9,32,VFORTH,"open-file",_OPEN_FILE);
	insertbui(BWD,10,32,VFORTH,"close-file",_CLOSE_FILE);
	insertbui(BWD,9,32,VFORTH,"read-line",_READ_LINE);
	insertbui(BWD,10,32,VFORTH,"write-line",_WRITE_LINE);
	insertbui(BWD,12,64,VFORTH,"compile-only",_COMPILE_ONLY);
	insertbui(BWD,9,64,VFORTH,"exec-only",_EXEC_ONLY);
	insertbui(BWD,5,64,VFORTH,"trace",_TRACE);
	insertbui(BWD,5,64,VFORTH,"debug",_DEBUG);
	insertbui(BWD,9,64,VFORTH,"searchlib",_SEARCH_LIB);
	insertbui(BWD,7,64,VFORTH,"require",_REQUIRE);
	insertbui(BWD,6,64,VFORTH,"ensure",_ENSURE);
	insertbui(BWD,6,64,VFORTH,"search",_SEARCH);
	insertbui(BWD,7,64,VFORTH,"include",_INCLUDE);
	insertbui(BWD,11,64,VFORTH,"save-system",_SAVE_SYSTEM);
	insertbui(BWD,11,64,VFORTH,"load-system",_LOAD_SYSTEM);
	insertbui(BWD,6,32,VFORTH,"cf-get",_CF_GET);
	insertbui(BWD,6,32,VFORTH,"cf-put",_CF_PUT);
	insertbui(BWD,3,32,VFORTH,"bye",_BYE);
	insertbui(BWD,3,32,VFORTH,"use",_USE);
	insertbui(BWD,3,32,VFORTH,"lib",_LIB);
	insertbui(BWD,4,32,VFORTH,"port",_PORT);
	insertbui(BWD,4,32,VFORTH,"send",_SEND);
	insertbui(BWD,6,32,VFORTH,"listen",_LISTEN);
	insertbui(BWD,4,32,VFORTH,"info",_INFO);
	insertbui(BWD,4,64,VFORTH,"edit",_EDIT);
	LAST_(PREVHERE);
	LAST_BUILTIN=PREVHERE;
	LOWMEM=HERE; 
}


/* _builtin()
 * execute built-in core words 
 * addr is the token address of coded word */
void _builtin(unsigned short addr) {
	int code=mem[addr];
	switch (code) {
		case _NOOP:
		case _CLOSE_PAREN:		
		case _TASK:             _noop(); break; // NOOP, ) \ comments 
		case _79_STANDARD:	_79_standard(); break;
		case _USE:		_use(); break;
		case _SOUND:		_qsound(); break;
		case _SEARCH_LIB:	_search_lib(); break;
		case _LIB:		_lib(); break;
		case _BEGIN:		_begin(); break;
		case _BYE:		_bye(); break;
		case _CF_GET:		_cf_get(); break;
		case _CF_PUT:		_cf_put(); break;
		case _RT_BEGIN:		_rt_begin(); break;
		case _BELL:		_bell(); break;
		case _INCLUDE:		_include(); break;
		case _REQUIRE:		_require(TRUE); break;
		case _ENSURE:		_require(FALSE); break;
		case _SEARCH:		_search(); break;
		case _SEMI_S:		_semi_s(); break;
		case _PAGE:		_page(); break;
		case _PAREN:		 isOpen=TRUE; _paren(); break;
		case _MESSAGE:		_message(); break;
		case _ERROR:		_error(); break;
		case _TRACE:		_trace(); break;
		case _QUESTION_ERROR:	_question_error(); break;
		case _QUESTION_CSP:	_question_csp(); break;
		case _STORE_CSP:	_store_csp(); break;
		case _KEY:		_key(); break;
		case _DEBUG:		_debug(); break;
		case _PUSH_STRING_ADDR:	_push_string_addr(addr); break;	
		case _BLOCK_READ:	_block_read(); break;	
		case _BLOCK_WRITE:	_block_write(); break;	
		case _QUERY:		_query(); break;
		case _DATE:		_date(); break;
		case _EDIT:		_edit(); break;
		case _CASE:		_case(); break;
		case _RT_CASE:		_rt_case(); break;
		case _OF:		_of(); break;
		case _RT_OF:		_rt_of(); break;
		case _ENDOF:		_endof(); break;
		case _RT_ENDOF:		_rt_endof(); break;
		case _ENDCASE:		_endcase(); break;
		case _RT_ENDCASE:	_rt_endcase(); break;
		case _CLOCK:		_clock(); break;
		case _TIME:		_time(); break;
		case _DOT_S:		_dot_s(); break;
		case _INTERPRET:	_interpret(); break;
		case _VERSION:		_version(); break;
		case _WORD_ENDER:	_word_ender(); break;
		case _REPEAT:		_repeat(); break;
		case _RT_REPEAT:	_rt_repeat(addr); break;
		case _PLUS_STORE:	_plus_store(); break;
		case _MINUS_TO:		_minus_to(); break;
		case _WHILE:		_while(); break;
		case _RT_WHILE:		_rt_while(addr); break;
		case _UNTIL:		_until(); break;
		case _RT_UNTIL:		_rt_until(addr); break;
		case _AGAIN:		_again(); break;
		case _RT_AGAIN:		_rt_again(addr); break;
		case _DO:		_do(); break;
		case _RT_DO:		_rt_do(); break;
		case _LOOP:		_loop(); break;
		case _RT_LOOP:		_rt_loop(); break;
		case _PLUS_LOOP:	_plus_loop(); break;
		case _RT_PLUS_LOOP:	_rt_plus_loop(); break;
		case _I:		_i(); break;
		case _J:		_j(); break;
		case _LEAVE:		_leave(); break;
		case _THEN:		_then(); break;
		case _RT_THEN:		_rt_then(); break;
		case _ELSE:		_else(); break;
		case _RT_ELSE:		_rt_else(); break;
		case _IF:		_if(); break;
		case _RT_IF:		_rt_if(); break;
		case _INFO:		_info(); break;
		case _WHERE:		_where(); break;
		case _BANNER:		_banner(); break;
		case _LICENSE:		_license(); break;
		case _DEFERCOMPILE:	_defer_compile(); break;
		case _COMPILE:		_compile(); break;
		case _BRACKETS_COMPILE:	_brackets_compile(); break;
		case _ALLOT:		_allot(); break;
		case _IMMEDIATE:	_immediate(); break;
		case _COMPILE_ONLY:	_compile_only(); break;
		case _EXEC_ONLY:	_exec_only(); break;
		case _HOLD:		_hold(); break;
		case _SPACE:		_spaces(TRUE); break;
		case _SPACES:		_spaces(FALSE); break;
		case _SIGN:		_sign(); break;
		case _SHARP:		_sharp(); break;
		case _SHARP_S:		_sharp_s(); break;
		case _LESSER_SHARP:	_lesser_sharp(); break;
		case _SHARP_GREATER:	_sharp_greater(); break;
		case _QUESTION_DUP:	_question_dup(); break;
		case _D_DOT_R:		_d_dot_r(); break;
		case _U_DOT_R:		_u_dot_r(); break;
		case _DOT_R:		_dot_r(); break;
		case _D_ZERO_EQUAL:	_d_zero_equal(); break;
		case _D_EQUAL:		_d_equal(); break;
		case _D_STAR:		_d_star(); break;
		case _D_SLASH:		_d_slash(); break;
		case _DMAX:		_dmax(); break;
		case _DMIN:		_dmin(); break;
		case _D_U_LESSER:	_d_u_lesser(); break;
		case _CONVERT:		_convert(); break;
		case _QUIT:		_quit(); break;
		case _QUESTION_TERMINAL:_question_terminal(); break;
		case _ASCII:		_ascii(); break;
		case _ABORT:		_abort(); break;
		case _PAREN_ABORT:	_paren_abort(); break;
		case _EXPECT:		_expect(); break;
		case _FORTH:		_forth(); break;
		case _EDITOR:		_editor(); break;
		case _ORDER:		_order(); break;
		case _TOP:		_top(); break;
		case _ET:		_t(); break;
		case _ED:		_d(); break;
		case _EX:		_x(); break;
		case _EB:		_b(); break;
		case _ES:		_s(); break;
		case _EF:		_f(); break;
		case _EH:		_h(); break;
		case _EE:		_e(); break;
		case _ER:		_r(); break;
		case _EN:		_n(); break;
		case _EM:		_m(); break;
		case _EI:		__i(); break;
		case _STAMP:		_stamp(); break;
		case _BLOCK:		_block(); break;
		case _BUFFER:		_buffer(); break;
		case _UPDATE:		_update(); break;
		case _ETEXT:		_etext(); break; // TEXT (editor)
		case _SAVE_BUFFERS:	_save_buffers(); break;
		case _EMPTY_BUFFERS:	_empty_buffers(); break;
		case _LIST:		_list(); break;
		case _LOAD:		_load(); break;
		case _LOAD_SYSTEM:	_load_system(); break;
		case _SAVE_SYSTEM:	_save_system(); break;
		case _TILL:		_till(); break;
		case _CLEAR:		_clear(); break;
		case _COPY:		_copy(); break;
		case _EL:		_l(); break;
		case _EP:		_p(); break;
		case _EA:		_a(); break;
		case _EC:		_c(); break;
		case _LITERAL:		_literal(); break;
		case _DLITERAL:		_dliteral(); break;
		case _WORD:		_word(addr); break;
		case _ACT_WORD:		_act_word(addr); break;
		case _EXIT:		__exit(addr); break;
		case _HERE:		_here(); break;
		case _DEPTH:		_depth(); break;
		case _32_FETCH:		_32_fetch(); break;
		case _32_COMMA:		_32_comma(); break;
		case _32_STORE:		_32_store(); break;
		case _R_FETCH:		_r_fetch(); break;
		case _TWO_FETCH:	_two_fetch(); break;
		case _C_FETCH:		_c_fetch(); break;
		case _C_STORE:		_c_store(); break;
		case _R_TO:		_r_to(); break;
		case _TO_R:		_to_r(); break;
		case _ZERO_LESSER:	_zero_lesser(); break;
		case _LESSER:		_lesser(); break;
		case _GREATER:		_greater(); break;
		case _ZERO_GREATER:	_zero_greater(); break;
		case _ZERO_EQUAL:	_zero_equal(); break;
		case _AND:		_and(); break;
		case _XOR:		_xor(); break;
		case _OR:		_or(); break;
		case _EQUAL:		_equal(); break;
		case _C_COMMA:		_c_comma(); break;
		case _COMMA:		_comma(); break;
		case _TWO_STORE:	_two_store(); break;
		case _TWO_ROT:		_two_rot(); break;
		case _TWO_OVER:		_two_over(); break;
		case _TWO_DROP:		_two_drop(); break;
		case _TWO_DUP:		_two_dup(); break;
		case _TWO_SWAP:		_two_swap(); break;
		case _TYPE:		_type(); break;
		case _COUNT:		_count(); break;
		case _FILL:		_fill(); break;
		case _TRAILING:		_trailing(); break;
		case _MOVE:		_move(); break;
		case _CMOVE:		_cmove(); break;
		case _D_PLUS:		_d_plus(); break;
		case _D_LESSER:		_d_lesser(); break;
		case _D_MINUS:		_d_minus(); break;
		case _OPEN_SQUARE:	_open_square(); break;
		case _CLOSED_SQUARE:	_closed_square(); break;
		case _SEMICOLON:	_semicolon(); break;
		case _COLON:		_colon(); break;
		case _PAD:		_pad(); break;
		case _EXECUTE:		_execute(); break;
		case _EMIT:		_emit(); break;
		case _DUMP:		_dump(); break;
		case _TICK:		_tick(addr); break;
		case _FIND:		_find(); break;
		case _CR:		_cr(); break;
		case _D_DOT:		_d_dot(); break;
		case _U_DOT:		_u_dot(); break;
		case _U_D_DOT:		_u_d_dot(); break;
		case _U_D_SLASH:	_u_d_slash(); break;
		case _SHIFT:		_shift(); break;
		case _STAR_SLASH_MOD:	_star_slash_mod(); break;
		case _SLASH_MOD:	_slash_mod(); break;
		case _MOD:		_mod(); break;
		case _ABS:		_abs(); break;
		case _DABS:		_dabs(); break;
		case _MAX:		_max(); break;
		case _MIN:		_min(); break;
		case _ONE_PLUS:		_one_plus(); break;
		case _ONE_MINUS:	_one_minus(); break;
		case _TWO_PLUS:		_two_plus(); break;
		case _TWO_MINUS:	_two_minus(); break;
		case _QUESTION:		_question(); break;
		case _VARIABLE:		_variable(); break;
		case _CONSTANT:		_constant(); break;
		case _TWO_VARIABLE:	_two_variable(); break;
		case _TWO_CONSTANT:	_two_constant(); break;
		case _DECIMAL: 		_decimal(); break;
		case _HEX: 		_hex(); break;
		case _FETCH: 		_fetch(); break;
		case _STORE: 		_store(); break;
		case _DOT_QUOTES: 	_dot_quotes(addr); break;
		case _DNEGATE:		_d_negate(); break;
		case _NEGATE: 		_negate(); break;
		case _ROLL: 		_roll(); break;
		case _OVER: 		_over(); break;
		case _ROT: 		_rot(); break;
		case _PICK: 		_pick(); break;
		case _DROP: 		_drop(); break;
		case _SWAP: 		_swap(); break;
		case _DUP: 		_dup(); break;
		case _PLUS: 		_plus(); break;
		case _MINUS: 		_minus(); break;
		case _STAR: 		_star(); break;
		case _U_LESSER:		_u_lesser(); break;
		case _U_STAR: 		_u_star(); break;
		case _U_SLASH_MOD:	_u_slash_mod(); break;
		case _STAR_SLASH:	_star_slash(); break;
		case _M_STAR_SLASH:	_m_star_slash(); break;
		case _SLASH: 		_slash(); break;
		case _DOT: 		_dot(); break;
		case _H_DOT: 		_h_dot(); break;
		case _SEE: 		_see(); break;
		case _WORDS: 		_words(); break;
		case _SET_VOC_NUM:	_set_voc_number(addr); break; 
		case _CREATE: 		_create(addr); break;
		case _FORGET:		_forget(); break;
		case _VOCABULARY:	_vocabulary(addr); break;
		case _DEFINITIONS:	_definitions(); break;
		case _OPEN_FILE:	_open_file(); break;
		case _CLOSE_FILE:	_close_file(); break;
		case _READ_LINE:	_read_line(); break;
		case _WRITE_LINE:	_write_line(); break;
		case _DOES_TO:		_does_to(); break;
		case _PUSH_DOES:	_word_ender(); break;
		case _PORT:		_port(); break;
		case _LISTEN:		_listen(); break;
		case _SEND:		_send(); break;
		default: fprintf(stdout,"(code %d unknown) ",code);
	}
}


/* _add_word()
 * add a user/builtin vocabulary entry 
 * if type=BWD insert a builtin word (code contains builtin opcode)
 * if type=UWD insert a user-defined word (code contains ref. address) */
int _add_word(int type, unsigned short where, unsigned short addr, int L, int CB, int V, char *name, int code) {
	mem[where++]=addr;
	mem[where++]=CB;
	mem[where++]=V; // voc code
	mem[where++]=MARKVAR;
	mem[where++]=L;
	while (*name) mem[where++]=(int) *name++;
	mem[where++]=type;
	mem[where++]=code;
	mem[where++]='\0';
	WCOUNT++;
	return where;
}


/* _add_var()
 * add a built-in vocabulary single cell variable 
 * if type='k' it's a system single cell constant 
 * if type='d' it's a system single cell variable 
 * if type='v' it's a user single cell variable */
int _add_var(int type, unsigned short where, unsigned short addr, int L, int CB, int V, char *name, int val) {
	mem[where++]=addr;
	mem[where++]=CB;
	mem[where++]=V; // voc code
	mem[where++]=MARKVAR;
	mem[where++]=L;
	while (*name) mem[where++]=(int) *name++;
	mem[where++]=type;
	mem[where++]=val%BYTE;
	mem[where++]=val/BYTE;
	mem[where++]='\0';
	WCOUNT++;
	return where;
}


/* _add_2_var()
 * add a built-in vocabulary double cell variable 
 * if type='V' it's a user double cell variable
 * if type='C' it's a user double cell constant */
int _add_2_var(int type, unsigned short where, unsigned short addr, int L, int CB, int V, char *name, int v1, int v2) {
	mem[where++]=addr;
	mem[where++]=CB;
	mem[where++]=V; // voc code
	mem[where++]=MARKVAR;
	mem[where++]=L;
	while (*name) mem[where++]=(int) *name++;
	mem[where++]=type;
	mem[where++]=v1%BYTE;
	mem[where++]=v1/BYTE;
	mem[where++]=v2%BYTE;
	mem[where++]=v2/BYTE;
	mem[where++]='\0';
	WCOUNT++;
	return where;
}


/* _setup()
 * initialize qartus environment */ 
void _setup(void) {
	if (BASE<BASELOW) BASE=BASELOW;
	if (BASE>BASELIMIT) BASE=BASELIMIT;
	STATE=INTERPRET;
	_erase_stacks(ALL);
	PC_(0);
	to_in_(0);
	/* load default vocabulary only on first _setup() invoking */
	if (!isSetup) _add();
	isSetup=TRUE;
}



/* _get_and_push() - system workhorse for interpreting and pushing numbers;
 * evaluate number in 'word', parse its digits, consider dots and the
 * starting minus sign according to base 'base' in case of integers, push 
 * number on data stack as single or double cell number, according to its 
 * format. 
 * return TRUE if push went ok, FALSE otherwise */
int _get_and_push(char* word, int base) {
	int isDouble=FALSE;
	int sign=1;
	long long int num=0;
	char snum[ILINE+1], *s=snum;
	
	DPL=-1;
	/* get digits storing sign aside */
	if (*word=='-') sign=-1, word++;
	/* this is not required by the Standard */
	else if (*word=='+') word++;
	while (*word) {
		if (*word=='-' || *word=='+') return FALSE;
		/* the first dot makes the number as double and set DPL */
		else if (*word==',') {
			word++;
			continue;
		}
		while (*word=='.') {
			isDouble=TRUE, word++;
			if (DPL<0) DPL++;
			continue;
		}
		*s++=*word;
		if (DPL>=0 && *word!='.') DPL++;
		word++;
	}
	*s='\0';s=snum;
	/* convert to signed double int; the conversion MUST 
	 * consume all characters; if there's a remainder, the format was not 
	 * fully complaint to base */
	num=strtoll(s,&s,base);
	/* check if it's a real number */
	if (*s) return FALSE;
	/* compiling state: compile number in proper form */
	if (STATE==COMPILE) {
		if (isDouble) {
			long long int high, low;
			ll2int(sign*num,&high,&low);
			mem[HERE++]='D';
			mem[HERE++]=(int)(high%BYTE);
			mem[HERE++]=(int)high/BYTE;
			mem[HERE++]=(int)(low%BYTE);
			mem[HERE++]=(int)low/BYTE;
		}
		else {
			short numv=num*sign;
			mem[HERE++]='I';
			mem[HERE++]=(int)(numv%BYTE);
			mem[HERE++]=(int)numv/BYTE;
		}
		return TRUE;
	}
	/* interpret state and debug */
	if (isDebug) {
		_dot_s();
		fputc('\n',stdout);
		if (isDouble) 
			fprintf(stdout,"\nPC:%05d  Pushing double number --> %s%s%s <-- \n",_PC,REDCOLOR,snum,RESETCOLOR);
		else 
			fprintf(stdout,"\nPC:%05d  Pushing number --> %s%s%s <-- \n",_PC,REDCOLOR,snum,RESETCOLOR);
		waitKey();
	}
	/* interpret state: push number */
	if (isDouble) {
		long long int high, low;
		ll2int(sign*num,&high,&low);
		_push(low);
		_push(high);
	}
	else {
		short numv=sign*num;
		_push(numv);
	}
	return TRUE;
}


/* _empty_string()
 * system function */
void _empty_string(char *str, int len) {
	register unsigned int i=0;
	for (i=0; i<=len; i++) *(str+i)='\0';
}

/* _inc_to_in_()
 * Increment >IN by one position in the buffer */
void _inc_to_in_(void) {
	int v=to_in+1;
	to_in_(v);
}


/* _erase_input_buffer()
 * erase input buffer before _fill_buff() in _evaluate_input_buffer() */
void _erase_input_buffer(void) {
	int i;
	for (i=0; i<B_BUF; i++) mem[TERMINALINPUTBUFFERSTART+i]='\0';
}


/* _fill_buff()
 * scan buffered input line for first word starting from >IN (to_in), until 
 * character BL is met, which stops parsing; 
 * the word is stored in buffer bb 
 * return TRUE if buffer was filled, FALSE if it's empty */
int _fill_buff(char *bb) {
	int len=1; // 31 characters 1..31
	_empty_string(bb,ILINE);
	b=bb;
	/* discard trailing blanks */
	while (isBlank(line_buffer(to_in))) _inc_to_in_(); 
	while (line_buffer(to_in) && !isBlank(line_buffer(to_in)) &&\
			len<WORDNAMELEN) {
		*b++=tolower(line_buffer(to_in));
		_inc_to_in_();
		len++;
	}
	*b='\0'; b=bb;
	if (*b) return TRUE;
	return FALSE;
}


/* _compare()
 * compare the forth word contained in memory (identified by NFA 'wordstart'
 * address and word length len), with a C string 'word' 
 * return true if they match completely */
int _compare(unsigned short wordstart, int len, char* word) {
	int i=0;
	while (len>0) {
		if (tolower(mem[wordstart+i])!=tolower(*word)) 
			return FALSE;
		len--; word++; i++;
	}
	/* TRUE if len is null and word is consumed
	 * (that is if the match is up to the last char) */
	return !len && !*word;
}


/* _ncompare()
 * compare the forth word contained in memory (identified by LFA 'addr'
 * address), with a C string 'word' for length 'len'
 * return addr if they match, zero otherwise */
int _ncompare(unsigned short addr, const char* word, int len) {
	int i=0;
	int wl=wordlen(addr);
	int ws=wordstart(addr);
	while (len>0) {
		if (tolower(mem[ws+i])!=tolower(*word) || i>=wl) 
			return FALSE;
		len--; word++; i++;
	}
	return addr;
}


/* _dupstr()
 * return a duplicate string of the word in the dictionary, given the LFA */
char *_dupstr(unsigned short addr) {
	int i=0;
	int wl=wordlen(addr);
	int ws=wordstart(addr);
	char *r;
	r=malloc(wl+1);
	while (wl>0) *(r+i)=mem[ws+i], i++, wl--;
	*(r+i)='\0';
	return r;
}


/* _find_word()
 * look for word in the current dictionary; if found, return address of it,
 * or else return 0 
 * Note: if mode = TRUE uses PREVHERE, otherwise uses LAST as the
 * starting address */
unsigned short _find_word(char* word, int mode) {
	int val, v=vocCount;
	int h=0; 
	int l=strlen(word);
	while (v>=0) {
		val=(int)pow(2,v);
		if (vocChain[(CONTEXT)] & val) {
			/* last voc entry */
			if (mode) h=PREVHERE;  
			else h=LAST;
			while (h>0) {
				if (!isSmudge(h) && voccode(h)==v && wordlen(h)==l && _compare(wordstart(h),wordlen(h),word)) 
					return h;
				h=prev(h);
			}
		}
		v--;
	}
	return 0; // not found
}


/* _ok()
 * print suitable confirmation */
void _ok(void) {
	/* avoid printing ok upon bye (which is ugly!) */
	if (!strcmp(word,"bye")) return;
	/* avoid printing ok in case of error (which is meaningless!) */
	if (isError) {
		isError=FALSE;
		return;
	}
	/* specialize */
	if (STATE==INTERPRET) fprintf(stdout,"%s",OK);
	else if (STATE==COMPILE) fprintf(stdout,"%s",OK_COMP);
}


/* word ender 
 * Note: at present this is simply a placeholder */
void _word_ender(void) {
	;
}


/* remove nasty characters at the end */
void _remove_nasty(char *str) {
	while (str[strlen(str)-1]==10 || str[strlen(str)-1]==13)
		str[strlen(str)-1]='\0';
}


/* _exec_one()
 * execute one instruction regardless of STATE
 * return number of added steps to the program counter */
void _exec_one(void) {
	/* built-in words */
	if (mem[_PC]==BWD) {
		if (isDebug && mem[_PC+1]!=_NOOP) {
			_dot_s();
			fputc('\n',stdout);
			fprintf(stdout,"\nPC:%05d  Executing built-in word --> %s",_PC,REDCOLOR);
			_print_builtin_name(mem[_PC+1]);
			fprintf(stdout,"%s%s",RESETCOLOR," <-- \n");
			waitKey();
		}
		_builtin(_PC+1); // pass code
	}
	/* user definition 
	 * Note: here PC is backed up twice: the first with cfpush() to 
	 * ensure CREATE/DOES> can choose the correct path in DOES> sections
	 * inherited by previous CREATE words; the second with oldPC, because
	 * this is the correct value needed to regain the correct position
	 * after EXIT; the first cfpush() value can be altered by DOES>
	 * and so it is not available as a correct return value, while oldPC
	 * is local and restores _PC ad hoc. */
	else if (mem[_PC]==UWD) {
		int bt=traceVal;
		int oldPC=_PC;
		if (isDebug) {
			_dot_s();
			fputc('\n',stdout);
			fprintf(stdout,"\nPC:%05d  Executing user word --> %s",_PC,REDCOLOR);
			_print_word_name(mem[_PC+1]);
			fprintf(stdout,"%s%s",RESETCOLOR," <-- \n");
			waitKey();
		}
		else if (traceVal!=mem[_PC+1]) traceVal=0; 
		_cfpush(_PC);
		countWords++;
		_exec_flow(codestart(mem[_PC+1]));
		countWords--;
		traceVal=bt;
		_cfpop();
		PC_(oldPC);
	}
	/* single-cell integer */
	else if (mem[_PC]=='I') {
		int v;
		if (isDebug) {
			_dot_s();
			fputc('\n',stdout);
			fprintf(stdout,"\nPC:%05d  Pushing number --> %s",_PC,REDCOLOR);
			_print(mem[_PC+1]+BYTE*mem[_PC+2],BASE,NULL);
			fprintf(stdout,"%s%s",RESETCOLOR," <-- \n");
			waitKey();
		}
		_push(mem[_PC+1]+BYTE*mem[_PC+2]);
		v=_PC+1;
		PC_(v);
	}
	/* double-cell integer */
	else if (mem[_PC]=='D') { 
		int v;
		if (isDebug) {
			_dot_s();
			fputc('\n',stdout);
			fprintf(stdout,"\nPC:%05d Pushing double number --> %s",_PC,REDCOLOR);
			_print(int2ll(mem[_PC+1]+BYTE*mem[_PC+2],mem[_PC+3]+BYTE*mem[_PC+4]),BASE,NULL);
			fprintf(stdout,"%s%s",RESETCOLOR," <-- \n");
			waitKey();
		}
		_push(mem[_PC+3]+BYTE*mem[_PC+4]);  // low
		_push(mem[_PC+1]+BYTE*mem[_PC+2]);  // high
		v=_PC+3;	
		PC_(v);
	}
	/* system variable */
	else if (mem[_PC]=='d') {
		if (isDebug) {
			_dot_s();
			fputc('\n',stdout);
			fprintf(stdout,"\nPC:%05d Calling system var --> %s",_PC,REDCOLOR);
			if (_PC+1<LOWMEM) {
				int pc=_PC+1;
				while (mem[pc]&&mem[pc]!=MARKVAR && pc<LOWMEM) 
					pc--;
				pc-=3;
				_print_word_name(pc);
			}
			else _print_word_name(mem[_PC+1]);

			fprintf(stdout,"%s%s",RESETCOLOR," <-- \n");
			waitKey();
		}
		if (_PC<LOWVAR) {
			int v=_PC+1;
			_push(_PC+1);
			PC_(v);
		}
		else _push(codestart(mem[_PC+1])+1);
	}
	/* single-cell user variable */
	else if (mem[_PC]=='v') {
		int v=_PC+1;
		_push(_PC+1);
		PC_(v);
	}
	/* double-cell user variable */
	else if (mem[_PC]=='V') {
		int v=_PC+3;
		_push(_PC+1);
		PC_(v);
	}
	/* single-cell system constant */
	else if (mem[_PC]=='k') { 
		int v;
		_push(mem[_PC+1]+BYTE*mem[_PC+2]);
		v=_PC+1;
		PC_(v);
	}
	/* single-cell user constant */
	else if (mem[_PC]=='c') { 
		int v;
		_push(mem[_PC+1]+BYTE*mem[_PC+2]);
		v=_PC+1;
		PC_(v);
	}
	/* double-cell user constant */
	else if (mem[_PC]=='C') {
		int v;
		_push(mem[_PC+3]+BYTE*mem[_PC+4]); // low
		_push(mem[_PC+1]+BYTE*mem[_PC+2]); // high
		v=_PC+3;
		PC_(v);
	}
}


/* _exec_flow()
 * the user word code parser
 * addr is the CFA address of word (the code start address) */
void _exec_flow(unsigned short addr) {
	PC_(addr);
	if (!currPC) currPC=_PC;
	/* interpret word program */
	while (mem[_PC]>0 && _PC<HIGHMEM && !isBye && !isQuit && !isInterrupt){
		int v;
		/* skip VOC_MARK, that serves as the hook for 
		 * returning the address */
		if (mem[_PC]==VOC_MARK) {
			v=_PC+2;
			PC_(v);
			continue;
		}
		/* close word in case of DOES> */
		else if (mem[_PC]==BWD && mem[_PC+1]==_PUSH_DOES) {
			_word_ender();
			return;
		}
		/* general close word routine 
		 * Note: this assures word execution is completed */
		else if (mem[_PC]==BWD && mem[_PC+1]==_WORD_ENDER) {
			_word_ender();
			return;
		}
		/* parse directly-compiled EXIT */
		else if (mem[_PC]==BWD && mem[_PC+1]==_EXIT) {
			_word_ender();
			return;
		}
		/* execute indirectly-compiled EXIT through COMPILE */
		else if (mem[_PC]==UWD && mem[_PC+1]==EXIT_CODE) {
			_doerr(loops>0,5);
			return;
		}
		/* execute the DEBUG function if required */
		else if (mem[_PC]==BWD && mem[_PC+1]==_DEBUG) {
			isDebug=FALSE;
			_exec_one();
		}
		/* execute the single (traced) instruction
		 * Note: in case of TRACE on addr, set up a temporary
		 * debugging sessions, resetting the original 
		 * debug state afterwards */
		else {
			int bd=isDebug;
			if (codestart(traceVal)==addr) {
				isDebug=TRUE;
				fputc('\n',stdout);
			}
			else isDebug=FALSE;
			_exec_one();
			if (!wasTrace) isDebug=bd;
			else isDebug=FALSE;
			wasTrace=FALSE;
		}

		/* update the program counter */
		v=_PC+2;
		PC_(v);
	}
}


/* exec_imm()
 * exec user immediate words (builtin immediate words depend on STATE) 
 * addr is the LFA address of the word */
void _exec_imm(unsigned short addr) {
	int bs=STATE, cbs;
	int bpc=_PC;
	PC_(codestart(addr));
	STATE=INTERPRET; // don't change!
	cbs=STATE;
	while (mem[_PC]>0 && !isBye && ! isQuit && !isInterrupt) {
		int v;
		/* execute word closing independently 
		 * Note: this assures it is completed (EXIT feature) */
		if (mem[_PC]==BWD && mem[_PC+1]==_WORD_ENDER) {
			_word_ender();
		}
		/* skip VOC_MARK, that serves as the hook for 
		 * returning the address */
		else if (mem[_PC]==VOC_MARK) {
			v=_PC+2;
			PC_(v);
			continue;
		}
		/* execute the single istruction and update the 
		 * program counter */
		else {
			_exec_one();
		}
		v=_PC+2;
		PC_(v);
	}
	if (cbs==STATE && ! provokeSTATE) {
		STATE=bs;
		provokeSTATE=FALSE;
	}
	PC_(bpc);
}


/* _exec()
 * the code parser for word type
 * addr is the LFA address of word (the start address) */
void _exec(unsigned short addr) {
	/* compile-only words (if they are not immediate) */
	if (isComp(addr) && STATE==INTERPRET && countWords<2 && !isExecute) {
		_perror(15,to_in);
		return;
	}
	/* execution-only words cannot be compiled*/
	else if (STATE==COMPILE && isExec(addr)){
		_perror(13,to_in);
		return;
	}
	/* compile COMPILE code */
	if (STATE==COMPILE && compile_ON) {
		mem[HERE++]=addr;
		compile_ON=FALSE;
	}
	/* execute immediate words 
	 * (since it calls _exec_one it works regardless of state) */
	else if (isImm(addr)) {
		if (addr<LOWMEM) _builtin(codestart(addr)+1);
		else {
			_exec_imm(addr);
		}
	}
	/* words in interpretation state */
	else if (STATE==INTERPRET) {
		unsigned short oldPC=_PC;
		// exec_flow() sets PC on its own
		_exec_flow(codestart(addr));
		PC_(oldPC);
	}
	/* compilation state */
	else if (STATE==COMPILE) {
		/* compile system variables */
		if (addr<LOWVAR) {
			mem[HERE++]='d';
			mem[HERE++]=addr;
		}
		/* exec added compilation behavior (only on built-in word
		 * that require this; for instance, constants don't) */
		else if (isAddComp(addr)) _builtin(codestart(addr)+1);
		/* default compile */
		else {
			mem[HERE++]=UWD;
			mem[HERE++]=addr;
			/* Compile DOESMARK and a zero as the setup of DOES>. 
			 * See note at _create() (compile section) */
			if (isDef(addr)) {
				mem[HERE++]=DOESMARK;
				mem[HERE++]=0;
				mem[currentDef+1]|=B_DEFINING;
			}
		}
	}
}


/* _database_search()
 * scan database for word 'word' passed as argument; if found, call exec
 * if not, try to convert it to number, otherwise signal the error 
 * Note: the class of super-immediate words are not Forth words, and have a 
 * special debug meaning */
void _database_search(char* word) {
	unsigned short addr=0;
	if (!*word || isQuit || isInterrupt) return;
	/* super-immediate words */
	/* COLD */
	if (!strcmp(word,"cold")) {
		_cold();
		return;
	}
	/* BYE */
	else if (!strcmp(word,"bye")) {
		_bye();
		return;
	}
	/* SH */
	else if (!strcasecmp(word,"sh")) {
		int in=to_in+1;
		char bb[B_BUF+1], *b=bb;
		fputc('\n',stdout);
		fputc('\n',stdout);
		to_in_(in);
		while (isblank(line_buffer(to_in))) _inc_to_in_();
		while (line_buffer(to_in) && to_in<B_BUF) {
			*b++=line_buffer(to_in);
			_inc_to_in_();
		}
		*b='\0';
		system(bb);
		OUT=0;
		isQuit=TRUE;
		return;
	}
	/* SYS/MON */
	else if (!strncmp(word,"sys",3) || !strncmp(word,"mon",3)) {
		struct termios orig_termios;
		unsigned short scr=SCR1+SCR*BYTE;
		int i;
		tcgetattr(STDIN_FILENO, &orig_termios);
		tcsetattr(STDIN_FILENO, TCSANOW, &base_termios);
		/* ensure that block-edits are saved before calling
		 * the shell, in order to avoid data dispersion */
		for (i=1; i<=3; i++) {   // BFR range
			if (_buffer_data[i].update) {
				save_buffer(scr,i);
			}
			_buffer_data[i].update=FALSE;
		}
		/* enter shell mode */
		fprintf(stdout,"%s","\n\nEntering shell mode.\n\n");
		system(SHELL);
		fprintf(stdout,"\n%sqartus restored.\n\n",RESETCOLOR);
		tcsetattr(STDIN_FILENO, TCSANOW, &orig_termios);
		return;
	}
	/* CF@ */
	else if (!strcmp(word,"cf@")) {
		_print_cf();
		return;
	}
	/* DS@ */
	else if (!strcmp(word,"ds@")) {
		_uprint_s();
		return;
	}
	/* RS@ */
	else if (!strcmp(word,"rs@")) {
		_uprint_r();
		return;
	}
	/* GETSTATE */
	else if (!strcmp(word,"getstate")) {
		fprintf(stdout,"%d ",STATE);
		OUT+=2;
		return;
	}
	/* (SOUND) */
	else if (!strcmp(word,"(sound)")) {
		_qsound();
		return;
	}
	/* DS */
	else if (!strcmp(word,"ds")) {
		_push(ds);
		return;
	}
	/* PREVHERE */
	else if (!strcmp(word,"prevhere")) {
		_push(PREVHERE);
		return;
	}
	/* @STAT */
	else if (!strcmp(word,"@stat")) {
		isStat=TRUE;
		return;
	}
	/* CHECK-DICT */
	else if (!strcmp(word,"check-dict")) {
		unsigned short addr1, addr2;
		int i=1, count=0, wcount=1;
		int oldprev=PREVHERE;
		while (i<HIGHMEM && mem[i]!=MARKVAR) i++;
		addr1=i;
		fprintf(stdout,"\nChecking and fixing dictionary (currently %d words):\n",WCOUNT);
		/* check the word tree starting from BASE (at addr=1)*/
	        do {
			i=addr1+1;
			while (i<HIGHMEM && mem[i]!=MARKVAR) i++;
			addr2=i;
			if (i>=HIGHMEM) break;
			if (mem[addr2-3]!=addr1-3) {
				PREVHERE=addr1-3;
				LAST_(addr1-3);
				HERE=addr2-3;
				break;
			}
			else {
				fputc('.',stdout);
				count++;
				wcount++;
				if (count>SCREEN) {
					fputc('\n',stdout);
					count=0;
				}
				addr1=addr2;
			}
		} while	(TRUE);
		if (oldprev==PREVHERE) {
			fprintf(stdout,"%s","\nDictionary ok.\n");
			fprintf(stdout,"%d words checked.\n",wcount);
		}
		else {
			int w=WCOUNT-wcount;
			fprintf(stdout,"%s","\nDictionary fixed. Last word is now: ");
			_print_word_name(PREVHERE);
			fputc('\n',stdout);
			if (w!=1) fprintf(stdout,"%d words were removed.\n",w);
			else fprintf(stdout,"%s","1 word was removed.\n");
			fprintf(stdout,"%d words remaining.\n",wcount);
		}
		WCOUNT=wcount;
		return;
	}
	/* is word a dictionary entry? */
	addr=_find_word(word,FALSE);
	if (addr>0) {
		countWords++;
		_exec(addr);
		countWords--;
		fflush(stdout);
	}
	/* or ...is it any number? */
	else if (!_not_a_number(word)) {
		/* if number is illegal it's not pushed */
		if (!_get_and_push(word,BASE)) {
			fprintf(stdout," %s  ? ",word);
			_perror(-1,-1);
		}
	}
	/* if not, it's not a forth word */
	else {
		fprintf(stdout," %s  ? ",word);
		_perror(-1,-1);
	}
}


/* the core of the buffer interpretation process, which selects next word in 
 * the input buffer (user input string) and processes it until buffer is over; 
 * to_in is a variable that corresponds to the address >IN, required
 * by the standard, and is used also for the precise position of error index 
 * shown by function _perror() (pointing to first character of unresolved word) 
 * inline comments and paren comments (immediate) are intercepted here */
void _evaluate_input_buffer(void) {
	int i=0;
	PC_(0);
	currPC=OUT=0;  // erase program counters
	stopKey=FALSE; // debug variable
	/* fill back buffer (for various usages) */
	b=backbuffer;
	while (line_buffer(i) && i<B_BUF-1) *(b+i)=line_buffer(i), i++;
	while (i<B_BUF-1) *(b+i)='\0', i++;
	*(b+i)=0;
	/* parse paren comments */
	if (isOpen) _paren();
	/* evaluate by calling the database searcher */
	while (line_buffer(to_in) && !isBye && !isInterrupt && !isMinusTo && !isSemiS) {
		/* _fill_buff() updates >IN automatically */
		_fill_buff(word);
		/* parse slash comments */
		if (!strcmp(word,"\\")) return;
		/* invoke the database */
		if (*word) _database_search(word);
		if (isQuit) break;
	}
	/* print last stack state in case of debug */
	if (isDebug) {
		_dot_s();
	}
	isInterrupt=FALSE;
	/* execute indirectly '-->' by previously resetting its flag */
	if (!isFile && isMinusTo) {
		isMinusTo=FALSE;
		_push((BLK1+BYTE*BLK)+1);
		_load();
	}
	/* reset --> and ;S flags */
	isMinusTo=FALSE;
	isSemiS=FALSE;
}


/* the core of the LOAD buffer interpretation process, which copies the whole
 * block in the line input buffer and passes the information to
 * _evaluate_input_buffer, which executes the line. */
void _evaluate_block_buffer(int curr_to_in) {
	int i=OFFSET1+OFFSET*BYTE;
	int pos;
	to_in_(0);
	/* copy to line input buffer */
	while (mem[_buffer_data[BFR].start+i] && i<B_BUF) {
	    line_buffer(to_in)=mem[_buffer_data[BFR].start+i];
	    _inc_to_in_();
	    i++;
	}
	line_buffer(to_in)='\0';
	pos=0;
	/* remove slash comments until the end of line */
	while (pos<B_BUF) {
		if (line_buffer(pos)=='\\') {
			int i;
			for (i=pos;i<64*((int)(pos/64)+1);i++) {
				line_buffer(i)=BL;
			}
			pos=64*((int)(pos/64)+1)-1;
		}
		pos++;
	}
	/* set #TIB */
	SHARP_TIB1=to_in%BYTE;
	SHARP_TIB=to_in/BYTE;
	/* Reset current >IN and evaluate */
	to_in_(curr_to_in);
	if (line_buffer(to_in)) _evaluate_input_buffer();
}


/* _file_open()
 * open a file and try adding extension .ft if not found;
 * return the file descriptor of the opened file, if found in either way,
 * or null if neither is found */
FILE* _file_open(char* original) {
	char fname[ILINE+1];
	FILE *ff;
       
	/* search in current directory */
	strcpy(fname,original);
	ff = fopen(fname, "r");
	if (!ff) {
		strcat(fname, ".ft");
		ff=fopen(fname, "r");
		if (!ff) return NULL;
	}
	return ff;
}


/* _parse_file()
 * the core of the file loading process, which reads each line of the
 * opened file and parse each non-void line as a forth sequence of words */
void _parse_file(char* filename) {
	char *p;
	char iline[B_BUF+1];
	FILE *f=NULL;

	/* open the file*/
	f = _file_open(filename);
	if (!f) { 
		if (filename) {
			f = _file_open(filename);
			if (!f) { 
			     	fprintf(stdout,"? file %s not found.\n",filename);
				return;
	        	}
	    	}
		else return;
        	}

	isFile=TRUE;
	/* parse file */
	while (fgets(iline,B_BUF,f) && !isBye && !isInterrupt && !isStopFile) {
		_remove_nasty(iline);
		p=iline;
		to_in_(0);
		/* copy to line input buffer */
		while (*p) {
			line_buffer(to_in)=*p++;
			_inc_to_in_();
		}
		line_buffer(to_in)='\0';
		/* set #TIB */
		SHARP_TIB1=to_in%BYTE;
		SHARP_TIB=to_in/BYTE;
		/* reset current >IN and evaluate */
		to_in_(0);
		if (line_buffer(to_in)) _evaluate_input_buffer();
		if (isQuit) break;
		if (isStopFile) break;
	}
	/* close file and reset >IN */
	if (f) fclose(f);
	isFile=FALSE;

	/* in case of interrupt, don't do anything else */
	if (isInterrupt) return;
	
	/* if state is still compile, there is 
	 * an unfinished colon definition */
	if (STATE==COMPILE && !isQuit) {
		_perror(48,to_in); // sets STATE
	}
	isQuit=FALSE;
	isStopFile=FALSE;
}


/* _process_input()
 * substitute for rl_newline(), which augments its functionality in the
 * keyboard input by the current user; it's never called directly; rather, it's
 * called by readline when the Enter Key is pressed.
 * Set up input buffer and pass control to _evaluate_input_buffer();
 * Values c and k are needed for readline binding, but are not used here.
 * Note: thanks to Chet Ramey, for some tips about using readline correctly */
int _process_input(int c, int k) {
	char *p=rl_line_buffer;
	int i=0, l=strlen(rl_line_buffer);

	/* redisplay, that is erase current line and repaint
         * Note: rl_redisplay() doesn't meet my standards */
	while(l>0) {
		fputc('\b',stdout);
		l--;
	}
	fprintf(stdout,"%s",rl_line_buffer);

	/* proceed with evaluation of the input line */
	setsig(); // readline seems to reset signals, someway...
	if (!isPage) fputc(' ',stdout); // space to separate input from output
	/* fill qartus input buffer for evaluation */
	i=0;
	while (*p) line_buffer(i++)=*p++;
	line_buffer(i)='\0';
	/* set #TIB */
	SHARP_TIB1=i%BYTE;
	SHARP_TIB=i/BYTE;
	/* resetting >IN */
	to_in_(0);
	/* evaluate */
	isError=FALSE;
	RunningErrorInstance=FALSE;
	_evaluate_input_buffer();
	/* print ok if QUIT was not used */
	if (!isQuit && !isPage) _ok();
	isQuit=isPage=FALSE;
	rl_newline(c,k); // print a New Line for the new input
	return 0;
}


/* _idle()
 * The idle cycle: wait for the user keyboard input; the input string is 
 * automatically processed by readline by calling _process_input()
 * (see tf_startup_hook() ) */
void _idle(void){
	while (!isBye) {
		/* enable autocomplete */
		rl_bind_key('\t',rl_complete);
		/* read input line */
		inputline=readline(prompt);
		if (!inputline) continue;
		if (!*inputline) continue;
		//if (*inputline==EOF) continue;
		/* manage history */
		if (inputline && *inputline) {
			rl_replace_line(inputline,0);
			add_history(inputline);
		}
	}
}


/* tf_startup_hook()
 * set up parameters for readline
 * I'm not really sure it's necessary, but it seems so... */
int tf_startup_hook(void) {
	rl_bind_key(13,_process_input);
	rl_bind_key(10,_process_input);
	return 0;
}


/* tf_completion()
 * readline completion function */
static char **tf_completion(const char *text, int start, int end) {
	int ret=0;
	if (!strchr(text,' ')) rl_attempted_completion_over=0;
	else rl_attempted_completion_over=1;
	/* set return type */
	if (!rl_attempted_completion_over) ret=1;
	else ret=0;

	if (ret==1) {
		rl_attempted_completion_over=0;
		return rl_completion_matches(text, tf_generator);
	}
	else {
		rl_attempted_completion_over=1;
		return rl_completion_matches(text, tf_fgenerator);
	}
	return rl_completion_matches(text, tf_generator);
}


/* tf_generator()
 * readline generator function */
char* tf_generator(const char *text, int state) {
	static int len=0;
	static int h;
	if (!state) {
		h=PREVHERE;
		len=strlen(text);
	}
	/* search strings */
	while (h>0) {
		int res=_ncompare(h,text,len);
		h=prev(h);
		if (res) {
			return _dupstr(res);
		}
	}
 	/* If no names matched, then return NULL. */
	return ((char *)NULL);
}


/* tf_fgenerator()
 * readline file generator function */
char* tf_fgenerator(const char *text, int state) {
 	/* If no names matched, then return NULL. */
	return ((char *)NULL);
}
 

/* sigint()
 * interrupts treatment */
void sigint(int sig) {
	int i=0;
	fprintf(stdout,"%s","\n? user interrupt.\nType Enter a few times to regain access to qartus");
	isInterrupt=TRUE;
	qquit(sig);
	while (i<B_BUF) line_buffer(i++)='\0';
	to_in_(0);
	rl_on_new_line();
	rl_point=rl_end;
	_perror(-1,-1);
}


/* setsig()
 * interrupts treatment */
void setsig(void) {
	new_action.sa_handler = sigint;
   	sigemptyset(&new_action.sa_mask);
   	new_action.sa_flags = 0;
	sigaction(SIGINT, &new_action, NULL);
	sigaction(SIGHUP, &new_action, NULL);
}


/* _print_help()
 * help on qartus usage; execute -h/--help options */
void _print_help(int flag) {
	if(flag) fprintf(stdout,"%s\n",LOGO);
	fprintf(stdout,"%s","\nUsage: qartus [options] [file1] [file2] [file3] ...\n");
	fprintf(stdout,"%s","Options:\n\n");
	fprintf(stdout,"%s","      --command=<str>            execute words in <str> after file loading\n");
	fprintf(stdout,"%s","      --conditions               print GPL3 redistribution conditions\n");
	fprintf(stdout,"%s","      --clear                    clear the screen before session start\n");
	fprintf(stdout,"%s","      --config                   install library and blocks in system dir\n");
	fprintf(stdout,"%s","      --errors-list              print a list of all the system error strings\n");
	fprintf(stdout,"%s","  -f  --fig-forth                enable the fig-compatibility at start\n");
	fprintf(stdout,"%s","  -h, --help                     display this information and exit\n");
	fprintf(stdout,"%s","      --init=<str>               execute words in <str> before file loading\n");
	fprintf(stdout,"%s","  -k  --keep                     enable both --clear and --no-banner\n");
	fprintf(stdout,"%s","  -m  --micro                    set CLOCK to return microseconds\n");
	fprintf(stdout,"%s","  -n <n> --newline=<n>           print <n> new lines after each file execution\n");
 	fprintf(stdout,"%s","      --no-banner                start without the textual banner\n");
	fprintf(stdout,"%s","  -Q, --no-quiet                 don't close session even if an error is found\n");
	fprintf(stdout,"%s","      --pad=<n>                  use <n> as the default PAD length\n");
	fprintf(stdout,"%s","      --prompt=<str>             set <str> as prompt (default is a null string)\n");
	fprintf(stdout,"%s","  -q, --quiet                    close session after executing all files\n");
	fprintf(stdout,"%s","  -r <n> --return-stack-size=<n> set return stack size (range 2048..16384)\n");
	fprintf(stdout,"%s","  -s <n> --stack-size=<n>        set data stack size (range 2048..32767)\n");
	fprintf(stdout,"%s","      --statistics               print execution statistics before exiting\n");
	fprintf(stdout,"%s","  -t  --timing                   print execution timing at exit\n");
	fprintf(stdout,"%s","  -v, --version                  display version and exit\n");
	fprintf(stdout,"%s","      --warranty                 print GPL3 warranty conditions\n\n");
	fprintf(stdout,"%s","The [file] items are common textual files containing Forth code.\n"); 
	fprintf(stdout,"%s","Numbers <n> are integer values; they are checked and reset if unfitting.\n"); 
	fprintf(stdout,"%s","If strings <str> contain spaces, they must be enclosed in double quotes.\n");
	fprintf(stdout,"%s","Type BYE to exit. Type WORDS to list all available dictionary entries.\n");
	fprintf(stdout,"%s","Type man qartus to consult the man page. There is also a pdf manual.\n");
	fprintf(stdout,"%s","If your program seems to hang, type CTRL-C and then ENTER a few times,\n");
	fprintf(stdout,"%s","until you see 'ok'. Type COLD anytime to restart a new default state.\n");
	if (flag) fprintf(stdout,"%s","Forth? qartus!\n\n");
	fprintf(stdout,"%s","\n");
}


void  _print_license(void) {
	fprintf(stdout,"\n%s\n","This program is free software; you can redistribute it and/or modify\nit under the terms of the GNU General Public License as published by\nthe Free Software Foundation; either version 3 of the License, or\n(at your option) any later version.\n\nThis program is distributed in the hope that it will be useful,\nbut WITHOUT ANY WARRANTY; without even the implied warranty of\nMERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\nGNU General Public License for more details.\n\nYou should have received a copy of the GNU General Public License\nalong with this program. If not, see http://www.gnu.org/licenses/.");	
}


/* _print_warranty()
 * warranty message (required by GPL3); execute --warranty option */
void _print_warranty(int flag) {
	if (flag) _print_version(FALSE);
	fprintf(stdout,"%s","Warranty\n--------\nTHERE IS NO WARRANTY FOR THE PROGRAM, TO THE EXTENT PERMITTED BY\nAPPLICABLE LAW. EXCEPT WHEN OTHERWISE STATED IN WRITING THE COPYRIGHT\nHOLDERS AND/OR OTHER PARTIES PROVIDE THE PROGRAM \"AS IS\" WITHOUT\nWARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING, BUT NOT\nLIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A\nPARTICULAR PURPOSE. THE ENTIRE RISK AS TO THE QUALITY AND PERFORMANCE OF\nTHE PROGRAM IS WITH YOU. SHOULD THE PROGRAM PROVE DEFECTIVE, YOU ASSUME\nTHE COST OF ALL NECESSARY SERVICING, REPAIR OR CORRECTION.\n");
}


/* _print_version()
 * version message; execute -v/--version options */
void _print_version(int flag) {
	fprintf(stdout,"%s\n",LOGO);
	fprintf(stdout,"\nqartus version %d.%d-%s (built on %s, %s); ", VERSION_MAJ,VERSION_MIN, VERSION_STR,__DATE__, __TIME__);
	if (flag) fprintf(stdout,"%s","it comes\nwith ABSOLUTELY NO WARRANTY; for details type 'qartus --warranty'.\n");
	if (flag) fprintf(stdout,"%s","This is free software, and you are welcome to redistribute it\n");
	if (flag) fprintf(stdout,"%s","under certain conditions; type 'qartus --conditions' for details.\n");
	if (!flag) fprintf(stdout,"%s","\n");
	fprintf(stdout,"%s","qartus is released under the terms of the GNU General Public License 3.\n");
	fprintf(stdout,"Copyleft (C) %s Antonio Maschio. For bug reporting, please write to\n",YEARS);
	if (flag) fprintf(stdout,"%s","<ing dot antonio dot maschio at gmail dot com>\nIt's GPL! Enjoy!\n");
	else fprintf(stdout,"%s","<ing dot antonio dot maschio at gmail dot com>.\n");
	if (!flag) fprintf(stdout,"%s","\n");
}



/* _print_conditions()
 * conditions message (required by GPL3); execute --conditions option */
void _print_conditions(int flag) {
	if (flag) _print_version(FALSE);
	fprintf(stdout,"%s","Conditions\n----------\n");
	fprintf(stdout,"%s","You may make, run and propagate covered works that you do not convey,\n");
	fprintf(stdout,"%s","without conditions so long as your license otherwise remains in force.\n");
	fprintf(stdout,"%s","You may convey covered works to others for the sole purpose of having\n");
	fprintf(stdout,"%s","them make modifications exclusively for you, or provide you with\n");
	fprintf(stdout,"%s","facilities for running those works, provided that you comply with the\n");
	fprintf(stdout,"%s","terms of this License in conveying all material for which you do not\n");
	fprintf(stdout,"%s","control copyright. Those thus making or running the covered works for\n");
	fprintf(stdout,"%s","you must do so exclusively on your behalf, under your direction and\n");
	fprintf(stdout,"%s","control, on terms that prohibit them from making any copies of your\n");
	fprintf(stdout,"%s","copyrighted material outside their relationship with you.\n");
}


/* print_long_banner() */
void print_long_banner(void) {
	fprintf(stdout,"\nqartus %d.%d-%s - Copyleft (C) %s Antonio Maschio - GPL3 license\n",VERSION_MAJ,VERSION_MIN,VERSION_STR,YEARS);	
	fprintf(stdout,"%s","qartus is an implementation of Forth-79 with Double-Number-Standard Extension");
	if (isFigCompatEnabled) fprintf(stdout,",\n%s\n","in this session with fig-forth emulation for what about the execution process,");
	else fprintf(stdout,"%s","\n");
	fprintf(stdout,"%s\n","coming free, with sources, and with ABSOLUTELY NO WARRANTY.");
	fprintf(stdout,"%s\n\n","For details, type 'license'; type 'bye' to exit.");
}


/* doConfig()
 * perform --config for Linux/Unix/Cygwin shell execution
 * Note: this function assumes to be in the installation directory of qartus 
 * to be run with --config
 * Note: these directories are for a Linux/Unix/CygWin machine
 * HID_PATH   =.qartus/
 * RES_NAME   =.qartus.rc
 * BASEHOME   =/home/<uname>/
 * HID_HOME   =/home/<uname>/.qartus/
 * RES_PATH   =/home/<uname>/.qartus.rc
 * LIBS_PATH  =/home/<uname>/.qartus/qlibrary.ft
 * BLOCKS_PATH=/home/<uname>/.qartus/blocks/
 **/
void doConfig(void) {
	int done=TRUE;
	FILE *fi;
	DIR *dir;
	char comm[ILINE];
	
	/* Linux/Unix/CygWin banner */
	fprintf(stdout,"\nqartus %d.%d-%s - Copyleft (C) %s Antonio Maschio - GPL3 license\n",VERSION_MAJ,VERSION_MIN,VERSION_STR,YEARS);	
	fprintf(stdout,"%s\n","Wait for the creation of the system environment configuration under shell...");

	/* create environment (variables were created in main() )*/
	fprintf(stdout,"%s\n","Creating main directories...");
	mkdir(HID_HOME,0755);
	mkdir(BLOCKS_PATH,0755);	

	/* check for library file and blocks dir and also blocks dir
	 * in the installation folder; if any is not found, set done=FALSE */
	fprintf(stdout,"%s\n","Checking newly created directories and sources consistency...");
	fi=fopen(LIBS_NAME,"r");
	if (!fi) done=FALSE;
	else fclose(fi);
	dir = opendir(BLOCKS_PATH);
	if (!dir) done=FALSE;
	else closedir(dir);
	dir = opendir("./blocks/");
	if (!dir) done=FALSE;
	else closedir(dir);

	/* copy library file (through shell) in overwrite mode 
	 * to the default hidden directory and copy all blocks in the 
	 * blocks directory */
	fprintf(stdout,"%s\n","Copying...");
	if (done) {
		strcpy(comm,"cp -f ");
		strcat(comm,LIBS_NAME);
		strcat(comm," ");
		strcat(comm,LIBS_PATH);
		system(comm);
		strcpy(comm,"cp -f blocks/* ");
		strcat(comm,BLOCKS_PATH);
		system(comm);
		fprintf(stdout,"%s\n\n","Installation complete. Happy qartus!");
	}
	else {
		fprintf(stdout,"\n%s\n","Installation couldn't be completed.");
		fprintf(stdout,"%s\n","Causes: either you are not in the installation folder...");
		fprintf(stdout,"%s\n","...or you have not the rights to create the default directories...");
		fprintf(stdout,"%s\n\n","...or there is an error in the file system. Check which is your case...");
	}
}


/* doWinConfig()
 * perform --config for Windows Prompt execution
 * Note: this function assumes to be in the installation directory of qartus 
 * to be run with --config
 * Note: these directories are for a Windows machine
 * BASEHOME   =C:\Users\<uname>\AppData\Roaming\ == $APPDATA
 * HID_HOME   =C:\Users\<uname>\AppData\Roaming\qartus\ $APPDATA+HID_PATH
 * RES_PATH   =C:\Users\<uname>\AppData\Roaming\qartus\_qartus.rc
 * LIBS_PATH  =C:\Users\<uname>\AppData\Roaming\qartus\qlibrary.ft
 * BLOCKS_PATH=C:\Users\<uname>\AppData\Roaming\qartus\blocks/ 
 * This function was written by Antonio Maschio and Bruce Axtens
 * January 2020 */
void doWinConfig(char *appdata) {
	char *comm = alloca(4*ILINE);
	int done=TRUE;
	char *resFolder = alloca(4*ILINE);
	char *blocksFolder = alloca(4*ILINE);
	char *buffer = alloca(4*ILINE);
	char *ptr = buffer;
	FILE *fi;
	DIR *dir;

	/* Windows banner */
	fprintf(stdout, "\nqartus %d.%d-%s - Copyleft (C) %s Antonio Maschio - GPL3 license\n", VERSION_MAJ, VERSION_MIN, VERSION_STR, YEARS);
	fprintf(stdout, "%s\n\n", "Wait for system environment configuration under the Windows Prompt...");
	
	/* Create qartus folder in appdata roaming. 
	 * Note appdata has trailing backslash */
	strncpy(BASEHOME,appdata,4*ILINE);
	strcpy(RES_PATH,BASEHOME);
	strcat(RES_PATH,"_qartus.rc");
	strcpy(resFolder, appdata);
	strcat(resFolder, "qartus");
	strncpy(LIBS_PATH,resFolder,4*ILINE);
	strcat(LIBS_PATH,LIBS_NAME);
	strncpy(HID_HOME,resFolder,4*ILINE);
	strncpy(HISTORY_PATH,resFolder,4*ILINE);
	strcat(HISTORY_PATH,"_qartus_history");
	mkdir(resFolder, 0755);
	//fprintf(stdout, "mkdir of %s = %d\n", resFolder, result);

	// create blocks folder in appdata roaming qartus
	strcpy(blocksFolder, resFolder);
	strcat(blocksFolder, "\\");
	strcat(blocksFolder, "blocks");
	strncpy(BLOCKS_PATH,blocksFolder,4*ILINE);
	mkdir(blocksFolder, 0755);
	//fprintf(stdout, "mkdir of %s = %d\n", blocksFolder, result);

	/* check for library file and blocks dir; 
	 * if not found, set done=FALSE */
	fprintf(stdout,"%s\n","Checking newly created directories and sources consistency...");
	fi=fopen(LIBS_NAME,"r");
	if (!fi) done=FALSE;
	else fclose(fi);
	dir = opendir(BLOCKS_PATH);
	if (!dir) done=FALSE;
	else closedir(dir);
	dir = opendir(".\\blocks\\");
	if (!dir) done=FALSE;
	else closedir(dir);

	/* copy library file (through Windows Prompt) in overwrite mode 
	 * to the default hidden directory */
	fprintf(stdout,"%s\n","Copying...");
	if (done) {
		strcpy(buffer, "cmd /c copy ");
		strcat(buffer, LIBS_NAME);
		strcat(buffer, " ");
		strcat(buffer, resFolder);
		/* erase command and rebuild it for Windows usage */
		strcpy(comm, "");
		ptr = buffer;
		while (*ptr != 0) {
	        	if (*ptr == '\\') {
	            		strcat(comm, "\\");
	            		strcat(comm, "\\");
	        	}
			else {
				strncat(comm, (const char *)ptr, 1);
			}
			ptr++;
		}
		//fprintf(stdout, "comm = %s\n", comm);
		system(comm);
	
		/* copy all blocks in the blocks directory (appdata roaming) */
		strcpy(buffer, "cmd /c copy ");
		strcat(buffer, "blocks");
		strcat(buffer, "\\*.*");
		strcat(buffer, " ");
		strcat(buffer, "\"");
		strcat(buffer, blocksFolder);
		strcat(buffer, "\"");
		/* erase command and rebuild it for Windows usage */
		strcpy(comm, "");
		ptr = buffer;
		while (*ptr != 0) {
	        	if (*ptr == '\\') {
	            		strcat(comm, "\\");
	            		strcat(comm, "\\");
	        	}
	        	else {
	            		strncat(comm, (const char *)ptr, 1);
	        	}
	        	ptr++;
	    	}
		//fprintf(stdout, "comm = %s\n", comm);
		system(comm);
	}
	else {
		fprintf(stdout,"\n%s\n","Installation couldn't be completed.");
		fprintf(stdout,"%s\n","Causes: either you are not in the installation folder...");
		fprintf(stdout,"%s\n\n","...or there is an error in the file system. Check which is your case...");
	}
}


/* perform @STAT */
void doStat(void) {
	int i=0, l=0, hour=0, min=0, sec=0;
	int wnum=0, vnum=0, bnum=0;
	char sysname[ILINE+1], output[ILINE+1], uname[ILINE+1];
	long long int now;
	double val;
	int ehour,emin,esec;
	int eyear,emon,eday;

	/* record some time data */
	gettimeofday(&tv, NULL);
	now = (long long int)(tv.tv_sec*1000.0 + tv.tv_usec/1000.0);
	val=(now-beginS)/1000.0; // seconds needed here

	/* calculate session time */
	hour=(int)(val/60.0/60.0);
	val=val-hour*60.0*60.0;
	min=(int)(val/60.0);
	sec=(int)(val-min*60.0);
	
	/* retrieve system and user name */
	gethostname(sysname,ILINE);
	strcpy(uname,getenv("LOGNAME"));
	
	/* calculate dictionary data */
	vnum=NVOCS-2;
	wnum=_count_peculiar(0,LOWMEM-1);
	i=1;
	while (i<4) {
	       	if(_buffer_data[i].in_use) bnum++;
		i++;
	}

	/* register end clock */
	gettimeofday(&tv, NULL);
	ehour=getHour();
	emin=getMin();
	esec=getSec();
	eyear=getYear();
	emon=getMonth();
	eday=getDay();

	/* print statistics */
	sprintf(output, "Statistics for the session of %d %s %d",getDay(), month[getMonth()],getYear());
	l=strlen(output);
	fprintf(stdout,"\n%s\n",output);
	i=0; while (i<l) fputc('-',stdout), i++;
	fprintf(stdout,"%s","\n");
	fprintf(stdout,"User name:         %s\n",uname);
	fprintf(stdout,"System name:       %s\n",sysname);
	fprintf(stdout,"qartus version:    %d.%02d %s\n",VERSION_MAJ,VERSION_MIN,VERSION_STR);
	fprintf(stdout,"compiled on        %s   %s\n\n",__DATE__,__TIME__);
	fprintf(stdout,"Session started on %02d/%02d/%04d at %02d:%02d:%02d\n",sday,smon,syear,shour,smin,ssec);
	fprintf(stdout,"Session ended on   %02d/%02d/%04d at %02d:%02d:%02d\n",eday,emon,eyear,ehour,emin,esec);
	fprintf(stdout,"execution time:    %02d:%02d:%02d\n\n",hour,min,sec);
	fprintf(stdout,"added dictionary words  : %5d\n",wnum);
	fprintf(stdout,"added vocabularies      : %5d\n",vnum);
	fprintf(stdout,"used buffers            : %5d\n\n",bnum);
	fprintf(stdout,"dictionary size         : %5d\n",HIGHMEM);
	fprintf(stdout,"  used memory           : %5d\n",HERE-1);
	fprintf(stdout,"  free memory           : %5d\n",HIGHMEM-HERE+1);
	fprintf(stdout,"data stack size         : %5d\n",DSSTACKLENGTH);
	fprintf(stdout,"return stack size       : %5d\n",RSSTACKLENGTH);
	i=0; while (i<l) fputc('-',stdout), i++;
	fprintf(stdout,"%s","\n\n");
}


/* do_close() 
 * perform closing operations */
void do_close(void) {
	tcsetattr(STDIN_FILENO, TCSANOW, &base_termios);
	if (isStat) doStat();
}


/* MAIN
 * "Per correr miglior acque alza le vele
 *  omai la navicella del mio ingegno..."
 * Dante, Purgatorio, Canto I */
int main(int argc, char **argv, char **envp) {
	int i, c, ac = argc;
	int exitNow=FALSE, exitSin=FALSE;
	typedef int Function ();
	FILE *fres;

	/* register closing function */
	atexit(do_close);

	/* register signals mask */
	sigaction(SIGSTOP, NULL, &old_action);

	/* register terminal settings */
	tcgetattr(STDIN_FILENO, &base_termios);
	
	/* set signal handler (CTRL-C interrupt) */
	setsig();

	/* register start clock 
	 * Note: beginS is in milliseconds, beginM is in microseconds */
	gettimeofday(&tv, NULL);
        beginS = (long long int)(tv.tv_sec*1000.0 + tv.tv_usec/1000.0);
        beginM = (long long int)(tv.tv_sec*1000000.0 + tv.tv_usec);
	NUMCLOCK=0; // set default for milliseconds
	shour=getHour();
	smin=getMin();
	ssec=getSec();
	syear=getYear();
	smon=getMonth();
	sday=getDay();
	
	/* qartus directories */
	struct passwd *pw = getpwuid(getuid());
	const char *hd = pw->pw_dir;
	/* set current directory (e.g. /home/user/progs/qartus/) */
	getcwd(currdir,(size_t)ILINE);
	strcat(currdir,"/");
	/* set home directory (e.g. /home/user/) */
	strcpy(BASEHOME,hd);
	strcat(BASEHOME,"/");
	/* set resource file name (e.g. /home/user/.qartus.rc) */
	strcpy(RES_PATH,BASEHOME);
	strcat(RES_PATH,RES_NAME);
	/* set base hidden directory $HOME/.qartus/ */
	strcpy(HID_HOME,BASEHOME);
	strcat(HID_HOME,HID_PATH);
	/* set base readline history file */
	strcpy(HISTORY_PATH,BASEHOME);
	strcat(HISTORY_PATH,HISTORY_NAME);
	/* set default lib name and create if not found */
	strcpy(LIBS_PATH,HID_HOME);
	strcat(LIBS_PATH,LIBS_NAME);
	/* set default blocks directory (e.g. /home/user/.qartus/blocks/) */
	strcpy(BLOCKS_PATH,HID_HOME);
	strcat(BLOCKS_PATH,BLOCKS_NAME);

	/* set default prompt */
	strcpy(prompt,PROMPT);
	cline[0]=0;
	/* parse interpreter arguments */
	while (--ac > 0 && '-' == (*++argv)[0]) {
		exitSin=FALSE;
		/* parse double-dash options */
		if (!strcmp(*argv, "--help")) {
			_print_help(TRUE);
			exit(EXIT_SUCCESS);
		}
		else if (!strcmp(*argv, "--version")) {
			_print_version(TRUE);
			exit(EXIT_SUCCESS);
		}
		else if (!strcmp(*argv, "--conditions")) {
			_print_conditions(TRUE);
			exit(EXIT_SUCCESS);
		}
		else if (!strcmp(*argv, "--warranty")) {
			_print_warranty(TRUE);
			exit(EXIT_SUCCESS);
		}
		/* option --config */
		else if (!strncmp(*argv,"--config",8)) {
			/* Linux, Unix and Cygwin can use doConfig()
			 * Windows under shell must use doWinConfig() */
			char appdata[ILINE];
			char infopath[ILINE];
			int i;
			/* set default values */
			appdata[0]=infopath[0]=0;
			/* search for APPDATA (and set INFOPATH) */
			for (i = 0; envp[i] != NULL; i++) {
				if (!strncasecmp(envp[i],"APPDATA",7)) {
					strncpy(appdata,envp[i]+8,ILINE-1);
					strcat(appdata,"\\");
				}
				if (!strncasecmp(envp[i],"INFOPATH",8)) {
					strncpy(infopath,envp[i]+9,ILINE-1);
				}
			}
			/* decide the configuration process */
			/* no APPDATA found: it's a Linux machine 
			 * (use default parameters) */
			if (!appdata[0]) {
				doConfig();
			}
			/* APPDATA found: it's a Windows machine */
			else {
				/* no INFOPATH found: it's a Windows OS 
				 * (pass ad-hoc parameters)*/
				if (!infopath[0]) {
					doWinConfig(appdata);
				}
				/* INFOPAT found: it's a CygWin environment 
			 	 * (use default Linux/Unix parameters) */
				else {
					doConfig();
				}
			}
			exit(EXIT_SUCCESS);
		}
		/* option --micro */
		else if (!strncmp(*argv,"--micro",7)) {
			NUMCLOCK=1; // set default for microseconds
			continue;
		}
		/* option --quiet */
		else if (!strncmp(*argv,"--quiet",7)) {
			exitNow=TRUE;
			continue;
		}
		/* option --no-quiet */
		else if (!strncmp(*argv,"--no-quiet",10)) {
			exitOnFile=FALSE;
			continue;
		}
		/* option --newline for print separation */
		else if (!strncmp(*argv,"--newline=",10)) {
			newline=strtol(*argv+10,NULL,10);
			isPrintSeparator=TRUE;
			continue;
		}
		/* option --stack-size */
		else if (!strncmp(*argv,"--stack-size=",13)) {
			DSSTACKLENGTH=strtol(*argv+13,NULL,10);
			if (DSSTACKLENGTH<ZDSSTACKLENGTH) 
				DSSTACKLENGTH=ZDSSTACKLENGTH;
			if (DSSTACKLENGTH>WDSSTACKMAX) 
				DSSTACKLENGTH=WDSSTACKMAX;
			continue;
		}
		/* option --return-stack-size */
		else if (!strncmp(*argv,"--return-stack-size=",20)) {
			RSSTACKLENGTH=strtol(*argv+20,NULL,10);
			if (RSSTACKLENGTH<ZRSSTACKLENGTH) 
				RSSTACKLENGTH=ZRSSTACKLENGTH;
			if (RSSTACKLENGTH>WRSSTACKMAX) 
				RSSTACKLENGTH=WRSSTACKMAX;
			continue;
		}
		/* option --print-execution-info */
		else if (!strncmp(*argv,"--statistics",12)) {
			isStat=TRUE;
			continue;
		}
		/* option --prompt */
		else if (!strncmp(*argv,"--prompt=",9)) {
			int qt=0;
			int j=9;
			if (**argv+j=='\"') j++, qt=1;
			else if (**argv+j=='\'') j++, qt=2;
			strcpy(prompt, *argv+j);
			prompt[strlen(prompt)]='\0';
			if (qt==1 && prompt[strlen(prompt)]=='\"')
				prompt[strlen(prompt)]='\0';
			else if (qt==2 && prompt[strlen(prompt)]=='\'')
				prompt[strlen(prompt)]='\0';
			continue;
		}
		/* option --fig-forth */
		else if (!strncmp(*argv,"--fig-forth",11)) {
			isFigCompatEnabled=TRUE;
			continue;
		}
		/* option --pad */
		else if (!strncmp(*argv,"--pad=",6)) {
			padlen=strtol(*argv+6,NULL,10);
			if (padlen<PADLEN) padlen=PADLEN;
			if (padlen>PADMAX) padlen=PADMAX;
			continue;
		}
		/* option --init - before file loading */
		else if (!strncmp(*argv,"--init=",7)) {
			int qt=0;	// quote type
			int j=7;
			if (**argv+j=='\"') j++, qt=1;
			else if (**argv+j=='\'') j++, qt=2;
			strcpy(nline, *argv+j);
			nline[strlen(nline)]='\0';
			if (qt==1 && nline[strlen(nline)]=='\"')
				nline[strlen(nline)]='\0';
			else if (qt==2 && nline[strlen(nline)]=='\'')
				nline[strlen(nline)]='\0';
			continue;
		}
		/* option --command - after file loading*/
		else if (!strncmp(*argv,"--command=",10)) {
			int qt=0;
			int j=10;
			if (**argv+j=='\"') j++, qt=1;
			else if (**argv+j=='\'') j++, qt=2;
			strcpy(cline, *argv+j);
			cline[strlen(cline)]='\0';
			if (qt==1 && cline[strlen(cline)]=='\"')
				cline[strlen(cline)]='\0';
			else if (qt==2 && cline[strlen(cline)]=='\'')
				cline[strlen(cline)]='\0';
			continue;
		}
		/* option --errors-list */
		else if (!strncmp(*argv,"--errors-list",13) || !strncmp(*argv,"--errors",8) || !strncmp(*argv,"--error-list",12)) {
			int j;
			fprintf(stdout,"%s","\nqartus's errors list:\n\n");
			for (j=0;j<=ERROR_NUM;j++) {
				fprintf(stdout,"%02d  %s\n",j,error_list[j]);
			}
			fprintf(stdout,"%s","\n");
			exit(0);
		}
		/* option --timing */
		else if (!strncmp(*argv,"--timing",8) || !strncmp(*argv,"--timer",7)) {
			isTotalTiming=TRUE;
			continue;
		}
		/* option --no-banner */
		else if (!strncmp(*argv,"--no-banner",11)) {
			isBanner=FALSE;
			continue;
		}
		/* option --clear */
		else if (!strncmp(*argv,"--clear",7)) {
			fprintf(stdout,"%s%s","[2J","[H");
			continue;
		}
		/* option --keep */
		else if (!strncmp(*argv,"--keep",6)) {
			fprintf(stdout,"%s%s","[2J","[H");
			isBanner=FALSE;
			continue;
		}
		/* option --qartus 
		 * UNDOCUMENTED */
		else if (!strncmp(*argv,"--qartus",8)) {
			fprintf(stdout,"\n%s\n","Okay, well, qartus is not the original name of this software...\nIt was born as 'tforth' (in the sense of 'tonibin's forth'), but I soon\nfound out that one Jef Raskin, in 1987, had written a forth engine\nand called it just 'tforth'.\nSo I had to change the name of my program to 'yforth' because\na long Internet search showned that apparently no other 'yforth'\nhad ever been created. It was not true.\nA certain Luca Padovani in 1997 created a program and gave it\nthe name 'yforth', and went on curing it until, as far as I can know, 2012.\nSo, again, I had to change the name of my program;\ndevelopping a tip told me by Bruce, I conceived the name 'qartus'.\nThis is the story; 'tforth' and 'yforth' are beautiful names\nthat I couldn't use but that I won't forget.\n\nNow, type qartus again, without this option. Thanks.");
			fprintf(stdout,"%s","\n");
			exit(0);
		}

		/* this feature is intentionally undocumented 
		 * I use it to generate the README file */
		else if (!strcmp(*argv, "--greetings")) {
			fprintf(stdout,"%s","Welcome, qartus user! I hope you\'ll like ");
			fprintf(stdout,"%s","this program.\n");
			fprintf(stdout,"%s","Here\'s some information about it ");
			fprintf(stdout,"%s","and its license...\n");
			_print_help(FALSE);
			fprintf(stdout,"%s","To compile type 'make' as user and ");
			fprintf(stdout,"%s","'make install' as root in qartus's source\n");
			fprintf(stdout,"%s","directory. For help about qartus's features, after");
			fprintf(stdout,"%s"," installation, see the manual.\nYou can also consult the man page typying 'man qartus' anywhere.\n\n");
			fprintf(stdout,"%s","\n");
			_print_warranty(FALSE);
			fprintf(stdout,"%s","\n\n");
			_print_conditions(FALSE);
			fprintf(stdout,"%s","\n\n");
			fprintf(stdout,"%s","It's GPL! Enjoy!\n\n\n");
			exit(EXIT_SUCCESS);
		}
		/* Ignore unrecognized options */
		else if (!strncmp(*argv,"--",2)) {
			fprintf(stdout,"unknown option %s. Ignoring.\n", *argv);
			continue;
		}
		
		/* parse single-dash options */ 
		while (!exitSin && (c = *++argv[0])) switch(c) {
			case 'h': 	_print_help(TRUE); 
					exit(EXIT_SUCCESS);
			case 'f':	isFigCompatEnabled=TRUE;
					break;
			case 'Q': 	exitOnFile=FALSE; 
					break;
			case 'k':	isBanner=FALSE;
					fprintf(stdout,"%s%s","[2J","[H");
					break;
			case 'm':	NUMCLOCK=1; 
					// set default for microseconds
					break;
			case 'n':
					newline=strtol(*argv+1,NULL,10);
					isPrintSeparator=TRUE;
					exitSin=TRUE;
					break;
			case 'q': 	exitNow=TRUE;
					break;
			case 'r': 	
					RSSTACKLENGTH=strtol(*argv+1,NULL,10);
					if (RSSTACKLENGTH<ZRSSTACKLENGTH) 
						RSSTACKLENGTH=ZRSSTACKLENGTH;
					if (RSSTACKLENGTH>WRSSTACKMAX) 
						RSSTACKLENGTH=WRSSTACKMAX;
					exitSin=TRUE;
					break;
			case 's': 	
					DSSTACKLENGTH=strtol(*argv+1,NULL,10);
					if (DSSTACKLENGTH<ZDSSTACKLENGTH) 
						DSSTACKLENGTH=ZDSSTACKLENGTH;
					if (DSSTACKLENGTH>WDSSTACKMAX) 
						DSSTACKLENGTH=WDSSTACKMAX;
					exitSin=TRUE;
					break;
			case 't': 	isTotalTiming=TRUE;
					break;
			case 'v': 	_print_version(TRUE); 
					exit(EXIT_SUCCESS);
			/* UNDOCUMENTED*/
			case 'V': 	fprintf(stdout,"%d.%d-%s",(int)VERSION_MAJ,(int)VERSION_MIN,VERSION_STR); 
					exit(EXIT_SUCCESS);
			default:  	fprintf(stdout,"unknown option -%c. Ignoring.\n", c); 
		}
	}
	
	/* SETUP PROGRAM MEMORY */
	HIGHMEM=	ADDRLIMIT-
			RSSTACKLENGTH-
			DSSTACKLENGTH-
			5*B_BUF-
			BUFFERSLENGTH-	
			1; // I subtract 1 because 0 is an address
	/* input buffer memory */
	TERMINALINPUTBUFFERSTART=HIGHMEM+1;
	TERMINALINPUTBUFFERLIMIT=HIGHMEM+B_BUF; 
	/* WORD buffer memory */
	WORDBUFFERSTART=TERMINALINPUTBUFFERLIMIT+1;
	WORDBUFFERLIMIT=TERMINALINPUTBUFFERLIMIT+B_BUF;
	/* scratch area memory */
	SCRATCHAREASTART=WORDBUFFERLIMIT+1;
	SCRATCHAREALIMIT=WORDBUFFERLIMIT+BUFFERSLENGTH;
	SC=SCRATCHAREASTART;
	/* block buffers memory */
	_buffer_data[1].start=SCRATCHAREALIMIT+1;
	_buffer_data[1].end=  SCRATCHAREALIMIT+B_BUF;
	_buffer_data[2].start=_buffer_data[1].end+1;
	_buffer_data[2].end=  _buffer_data[1].end+B_BUF;
	_buffer_data[3].start=_buffer_data[2].end+1;
	_buffer_data[3].end=  _buffer_data[2].end+B_BUF;
	_buffer_data[1].update=FALSE;
	_buffer_data[2].update=FALSE;
	_buffer_data[3].update=FALSE;
	_buffer_data[1].in_use=FALSE;
	_buffer_data[2].in_use=FALSE;
	_buffer_data[3].in_use=FALSE;
	_buffer_data[1].number=0;
	_buffer_data[2].number=0;
	_buffer_data[3].number=0;
	_buffer_data[1].order=3;
	_buffer_data[2].order=2;
	_buffer_data[3].order=1;
	/* data and return stacks */
	DATASTACKSTART=_buffer_data[3].end+1;
	DATASTACKLIMIT=_buffer_data[3].end+DSSTACKLENGTH;
	RETURNSTACKSTART=DATASTACKLIMIT+1;
	RETURNSTACKLIMIT=DATASTACKLIMIT+RSSTACKLENGTH;
	CFSTACKSTART=0; CFSTACKLIMIT=CFSTACKSTART+CFSTACKLENGTH;
	/* setup stack pointers */
	ds=DATASTACKSTART;
	rs=RETURNSTACKSTART;
	cfs=CFSTACKSTART;
	/* set up default vocabularies  */
	vocCount=1;
	strncpy(vocs[0].name,"forth",6);
	vocChain[0]=1; // 2^0
	strncpy(vocs[1].name,"editor",7);
	vocChain[1]=3; // 2^0 + 2^1

	/* reset channel pointers */
	i=0;
	while(i<=FILELIMIT) {
		yfile[i].state=FALSE;
		yfile[i].fd=NULL;
		i++;
	}

	/* STEP 0: first of all, populate dictionary with built-in words */
	_setup();
	/* setup readline stuff for interactive session */
	rl_attempted_completion_function = tf_completion;
	rl_startup_hook = (Function *) tf_startup_hook;
	/* note: read_history() creates HISTORY in case it does not exist */
	read_history(HISTORY_PATH);		
	
	/* STEP 1: execute init string if given */
	if (nline[0]) {
		int i=0;
		while (i<512 && nline[i]) { 
			line_buffer(i)=nline[i];
			i++;
		}
		while (i<512) { 
			line_buffer(i)=0;
			i++;
		}
		to_in_(0);
		_evaluate_input_buffer();
		fprintf(stdout,"%s","\n");
	}

	/* STEP 2: parse resource file */
	fres=fopen(RES_PATH,"r");
	if (!!fres) {
		isRes=TRUE;
		fclose(fres);
		_parse_file(RES_PATH);
	}	

	/* STEP 3:
	 * read files (if any) passed as arguments, and parse them in turn, 
	 * separating each file output newlines if required 
	 * (options --newline) */
	argc=0;
	while (argc<ac) {
		strncpy(filearray[fc],*(argv+argc),FILENAMELEN);
		fc++, argc++;
	}
	fr=0;
	while (fr<fc) {
		_parse_file(filearray[fr++]);
		if (isPrintSeparator && fr!=fc) {
			int i=0;
			while (i<newline) {
				fprintf(stdout,"%s","\n");
				i++;
			}
		}
	}
	
	/* STEP 4: perform checks
	 * exiting without going interactive in case of files execution:
	 * - isBye is set if 'bye' was used in one of the argument files
	 * - exitNow is explicitly set if option -q is used */
	if (exitNow || isBye) goto exitpoint;
	/* print banner 
	 * Note: banner is printed only in case of no files (and if not
	 * disabled by --no-banner) */
	if (!fc && isBanner) print_long_banner();
	else fprintf(stdout,"%s","\n");
	
	/* STEP 5: execute command string, if given */
	if (cline[0]) {
		int i=0;
		while (i<512 && cline[i]) { 
			line_buffer(i)=cline[i];
			i++;
		}
		while (i<512) { 
			line_buffer(i)=0;
			i++;
		}
		to_in_(0);
		_evaluate_input_buffer();
		fprintf(stdout,"%s","\n\n");
	}
	
	/* START IDLE CYCLE 
	 * Note: the STATE and BASE settings MUST be given here, or in case 
	 * of any error during compilation in files, the compilation state
	 * remains active, possibly with a BASE different than 10 or a
	 * STATE different from INTERPRET...*/
	STATE=INTERPRET;
	BASE=10;
	_idle(); // the magic gateway

exitpoint:
	/* CLOSING OPERATIONS */
	write_history(HISTORY_PATH);
	/* reset colors */
	fprintf(stdout,"%s",RESETCOLOR);

	/* record timing */
	if (isTotalTiming) {
		long long int res;
		gettimeofday(&tv, NULL);
        	mendS = (long long int)(tv.tv_sec*1000.0 + tv.tv_usec/1000.0);
		res=(long long int)(mendS-beginS);
		if (res<1000) fprintf(stdout,"\nTiming: %lld milliseconds.\n",res);
		else fprintf(stdout,"\nTiming: %.3f seconds.\n",res/1000.0);
	}
	/* finally, exit */
	fprintf(stdout,"%s","\n");
	return EXIT_SUCCESS;
}
